#include "formula.h"
#include "problem.h"
#include "sorting.h"
#include "strategy.h"
#include <unistd.h>
#include <pthread.h>
#include <stdio.h>
#include <stdlib.h>
#include <time.h>
#include <signal.h>

using namespace std;
 
#define MAX_THREADS 20

//the main mutex on which to lock
pthread_mutex_t count_mutex = PTHREAD_MUTEX_INITIALIZER;

void *full_para(void* ptr);
void *full_unit(void* ptr);
void *full_reso(void* ptr);
void *only_para(void* ptr);
void *full_sema(void* ptr);
void *full_sema_para(void* ptr);
void *just_para(void* ptr);
void leave(int sig);

class StrategyManager {
protected:
  static vector<BasicSOS<WEIGHTq> *> sos;
  static vector<pthread_t*> threads;
  static vector<int> iret;
  static vector<void* (*)(void*)> strategy;
  static bool freed;

public:
  StrategyManager() { freed = false; }

  ~StrategyManager() {
    if (freed) return;
    freed = true;
    
    for (unsigned i = 0; i < sos.size(); i++) {
      if (sos[i] == NULL)
	continue;
      delete sos[i];
      delete threads[i];
    }
  }    

  static bool satisfiable() {
    for (unsigned i = 0; i < sos.size(); i++)
      if (!sos[i]->satisfiable()) return false;
    return true;
  }
 
  //adds a new SOS and associates it with a strategy function 
  int add(BasicSOS<WEIGHTq> *ws, void* (*st)(void*)) {
    sos.push_back(ws);
    iret.push_back(0);
    strategy.push_back(st);
    threads.push_back(new pthread_t);
    return sos.size() - 1;
  }

  void remove(int i) {
    assert(sos[i]);
    delete sos[i];
    delete threads[i];
    sos[i] = NULL;
    threads[i] = NULL;
  }

  //executes the threads and returs a vector with status of each
  static vector<int> &run() {
    for (unsigned i = 0; i < sos.size(); i++)
      if (sos[i] != NULL)
	iret[i] = pthread_create(threads[i], NULL, *(strategy[i]), (void*)sos[i]);
    for (unsigned i = 0; i < sos.size(); i++)
      if (sos[i] != NULL)
	pthread_join(*(threads[i]), NULL);
    return iret;
  }

  int run(int i) {
    if (sos[i] == NULL)
      return -1;
    iret[i] = pthread_create(threads[i], NULL, *(strategy[i]), (void*)sos[i]);
    pthread_join(*(threads[i]), NULL);
    return iret[i];
  }
};

vector<BasicSOS<WEIGHTq> *> StrategyManager::sos;
vector<pthread_t*> StrategyManager::threads;
vector<int> StrategyManager::iret;
vector<void* (*)(void*)> StrategyManager::strategy;
bool StrategyManager::freed;

//all strategy managers need to inherit this class
class StrategyManagerMedium : public StrategyManager {
  protected:	
    vector<int> mythreads;

  public:
   ~StrategyManagerMedium() {  }
  
  void run() {
    for (unsigned j = 0; j < mythreads.size(); j++) {
      int i = mythreads[j];
      if (StrategyManager::sos[i] != NULL)
	StrategyManager::iret[i] = pthread_create(StrategyManager::threads[i],
		       	NULL, *(StrategyManager::strategy[i]), 
			(void*)StrategyManager::sos[i]);
    }
    for (unsigned j = 0; j < mythreads.size(); j++) {
      int i = mythreads[j];
      if (StrategyManager::sos[i] != NULL)
	pthread_join(*(StrategyManager::threads[i]), NULL);
    }
  }    
};

class StrategyManagerModel : public StrategyManagerMedium {
  
public:
  int interpretConjectureTRUE(Problem* problem) {
    TermList cons;
    problem->getConjecture(cons, false);
    for (unsigned i = 0; i < cons.size(); i++) {
      pthread_mutex_lock( &count_mutex );
      if (problem->equalsIndex > -1)
	mythreads.push_back(add((BasicSOS<WEIGHTq> *) 
			new ModelSOS<CRAPq>(problem, cons[i]), &full_para));
      else
	mythreads.push_back(add((BasicSOS<WEIGHTq> *) 
			new ModelSOS<CRAPq>(problem, cons[i]), &full_reso));
      pthread_mutex_unlock( &count_mutex );
    }
    return cons.size();
  }
	
  StrategyManagerModel(Problem* problem) {
    pthread_mutex_lock( &count_mutex );
    if (problem->equalsIndex > -1){
      mythreads.push_back(add((BasicSOS<WEIGHTq> *)new 
			      ModelSOS<WEIGHTq>(problem, false), &full_para));
    } else {
      mythreads.push_back(add((BasicSOS<WEIGHTq> *)new 
			      ModelSOS<WEIGHTq>(problem, false), &full_reso));
    }    
    pthread_mutex_unlock( &count_mutex );
  }
};

class StrategyManagerSemantic : public StrategyManagerMedium {
  
public:
  StrategyManagerSemantic(Problem* problem) {
    TermList cons;
    problem->getConjecture(cons, false);
    for (unsigned i = 0; i < cons.size(); i++) {
      pthread_mutex_lock( &count_mutex );
      	mythreads.push_back(add((BasicSOS<WEIGHTq> *)new 
	    SemanticResolutionSOS<CRAPq>(problem, cons[i]), &full_sema));
      pthread_mutex_unlock( &count_mutex );
    }
  }
};

class StrategyManagerUnitResolution : public StrategyManagerMedium {
public:
  StrategyManagerUnitResolution(Problem* problem) {
    pthread_mutex_lock( &count_mutex );
    mythreads.push_back(add((BasicSOS<WEIGHTq> *)
			    new UnitResolutionSOS<CRAPq>(problem), &full_unit));
    pthread_mutex_unlock( &count_mutex );
  }
};

class StrategyManagerPureEquality : public StrategyManagerMedium {

};

//----------------------------------------------------------------------

void *full_unit(void* ptr) {
  assert(ptr);	  
  UnitResolutionSOS<WEIGHTq> *SoS = (UnitResolutionSOS<WEIGHTq> *)ptr;
  while (!SoS->problem->problemSolved()) {   
    pthread_mutex_lock( &count_mutex );
    if (!SoS->first()) {
      pthread_mutex_unlock( &count_mutex );  
      return NULL;	    
    }	    
    SoS->setFirst();
    SoS->queryForMatch(_resolution);
    SoS->doResolution();
    pthread_mutex_unlock( &count_mutex );  
  }
  return NULL;
}

void *full_sema(void* ptr) {
  assert(ptr);	  
  SemanticResolutionSOS<WEIGHTq> *SoS = (SemanticResolutionSOS<WEIGHTq> *)ptr;
  while (!SoS->problem->problemSolved()) {
    pthread_mutex_lock( &count_mutex );
    if (!SoS->first()) {
      pthread_mutex_unlock( &count_mutex );  
      return NULL;	    
    }	  

    SoS->queryForMatch(_resolution);
    SoS->doResolution();
    pthread_mutex_unlock( &count_mutex );  
  }
  return NULL;
}

void *full_sema_para(void* ptr) {
  assert(ptr);	  
  SemanticResolutionSOS<WEIGHTq> *SoS = (SemanticResolutionSOS<WEIGHTq> *)ptr;
  while (!SoS->problem->problemSolved()) {
    pthread_mutex_lock( &count_mutex );
    if (!SoS->first()) {
      pthread_mutex_unlock( &count_mutex );  
      return NULL;	    
    }	  

    SoS->setFirst();
    if (SoS->doExtraction(_equality_resolution)) {
      pthread_mutex_unlock( &count_mutex );  
      continue;
    }

    ((BasicSOS<WEIGHTq>*)SoS)->queryForMatch(_paramodulation);
    if (((BasicSOS<WEIGHTq>*)SoS)->doResolution(_paramodulation)) {
      pthread_mutex_unlock( &count_mutex );  
      continue;
    }

    SoS->queryForMatch(_resolution);
    SoS->doResolution();

    pthread_mutex_unlock( &count_mutex );  
  }
  return NULL;
}

void *just_para(void* ptr) {
  assert(ptr);	
  BasicSOS<WEIGHTq> *SoS = (BasicSOS<WEIGHTq> *)ptr;
  while (!SoS->problem->problemSolved()) {   
    pthread_mutex_lock( &count_mutex );
    if (!SoS->first()) {
      pthread_mutex_unlock( &count_mutex );  
      return NULL;	    
    }	    
    SoS->setFirst();
 
    //the EQUALITY FACTORING  --- not nessesary, seldom used

    if (SoS->doExtraction(_equality_resolution)) {
      pthread_mutex_unlock( &count_mutex );  
      continue;
    }

    SoS->queryForMatch(_paramodulation);
    SoS->doResolution(_paramodulation);

    pthread_mutex_unlock( &count_mutex );  
  }
  return NULL;
}

void *full_para(void* ptr) {
  assert(ptr);	
  BasicSOS<WEIGHTq> *SoS = (BasicSOS<WEIGHTq> *)ptr;
  while (!SoS->problem->problemSolved()) {   
	  
    pthread_mutex_lock( &count_mutex );
    if (!SoS->first()) {
      pthread_mutex_unlock( &count_mutex );  
      return NULL;	    
    }	    
    SoS->setFirst();
 
    //the EQUALITY FACTORING  --- not nessesary, seldom used

    if (SoS->doExtraction(_equality_resolution)) {
      pthread_mutex_unlock( &count_mutex );  
      continue;
    }

    // if resolvant subsumed stop this resolution
    SoS->queryForMatch(_resolution);
    if (SoS->doResolution(_resolution)) {
      pthread_mutex_unlock( &count_mutex );  
      continue;
    }

    SoS->queryForMatch(_paramodulation);
    SoS->doResolution(_paramodulation);

    pthread_mutex_unlock( &count_mutex );  
  }
  return NULL;
}

void *full_reso(void* ptr) {
  assert(ptr);	
  BasicSOS<WEIGHTq> *SoS = (BasicSOS<WEIGHTq> *)ptr;
  while (!SoS->problem->problemSolved()) {
    pthread_mutex_lock( &count_mutex );
    if (!SoS->first()) {
      pthread_mutex_unlock( &count_mutex );  
      return NULL;	    
    }	    
    SoS->setFirst();
    SoS->queryForMatch(_resolution);
    SoS->doResolution(_resolution);
    pthread_mutex_unlock( &count_mutex );  
  }
  return NULL;
}

void leave(int sig) {
  exit(sig);
}

int main(int argc, char *argv[]) {
  assert(argc == 2);		  
  time_t start, end;
  time(&start);
  (void) signal(SIGXCPU,leave);
  (void) signal(SIGALRM,leave);
  Problem* problem = new Problem(argv[1], false, true);
  StrategyManagerUnitResolution unitman(problem);
  StrategyManagerModel modelman(problem);
  StrategyManagerSemantic seman(problem);
  if (!problem->problemSolved())
    StrategyManager::run();
  time(&end);
  if (problem->problemSolved()) {
    cout << "Proof found in: " << difftime(end, start) << " seconds" << endl;
    cout << "START OF PROOF SEQUENCE" << endl;
    problem->printDerivationTree();
    cout << "END OF PROOF SEQUENCE" << endl;
  } else if (StrategyManager::satisfiable()) {
    cout << "Problem unprovable: " << difftime(end, start) << " seconds" << endl;	  
  } else {
    cout << "Proof not found: " << difftime(end, start) << " seconds" << endl;
  }
  delete problem;
  exit(0);  
}
