/*---------------------------------------------------------------------------*
 * FILE: formula.h
 * AUTHOR: Yury Puzis
 *---------------------------------------------------------------------------*/
#ifndef _DATA_STRUCTURES_H_
#define _DATA_STRUCTURES_H_

#define MAX_VARS_IN_FORMULA 500
#define MAX_UNIVERSAL_VARS 1500
#define MAX_VARIABLES MAX_UNIVERSAL_VARS / 3

#include <stdio.h>
#include <iostream>
#include <cctype>
#include <vector>
#include <assert.h>
#include <map>
#include <set>
#include <algorithm>
#include <iterator>
#include "jjheader.h"
#include <string>
#include <math.h>
#include <ext/hash_map>
#include <ext/hash_set>
#include <queue>

using namespace std;
using __gnu_cxx::hash_map;
using __gnu_cxx::hash_set;
using __gnu_cxx::hash;

class Formula;

template <class T>
class FormulaList;

//important!!! when creating "new TermList" it is nessesary to define # of
//elements, even if it is 0 (a.k.a. new TermList(0))

typedef FormulaList<Formula*> TermList;
typedef TermList * TERMLIST;
typedef FormulaList<int> VarList;
typedef VarList * VARLIST;
typedef FormulaList<VARLIST> VarList_List;
typedef VarList_List * VARLIST_LIST;
typedef FormulaList<bool> FlagList;
typedef FlagList * FLAGLIST;

enum unitype {
  _symbol,	//does not differentiate predicates, functors and negation
  _connective,  //does not differentiate conjuction and disjunction
  _variable     //does not differentiate beetween existential and universal
};

enum FaustFormulaType {
   _negation,
   _conjunction,
   _disjunction,
   _universal,
   _existential,
   _predicate,
   _functor,
   _generic	//anything
};

enum FaustInferenceRules {
  _rewrite,
  _preprocessing,
  _cnf_conversion,
  _cnf_refutation,
  _resolution,
  _paramodulation,
  _factorization,
  _forward_subsumption_resolution,
  _skolemization,
  _equality_resolution
};

class CompareFormulaeByAdress {
 public:
  CompareFormulaeByAdress() { }

  bool operator() (Formula* formula1, Formula* formula2) {
    return formula1 < formula2;
  }
};

struct FormulaHash {
  int operator () (Formula* t) const {
    return (int) t;
  }
};

typedef hash_set<Formula*, FormulaHash> FSet;
typedef FSet * FSET;
typedef FSet::iterator FSetIter;

class HB {
 public:
  int name;
  int Arity;
  SYMBOLNODE data;
  FaustFormulaType type;  
  int NumberOfUses;
  bool _equality;
  bool symmetric;              //f.e. equality is simmetric
  int KBOweight;               //not # of literals, arbitrary
  std::string NameSymbol;
  
  ~HB();
  HB(int name, SYMBOLNODE _data, FaustFormulaType ft = _generic);
  HB(int name, char* _NameSymbol, int Arity, FaustFormulaType ft = _generic);
  HB();
  
  void set(SYMBOLNODE _data, FaustFormulaType ft = _generic);
  void set(char* _NameSymbol, int _Arity, FaustFormulaType ft = _generic);
  
  void print();

  void operator = (HB &source);
  
};

typedef int pIndex;

template <class T>
class FormulaList {
 private:
  unsigned capacity;
  unsigned lsize;

 public:
  T* list;
  
  FormulaList();
  FormulaList(int size, T f = 0);
  ~FormulaList();
 
  T operator[](unsigned i);
  bool operator == (FormulaList &t);
  void operator = (FormulaList<T> &that);

  T at(unsigned i);
  void clear();
  void set(unsigned i, T f);
  void set(unsigned i, T f, int times);
  void set (unsigned i, FormulaList<T> &that);
  void push(T f);
  void push(FormulaList<T> &t);
  T pop();
  void erase(unsigned ind);
  unsigned size();
  void grow(unsigned newsize);
  void resize(unsigned newsize);
  void sort();
  void print();
};

class FormulaSignature {
 private:
  map<pIndex, float> count;

 public:
  void add(pIndex new_symbol, float c = 1);
  void add(FormulaSignature* merger);
  void add(Formula* formula);
  void erase(pIndex symbol);
  void clear();
  float find(pIndex target);

  void print();
};

class FormulaTerm;

struct FormulaTermHash {
  int operator () (FormulaTerm* t) const {
    return (int) t;
  }
};

struct VoidHash {
  int operator () (void* t) const {
    return (int) t;
  }
};

class FormulaTerm {
 private:
  FormulaTerm* prev;		//the parent of this node in the tree

  typedef hash_map<FormulaTerm*, FormulaTerm*, FormulaTermHash> tmap;  
  typedef hash_map<pIndex, tmap*, hash<int> > smap;
  typedef hash_map<pIndex, Formula*, hash<int> > pmap;

  //hash_map1 + symbol ---> hash_map2
  //hash_map2 + terms  ---> next stub in tree
  smap *next;			//next term
  //hash_map1 + principal symbol ---> list of formulae (different by varmaps)
  pmap *principal;		//principal symbol
  //vector of CLASUSE+(map of literals to VARMAPs)
  TERMLIST disj;

  int depth;

  //this->f->KBO + KBO of all prev terms up to root
  int KBO;
  int weight;

 public:
  //this is a virtual link, not a hard link (hard is in pmap)
  Formula* f;			//the actual formula (here live terms)
  

  //locking term sequence, adding a new formula (or returning existing one)
  Formula* add(pIndex principal_symbol, vector<HB> &s);

  //ending literal sequence with a new formula
  //no need to check if one with similar varmap exists - 
  //that would mean the formulae are exactly the same, which cannot
  //be because one would subsume the other
  Formula* add(pIndex principal_symbol, vector<HB> &s, VARLIST_LIST vars);

  FormulaTerm* nextNode(Formula* g);
  FormulaTerm* prevTerm() { return prev; }
  bool isTerm() { return (prev); }
  FormulaTerm(Formula* f, FormulaTerm* prev, int depth, int prevKBO, int prevWeight);
  FormulaTerm() : prev(prev), next(0), principal(0), disj(0), depth(0), 
    KBO(0), weight(0), f(0) {}
  ~FormulaTerm();
};

typedef hash_set<FormulaTerm*, FormulaTermHash> TSet;
typedef TSet * TSET;

struct FInference {
  FormulaList<FaustInferenceRules> *inference;
  TERMLIST formula;

  FInference() { 
    formula = new TermList; 
    inference = new FormulaList<FaustInferenceRules>(); 
  }
  ~FInference() {
    delete formula;
    delete inference;
  }
  int size() {
    return formula->size();
  }
  void push(Formula* f) { 
    formula->push(f); 
  }
  void push(FaustInferenceRules r) { 
    inference->push(r); 
  }
};
 
//each thread uses Customs by adding it's own sos adress to the hashes
class Customs {
 private:
   //thread that considers this formula for his own leaves a mark here
  hash_set<void*, VoidHash> crossed_in;
  hash_set<void*, VoidHash> crossed_out;
 
 public:
  void arrive(void* p) { crossed_in.insert(p); }
  void depart(void* p) { crossed_out.insert(p); }
  
  bool checked_in(void* p) { return crossed_in.find(p) != crossed_in.end(); }
  bool stays_in(void* p) { return checked_in(p) && !checked_out(p); }
  bool checked_out(void* p) { return crossed_out.find(p) != crossed_out.end();}
  bool stays_out(void* p) { return !checked_in(p) && !checked_out(p); }
};

class Formula {
 private:
  Formula* id;			//clause that subsumes this
  int getFlattermHelper(TERMLIST flatterm, VARLIST vmap, int index);

 public:
  static int AD;                //global time
  int ADlocal;

  pIndex parent;                //also can be used as KBO's p 
  FormulaTerm *term;		//arguments lined list (leaf to root)
  //variable data, same order as traversing the literals from leaf to root
  VARLIST_LIST vars;  
  //Feature vector (occurence of each symbol)
  FormulaSignature* signature;
  //depth in the proof tree. in vars - sequential ordering
  //in "for" formulae - index of original f'
  int depth;
  int weight;		        //number of symbols  
  double crappiness;
  bool symmetric;               //f.e. equality is simmetric
  int wKBO;
  int pKBO;
  int arity;			//number of arguments / literals
  FaustFormulaType ftype;       //same as the one in parent (for ease of access)
  vector<FInference *> ancestor;
  TERMLIST bflatterm;
  //FINISH later - not a priority
  //hash_set<FormulaTerm*, FormulaTermHash> anc;  //linked list of ancestors


  //works like stamp in passport - you come in you get a stemp,
  //you come out you get a stemp, but the first one is not removed
  Customs TBU;
  //if TRUE then checked_in, if false then stays_in (in principle 
  //threads should mark just one of the two, but still can change their mind)
  Customs interpretation;

  Formula(pIndex parent, FaustFormulaType _type, int wKBO, int pKBO, int weight,
	  FormulaTerm* newTerm, VARLIST_LIST newMap, int arity, bool sym);
  Formula();
  ~Formula();

  FormulaSignature* symbolCount();  //returns occurence of each symbol  
  //for second order formulae this is used to point to formula that 
  //subsumed this one. for fanctors this is used for demodulation.
  Formula* getId();
  void setId(Formula *clause) { 
    id = clause; 
    if (bflatterm != NULL) {
      delete bflatterm;
      bflatterm = NULL;	
    }
  }
  //no need to check if this one is subsumed by anything - not 
  //supposed to be (neither is the subsumer)
  bool isSubsumed() {  return id != this; }
  /* 0 = unknown, >0 = true (depth/age),   <0 = false, negation is true
     negative formulas will not always get the <0 value when supposed to,
     since it is computationaly expensive (need to search) but it will be
     updated whenever possible */
  void setDepth(int depth);
  //flatten the formula, rename variables according to data in vars
  void getFlatterm(TermList &flatterm, VarList &lit_loc);
  void makeFlatterm();
  Formula* addAnc(FInference *a) { 
    ancestor.push_back(a); 
    return this; 
  }
  TERMLIST getAnc(unsigned i) { 
    assert(i < ancestor.size());
    return ancestor[i]->formula; 
  }
  FormulaTerm* firstTerm();

};

class CompareFormulaeByArity {
 public:
  CompareFormulaeByArity() { }

  bool operator() (Formula* formula1, Formula* formula2) {
    return formula1->arity > formula2->arity;
  }
};

//typedef priority_queue<Formula*, vector<Formula*>, CompareFormulaeByArity> ARITYq;
typedef vector<Formula*> ARITYq;

class FeatureIndexNode {
 public:
  map<int, FeatureIndexNode*> next;
  TERMLIST flist;
 FeatureIndexNode() : flist(0) { }
  ~FeatureIndexNode();   
};

typedef map<int, FeatureIndexNode*> fin;

class FeatureIndex {
 private:  
  vector<fin::iterator> iterstack;
  VarList indexstack;
  FormulaList<FeatureIndexNode*> nstack;    
  FeatureIndexNode root;
  int from, to, uniStart;

 public:
  //CONNECTIVES_NUM to LS, but skip from uniStart to uniStart+MAX_UNIVERSAL_VARS
  FeatureIndex(int from, int to, int uniskip) 
    : from(from), to(to), uniStart(uniskip) { } 

  void add(Formula* f);
  //adds to fset more formulas
  //fb=true=larger or equal=backward
  //fb=false=smaller or equal=forward
  //  void query(bool fb, int arity, FormulaSignature &fsig, FSet &fset);
  void query(bool fb, int arity, FormulaSignature &fsig, ARITYq &fset);
  
};



#endif
