#include "problem.h"
 
/*
  --------- structure of the database ---------
  AND
  FOR --- original disjunctions
  OR
  PREDICATE
  ...
  ~PREDICATE
  ...
  FUNCTOR
  ...
  SCHOELOM
  ...
*/
Problem::Problem(char* filename, bool SplitConjunctions, bool deepParsing) 
  : filename(filename), TrySubsumption(deepParsing) {

  assert(filename);
  //loading the problem
  Signature = NewSignature();
  SetNeedForNonLogicTokens(0);
  assert(Head = ParseFileOfFormulae(filename, NULL, Signature, 1, NULL));
  ConvertToECNF(&Head, Signature, SplitConjunctions);

  //creating 2nd order symbols
  symbols.push_back(HB(0, CON_AND, -1, _conjunction));
  andStore = 0;
  symbols.push_back(HB(1, CON_FOR, -1, _disjunction)); 
  forStore = 1;
  symbols.push_back(HB(2, CON_ORR, -1, _disjunction)); 
  cnfStore = 2;
  symbols.push_back(HB(3, CON_AND, -1, _conjunction));
  ancStore = 3;		//reserved. unused
  LS = CONNECTIVES_NUM;
  
  empty.parent = cnfStore;
  empty.ftype = _disjunction;

  facts = new TermList;
  questions = new TermList;
  contra = new TermList;
  formulae = new TermList;

  //creating predicate symbols
  equalsIndex = -1;
  int largestArity = 0;
  for (SYMBOLNODE node = NextInSignatureTree(Signature->Predicates, NULL); 
       node != NULL; node = NextInSignatureTree(Signature->Predicates, node)){
    symbols.push_back(HB(LS, node, _predicate));
    //allow autosorting of 'equal' predicate (to normalize)
    symbols[LS].symmetric = false;
    if (strcmp(node->NameSymbol, "$equal") == 0) {
      symbols[LS]._equality = true;
      symbols[LS].symmetric = true;
      equalsIndex = LS;
    }
    if (node->Arity > largestArity) largestArity = node->Arity;
    LS++;    
  }
  //creating negations for predicates
  predSymbols = LS - CONNECTIVES_NUM;
  for (unsigned i = CONNECTIVES_NUM; i < predSymbols + CONNECTIVES_NUM; i++) {
    assert(LS < 1000);
    symbols.push_back(HB(LS, CON_NOT, 1, _negation));
    LS++;
  }
  //creating functor symbols
  for (SYMBOLNODE node = NextInSignatureTree(Signature->Functions, NULL); 
       node != NULL; node = NextInSignatureTree(Signature->Functions, node)) {
    symbols.push_back(HB(LS, node, _functor));
    LS++;
    if (node->Arity > largestArity) largestArity = node->Arity;
  }
  funcSymbols = LS - CONNECTIVES_NUM - predSymbols * 2;
  //setting KBO ordering (w)  
  for (unsigned i = CONNECTIVES_NUM; i < LS; i++) {
    symbols[i].KBOweight = 2;
    // symbols[i].KBOweight = predSymbols + funcSymbols + 4 + 2 - i;
  }

  int uniStart = LS;
  //setting up variables
  for (int i = 0; i < MAX_UNIVERSAL_VARS; i++) {
    symbols.push_back(HB(LS, CON_ALL, -2, _universal));
    univars[i].parent = LS++;
    univars[i].depth = i;
    univars[i].weight = 1;
    univars[i].crappiness = 0.25;    
    univars[i].ftype = _universal;
    univars[i].term = &dataStore[3];
  }
  stack = new VARSTACK[MAX_VARS_IN_FORMULA];
  last_in_stack = 0;

  //  PrintListOfAnnotatedTSTPNodes(stdout,Signature,Head,tptp,1);

  //parsing the formulae
  int i = 0;
  for (LISTNODE JJformula = Head; JJformula; JJformula = JJformula->Next, i++){
    Formula* newf = parse(JJformula->jjnext);
    newf->depth = i;
    //nullify last_in_stack (might be >0 if non-quantified vars encountered)
    last_in_stack = 0;
    //add disjunction    
    if (symbols[newf->parent].type != _disjunction) 
      newf = makeFOFDisjunction(newf);
    newf->depth = i;
    formulae->push(newf);
    //store status of the formula (axiom, conjecture ... )
    status.push_back(JJformula->AnnotatedFormula->AnnotatedFormulaUnion.\
		     AnnotatedTSTPFormula.Status);
    cname.push_back(JJformula->AnnotatedFormula->AnnotatedFormulaUnion.\
		    AnnotatedTSTPFormula.Name);
  }
  
  //at this point LS is at it's max (not more scoelomizations)
  //so can create the feature index
  if (deepParsing) {
    featureIndex = new FeatureIndex(CONNECTIVES_NUM, LS, uniStart);
    i = 0;
    for (LISTNODE JJformula = Head; JJformula; JJformula = JJformula->Next, i++){
      Formula* newf = formulae->at(i);
      if (status[i] != conjecture) {
	if (status[i] == negated_conjecture) 
	  addConjunction(*newf, newf, *contra);	  
	else
	  addConjunction(*newf, newf, *facts);
	continue;
      }
      //store in the list of conjectures
      questions->push(newf);
      //negate and store in the list of negated conjectures (using ConvertECNF)
      NegateListNode(&JJformula, Signature, SplitConjunctions);
      Formula* con_newf = parse(JJformula->jjnext);
      if (symbols[con_newf->parent].type != _disjunction)
	con_newf = makeFOFDisjunction(con_newf);
      con_newf->depth = i;
      addConjunction(*con_newf, con_newf, *contra);
    }  
  }
  else {
    featureIndex = NULL;
  }

  delete [] stack;
}

Problem::~Problem() {
  delete formulae;
  delete facts;
  delete contra;
  delete questions;

  if (featureIndex)
    delete featureIndex;
  FreeListOfAnnotatedFormulae(&Head);
  FreeSignature(&Signature);
}
//--------------------------------------------------------------------------

/* ways to use:
 *						      > addAsTerm > add 
 * addConjunction > addForeign > createFlatterm > addFlatterm > subsume > add
 * ^----------------^-----------------------------^-----------------------^----
 * ^ new from jj    ^ from	 ^ to rename 	  ^ from		  ^from
 * 		      another	   varialbles	    resolution		   pars
 * 		      Problem	   				      (vanilla)
 * 		      or fresh
 * 		      from JJ
 */

/* accepts 
  1. parent of the list
  2. current position in flatterm
  3. a list of terms/predicates
  4. varset mapping 
  5. ancestors [if it is term]
  replace iterator with index
*/

  //assumes there is no _exactly_ identical formula (same set of literals)
  //(which also means no need to 1) look it up 2) sort literals )
Formula* Problem::addFlatterm(pIndex parent, int &i, TermList &flatterm,
			      VarList_List &varset) {
  VARLIST_LIST vars = NULL;
  int arity;
  if (parent == cnfStore) {
    arity = varset.size();
    if (arity == 0)
      return &empty;
    vars = new VarList_List(arity);
    //if this is a disjunction with no literals it is a contradiction
  } else {
    arity = symbols[parent].Arity;
  }
  FormulaTerm* terms;
  if (parent == cnfStore || parent == forStore || parent == andStore) {
    terms = &dataStore[parent];
    for (int k = 0; k < arity; k++) {
      pIndex p = flatterm[i]->parent;
      Formula* f = addFlatterm(p, ++i, flatterm, varset);
      terms = terms->nextNode(f);
      //if it is a predicate, store it's variable mapping pattern
      //(even if it has not terms, or just no variables!  - need consistency)
      vars->set(k, varset[arity-k-1]);
    }
    return terms->add(parent, symbols, vars);
  }
  terms = &dataStore[3];
  for (int k = 0; k < arity; k++)
    if (!termChar((int)flatterm[i])) {                   //a variable
      terms = terms->nextNode(&univars[(int) flatterm[i]]);
      i++;
    } else { 						//add the term
      pIndex p = flatterm[i]->parent;
      terms = terms->nextNode(addFlatterm(p, ++i, flatterm, varset));
    }
  return terms->add(parent, symbols);
}


Formula* Problem::addFlatClause(TermList &flatterm, 
				VarList_List &varset, 
				FInference* ancestors) {
  //adding the flatterm as formula
  int i = 0;
  Formula* f = addFlatterm(cnfStore, i, flatterm, varset);
  //storing flatterm by default. will get removed on demand
  //setting ancestors
  if (ancestors)
    f->addAnc(ancestors);
  //indexing (subsumption)
  featureIndex->add(f);
  return f;
}

Formula* Problem::doSubsumption(Unifier* unifier, FInference *ancestors, TermList &subsumed_clauses) {
  //--- check if this is a contradiction
  int arity = unifier->active_literals();
  if (arity == 0)
    return NULL;
  int old_arity = arity;
  //--- create ancestors list for this resolvant
  Formula* best_result = NULL;
  //subsumption
  FormulaSignature fsig;
  unifier->getSignature(fsig);
  unifier->subsumptionON();  
  unifier->push_state();  
  unifier->create_state();
  //the sets are sorted s.t. shorted formulas come first
  //this way we get forward subsumption resolution && backward
  //subsumption terminate faster
  ARITYq forwardSet, backwardSet;
  featureIndex->query(false, arity, fsig, forwardSet);
  //FORWARD + F-SUBS-RESOLUTION (trying to kill the new clause)
  ::sort(forwardSet.begin(), forwardSet.end(), CompareFormulaeByArity());
  for (int j = 0; j < forwardSet.size(); j++) {
    Formula* i = forwardSet[j];
    if (j > 0 && i == forwardSet[j-1])
      continue;
    //if arity of the candidate is bigger stop [list is sorted by arity]
    //(can be because arity of *i changes because of forward_subs_resolution)
    if (i->arity > arity) {
      break;
    }
    if (unifier->forwSubsume(i, ancestors->formula)) {
      best_result = i;
      break;
    }
    //need to update arity because forward_subs_res could have droped some literals
    arity = unifier->active_literals();
    //if just killed all the last literal just quit
    if (arity == 0)
      break;	
  }
  //if was forward-resolved, record it
  if (arity < old_arity)
    ancestors->push(_forward_subsumption_resolution);
  //if wasn't subsumed...  //BACKWARD (trying to kill something old)
  if (arity > 0 && !best_result) {	 
    featureIndex->query(true, arity, fsig, backwardSet);
    ::sort(backwardSet.begin(), backwardSet.end(), CompareFormulaeByArity());
    for (int j = 0; j < backwardSet.size(); j++) {
      Formula* i = backwardSet[j];
      if (j > 0 && i == backwardSet[j])
	continue;
      if (unifier->backSubsume(i))
	subsumed_clauses.push(i);
    }
  }
  unifier->subsumptionOFF();
  return best_result;
}

//  only 1 ancestor because more then 1 can only be if it is derived
//  while this one is transformed from ancestor
Formula* Problem::addForeign(Formula* f, Formula* ancestor) {
  TermList flatterm;
  //extracting flatterm representation, fixing variable names, and simplifying
  OneState sm[30];
  Unifier u(f, &symbols, sm);
  u.push_state();
  if (u.simplify())
    return NULL;
  //subsumption step.
  TermList subsumed_clauses;
  VARLIST_LIST varset = new VarList_List;
  FInference* ancestors = new FInference;
  ancestors->push(ancestor);
  ancestors->push(_rewrite);
  if (doSubsumption(&u, ancestors, subsumed_clauses) != NULL) {
    delete ancestors;
    delete varset;
    return NULL;
  }   
  u.createResolvant(flatterm, varset);
  f = addFlatClause(flatterm, *varset, ancestors);
  //set id's of subsumed clauses
  for (unsigned i = 0; i < subsumed_clauses.size(); i++)
    subsumed_clauses[i]->setId(f);
  delete varset;
  return f;
}

void Problem::addConjunction(Formula &f, Formula* p, TermList &out){
  Formula *newf;
  //case 1: bare predicate
  FaustFormulaType ftype = symbols[f.parent].type;
  if (ftype == _predicate || ftype == _negation)
    newf = makeFOFDisjunction(&f);
  else if (ftype == _disjunction && f.arity == 1 
	     && symbols[f.term->f->parent].type == _conjunction)
    newf = f.term->f;
  else 
    newf = &f;

  if (symbols[newf->parent].type == _conjunction) {
    for (FormulaTerm* t = newf->term; t && t->isTerm(); t = t->prevTerm())
      addConjunction(*t->f, p, out);
    return;
  }
  if ((newf = addForeign(newf, p)) == NULL) 
    return;
  setDepth(newf, 1);
  out.push(newf);
}

Formula* Problem::makeFOFDisjunction(Formula* f) {
  //  note: do not try to print f at this stage - it is not finished parsing
  FormulaTerm* newt = dataStore[forStore].nextNode(f);
  return newt->add(forStore, symbols, mirrorVars(newt));
}

void Problem::setDepth(Formula* f, int newDepth) {
  if (newDepth == 0) return;
  f->depth = newDepth;
  if (symbols[f->parent].type == _negation)
    setDepth(f->term->f,  -1 * newDepth);
  if ((symbols[f->parent].type == _disjunction && f->arity == 1) ||
      (symbols[f->parent].type == _conjunction))
    for (FormulaTerm* t = f->term; t && t->isTerm(); t = t->prevTerm())
      setDepth(t->f, newDepth);	    
}

//void Problem::addAncestors(Formula* f, TERMLIST ancestors) {
//  f->ancestor.insert(ancestors);
//}
//--------------------------------------------------------------------------

FaustFormulaType Problem::getType(pIndex h) {
  return symbols[h].type;
}

Formula& Problem::getClause(unsigned ClauseNumber) { 
  assert(ClauseNumber < formulae->size());
  return *(formulae->at(ClauseNumber)); 
}

void Problem::getProblem(TermList &list) {
  list.push(*formulae);
}

void Problem::getFacts(TermList &list) {
  list.push(*facts);
}

void Problem::getConjecture(TermList  &list, bool sign) {
  if (sign)
    list.push(*questions);
  else
    list.push(*contra);
}

bool Problem::problemSolved() {
  return (empty.ancestor.size() > 0);
}

StatusType Problem::getClauseStatus(unsigned ClauseNumber) {
  assert(ClauseNumber < formulae->size());
  return status[ClauseNumber];
}

char* Problem::getClauseName(unsigned ClauseNumber) {
  assert(ClauseNumber < formulae->size());
  return cname[ClauseNumber];
}
 
pIndex Problem::getOpposite(pIndex h) {
  assert (symbols[h].type == _negation || symbols[h].type == _predicate);
  if (symbols[h].type == _negation)
    return (h - predSymbols);
  return (h + predSymbols);
}

 
//-------------------------------------------------------------------

FSET Problem::RootTable::get(Formula* f, Formula* g) {
  hash_map<FormulaPair, FSET, FormulaPair>::iterator rlist;
  rlist = roots.find(FormulaPair(f, g));
  if (rlist != roots.end())
    return rlist->second;
  return NULL;
}

void Problem::RootTable::add(Formula* f, Formula* g, FSET t) {
  roots[FormulaPair(f, g)] = t;
}

Problem::RootTable::~RootTable() {
  hash_map<FormulaPair, FSET, FormulaPair>::iterator rlist;
  for (rlist = roots.begin(); rlist != roots.end(); rlist++)
    delete rlist->second;
}

Problem::FormulaPair::FormulaPair(Formula *f, Formula* g) {
  if (f < g) {
    this->f = f;
    this->g = g;
  }
  else {
    this->f = g;
    this->g = f;
  }
  hash = 37 * (int) this->f + (int) this->g;
}


Problem::SoloTable::~SoloTable() {  
  hash_map<Formula*, FSET, FormulaHash>::iterator rlist;  
  for (rlist = roots.begin(); rlist != roots.end(); rlist++)
    delete rlist->second; 
}

FSET Problem::SoloTable::get(Formula* f) {
  hash_map<Formula*, FSET, FormulaHash>::iterator rlist;  
  rlist = roots.find(f);
  if (rlist != roots.end())
    return rlist->second;
  return NULL;
}

void Problem::SoloTable::add(Formula* f, FSET t) {
  roots[f] = t;
}

//-------------------------------------------------------------------

void Problem::FormulaBag::queryHelper(FSet &curset, FSet &fout, FSet &ftbu) {
  FSetIter j = curset.begin();
  for (FSetIter i = j; i != curset.end(); i = j) {
    j++;
    Formula* g = (*i)->getId();
    if (*i != g) {      //subsumed
      Formula* m = NULL;
      if (j != curset.end())
	m = *j;	
      ftbu.insert(g);
      curset.erase(i); 
      if ((*i)->bflatterm) {
        delete (*i)->bflatterm;
        (*i)->bflatterm = NULL;
      }
      if (m)
	j = curset.find(m);
      else
	return;
    } else {
      fout.insert(*i);
    }
  }
}

void Problem::FormulaBag::query(Formula* f, FSet &fout, FSet &ftbu) {
  //get the opposite for each predicate
  for(FormulaTerm* t = f->term; t && t->isTerm(); t = t->prevTerm())
    queryHelper(vset->at(p->getOpposite(t->f->parent)), fout, ftbu);
}

void Problem::FormulaBag::queryEquality(Formula* f, FSet &fout, FSet &ftbu) {
  f->makeFlatterm();
  bool inEquals = false;
  for (int i = 0; i < f->bflatterm->size(); i++) {
    if (!termChar((int)f->bflatterm->at(i)))
      continue;
    pIndex par = f->bflatterm->at(i)->parent;
    if (f->bflatterm->at(i)->ftype == _functor) {
      queryHelper(eset->at(par), fout, ftbu);
      if (inEquals)
	queryHelper(vset->at(par), fout, ftbu);
    } else 
      inEquals = detectInEquals(f->bflatterm->at(i));
    if (f->bflatterm->at(i)->ftype == _negation)
      i++;
  }
  fout.erase(f);
}

//making sure that $equals (but not ~$equals) queries for anything
inline bool Problem::FormulaBag::detectInEquals(Formula* t) {
  if (t->ftype == _negation)
    return false;
  if (t->parent == p->equalsIndex)
    return true;   
  return false;  
}

void Problem::FormulaBag::add(Formula* f) {
  f->makeFlatterm();
  bool inEquals = false;
  for (int i = 0; i < f->bflatterm->size(); i++) {
    if (!termChar((int)f->bflatterm->at(i)))
      continue;
    pIndex par = f->bflatterm->at(i)->parent;
    if (inEquals)
      (*eset)[par].insert(f);
    else
      (*vset)[par].insert(f);
    if (f->bflatterm->at(i)->ftype == _predicate || f->bflatterm->at(i)->ftype == _negation) {
      inEquals = detectInEquals(f->bflatterm->at(i));
      if (f->bflatterm->at(i)->ftype == _negation)
	i++;
    }
  }
}

Problem::FormulaBag::FormulaBag(Problem *p) : p(p) {
  vset = new vector<FSet>(p->symbols.size());  
  eset = new vector<FSet>(p->symbols.size());  
}

Problem::FormulaBag::~FormulaBag() {
  delete vset;
  delete eset;
}

void Problem::FormulaBag::erase(Formula* f) {
  for (FormulaTerm* t = f->term; t && t->isTerm(); t = t->prevTerm())
    (*vset)[t->f->parent].erase(f);
}

//--------------------------------------------------------------------------

void Problem::printPredicateDetails(int i) {
  cout << symbols[i].NameSymbol << "/" << symbols[i].Arity 
       << " => references (+" 
       << symbols[i].NumberOfUses << ", -" 
       << symbols[i+predSymbols].NumberOfUses 
       << ") KBO ("
       << symbols[i].KBOweight << ", -"
       << symbols[i+predSymbols].KBOweight
       << ")"  << endl;
}
void Problem::printFunctorDetails(int i) {
  cout << symbols[i].NameSymbol << "/" << symbols[i].Arity 
       << " => references: " 
       << symbols[i].NumberOfUses
       << " KBO : " 
       << symbols[i].KBOweight
       << endl;
}

void Problem::print() {
  cout << "PROBLEM FILE: " << filename << endl;
  cout << "--- Problem -----------------" << endl;
  PrintListOfAnnotatedTSTPNodes(stdout, Signature, Head, tptp, 1);
  cout << "--- Formulas (all types) -----------------" << endl;
  //  printFormulaList(&formulae);
  cout << endl;
  cout << "--- Conjecture -----------------" << endl;
  //  printFormulaList(questions);
  cout << endl;
  cout << "--- Conjecture (negated) -----------------" << endl;
  printFormulaList(contra);
  cout << endl;
  cout << "--- Parsed formulae (Facts) -----------------" << endl;
  printFormulaList(facts);
  cout << endl << "--- Statistics ---------------" << endl;
  cout << "(Conjunctions): " << symbols[andStore].NumberOfUses << endl;
  cout << "(Disjunctions): " << symbols[forStore].NumberOfUses << endl;
  cout << "  Disjunctions: " << symbols[cnfStore].NumberOfUses << endl;
  cout << "    Predicates: " << predSymbols << endl;
  for (unsigned i = CONNECTIVES_NUM; i < predSymbols + CONNECTIVES_NUM; i++)
    printPredicateDetails(i);
  cout << "      Functors: " << funcSymbols << endl;
  int scNum = LS - funcSymbols - CONNECTIVES_NUM - predSymbols * 2;
  for (unsigned i = CONNECTIVES_NUM + predSymbols * 2; i < LS - scNum; i++)
    printFunctorDetails(i);
  cout << "     Schoeloms: " << scNum - MAX_UNIVERSAL_VARS  << endl;
  for (unsigned i = funcSymbols + CONNECTIVES_NUM + predSymbols * 2 ; i<LS;i++) {
    if (symbols[i].type != _universal)
      printFunctorDetails(i);
  }
  cout << endl << "--- Solutions ----------------" << endl;
  printDerivationTree(&empty);
  cout << endl << "------------------------------" << endl;
}

void Problem::constructSymbol(pIndex p, char* buffer, int &i) {
   for(unsigned j = 0; j < symbols[p].NameSymbol.length(); j++, i++)
     buffer[i] = tolower(symbols[p].NameSymbol[j]); 
}

void Problem::constructFlatLogic(Formula* f, char* buffer, int &i) {
  if (symbols[f->parent].type == _disjunction && f->arity == 0) {
    sprintf(buffer + i, "$false");
    i += 6;
    return;
  }    
  TermList flatterm;

  VarList lit_loc;  //dummy
  f->getFlatterm(flatterm, lit_loc);
  vector<int> ket;         //closing bracket
  vector<char> con;        //connective for this group
  ket.push_back(flatterm.size() - 1);
  buffer[i++] = '(';
  if (symbols[f->parent].type == _disjunction)
    con.push_back('|');
  else if (symbols[f->parent].type == _conjunction)
    con.push_back('&');
  else 
    con.push_back(',');
  for (unsigned j = 0; j < flatterm.size(); j++) {
    //write variable
    int fv = (int) flatterm[j];      
    //    if (fv < MAX_VARS_IN_FORMULA && fv >= 0) {
    if (!termChar(fv)) {
      if (fv < 26)
	buffer[i++] = (char) (fv + 65);
      else {
	int quot = div(fv, 26).quot;
	buffer[i++] = (char) (quot + 65);
	buffer[i++] = (char) (fv - quot * 26 + 97);
      }
	
    }
    else {
      //write symbol
      FaustFormulaType ftype = symbols[flatterm[j]->parent].type;
      if (ftype != _disjunction && ftype != _conjunction)
	constructSymbol(flatterm[j]->parent, buffer, i);
      //start new "("
      if (ftype != _negation && flatterm[j]->arity > 0) {
	buffer[i++] = '(';
	ket.push_back(j + flatterm[j]->weight - 1);
	if (ftype == _conjunction)
	  con.push_back('&');
	else if (ftype == _disjunction)
	  con.push_back('|');
	else
	  con.push_back(',');
	continue;
      }
      if (ftype == _negation)
	continue;
    }
    //end ")"
    while (ket.size() > 0 && ket.back() == j) {
      buffer[i++] = ')';
      ket.pop_back();
      con.pop_back();
    }
    //connect
    if (con.size() > 0)
      buffer[i++] = con.back();
  }
  //  for (int j = 0; j < ket.size(); j++)
  //    buffer[i++] = ')';
}

Formula* Problem::getDisjunction(Formula* f) {
  if (symbols[f->parent].type == _disjunction)
    return f;
  return dataStore[cnfStore].nextNode(f)->add(cnfStore, symbols, NULL);
}

int Problem::printFormulaName(Formula* f, char* buffer) {
  if (f->parent == forStore) {
    sprintf(buffer, "%s", getClauseName(f->depth));
    return strlen(getClauseName(f->depth));
  } else if (f == &empty) {
    sprintf(buffer, "contradiction");
    return 13;
  } else if (f == &theoryequality) {
    sprintf(buffer, "theory(equality)");
    return 16;
  }  else {
    sprintf(buffer, "%d", abs((int)f));
    return 9;
  }
}

int Problem::printFormulaInference(Formula* f, char* buffer){
  FInference* fi = f->ancestor[0];
  int j = 0;
  for (unsigned i = 0; i < fi->inference->size(); i++) {
    if (i > 0) {
      sprintf(buffer + j, "__");
      j += 2;
    }
    switch(fi->inference->at(i)) {
    case _rewrite:
      sprintf(buffer + j, "rewrite");
      j += 7;
      break;
    case _preprocessing:
      sprintf(buffer + j, "preprocessing");
      j += 13;
      break;
    case _cnf_conversion:
      sprintf(buffer + j, "cnf_conversion");
      j += 14;
      break;
    case _cnf_refutation:
      sprintf(buffer + j, "cnf_refutation");
      j += 14;
      break;
    case _resolution:
      sprintf(buffer + j, "resolution");
      j += 10;
      break;
    case _paramodulation:
      sprintf(buffer + j, "paramodulation");
      j += 14;
      break;
    case _factorization:
      sprintf(buffer + j, "factorization");
      j += 13;
      break;
    case _skolemization:
      sprintf(buffer + j, "skolemization");
      j += 13;
      break;
    case _forward_subsumption_resolution:
      sprintf(buffer + j, "forward_subsumption_resolution");
      j += 30;
      break;
    case _equality_resolution:
      sprintf(buffer + j, "equality_resolution");
      j += 19;
    }
  }
  return j;
}

void Problem::constructParents(Formula* f, char* buffer, int &i) {
  //  TERMLIST chosenAncestor = *f->ancestor.begin();
  TERMLIST chosenAncestor = f->getAnc(0);
  sprintf(buffer + i, ",[status(thm)],[");
  i += 16;
  i += printFormulaName(chosenAncestor->at(0), buffer + i);
  //  sprintf(buffer + i, "%d", (int)(chosenAncestor->at(0)));
  //  i += 9;
  for (unsigned k = 1; k < chosenAncestor->size(); k++) {
    buffer[i++] = ',';
    i += printFormulaName(chosenAncestor->at(k), buffer + i);
    /*
    sprintf(buffer + i, ",%d",(int)(getDisjunction(chosenAncestor->at(k))));
    i += 10;
    */
  }  
  sprintf(buffer + i, "]");
  i += 1;
}

void Problem::printFormula(Formula* f, bool extra) {
  assert(f);
  //turn formula to formatted string
  char buffer[100024];
  //  stack = new VARSTACK[MAX_VARS_IN_FORMULA];
  last_in_stack = 0;
  int i = 0;
  sprintf(buffer, "cnf(");
  i += 4;  
  i += printFormulaName(f, buffer + i);
  if (f->ancestor.size() > 0) {
    sprintf(buffer + i, ",derived,");
    i += 9;
    constructFlatLogic(f, buffer, i);
    if (extra) {
      sprintf(buffer + i, ",inference(");
      i += 11;
      i += printFormulaInference(f, buffer + i);
      constructParents(f, buffer, i);
      sprintf(buffer + i, "),[]).");
      i += 6;
    } else {
      sprintf(buffer + i, ").");
      i += 2;
    }
  } else {
    sprintf(buffer + i, ",initial,");
    i += 9;
    constructFlatLogic(f, buffer, i);
    if (extra) {
      sprintf(buffer + i, ",file('%s',unknown),[]).", filename);
      i += 22 + strlen(filename);
    } else {
      sprintf(buffer + i, ").");
      i += 2;
    }
  }  
  buffer[i+1] = '\0';
  cout << buffer << endl;

  /*
  //send it over a stream to JJParser
  FILE* MyStream = fmemopen (buffer, strlen (buffer), "r");
  READFILE Stream = OpenFILEFile(MyStream);
  //ANNOTATEDFORMULA AnnotatedFormula = ParseAnnotatedFormula(Stream,Signature);
  ANNOTATEDFORMULA AnnotatedFormula = ParseAndUseAnnotatedFormula(Stream,Signature);
  
  //output the formula from jjparser
  PrintAnnotatedTSTPNode(stdout, AnnotatedFormula, tptp, 1);

  //release all memeory
  FreeAnnotatedFormula(&AnnotatedFormula);
  fclose(MyStream);
  CloseFile(Stream);
  */
  /*
  for (int k = 0; k < last_in_stack; k++) {
    delete [] stack[k].jjnode->NameSymbol;
    delete stack[k].jjnode;
  }
  delete [] stack;
  */
}

void Problem::printDerivationTree(Formula* f) {
  FSet tree;
  if (f == NULL) 
    makeDerivationTree(&empty, tree);
  else
    makeDerivationTree(f, tree);
}

void Problem::makeDerivationTree(Formula* f, FSet &tree) {
  //1. do not add/print ancestors of something already in tree (infi. loop)
  //   x + y = z AND z + t = x
  if (!tree.insert(f).second)
    return;
  //if there are any ancestors, print them first
  if (f->ancestor.size()> 0) {
    //    TERMLIST chosenAncestor = *f->ancestor.begin();
    TERMLIST chosenAncestor = f->getAnc(0);
    for (unsigned k = 0; k < chosenAncestor->size(); k++) {
      Formula* g = chosenAncestor->at(k);
      //skip 'theory(equality)' record (will look like ancestor)
      if (g == &theoryequality)
	continue;
      g = getDisjunction(g);
      makeDerivationTree(g, tree);
    }
  }
  //print this one
  printFormula(f, true);
  cout << endl;
}

void Problem::printFormulaList(TERMLIST l, bool extra){
  for (unsigned i = 0; i < l->size(); i++) {
    printFormula((*l)[i], extra);
    cout << endl;
  }
}

void Problem::storeString(int ClauseNumber, char *symbol){
  ClauseNumber++;
  AddUsefulInformationToAnnotatedFormula(
    GetAnnotatedFormulaFromListByNumber(Head,ClauseNumber),Signature,symbol);
}

//------------------------------------------------------------------------:w

VARLIST Problem::mirrorVarsHelper(Formula* f, VARLIST v) {
  if (symbols[f->parent].type == _universal) {
    if ((unsigned) f->depth >= v->size()) 
      v->resize(f->depth + 1);
    v->set(f->depth, f->depth);
    return v;
  } 
  for (FormulaTerm* t = f->term; t && t->isTerm(); t = t->prevTerm())
    mirrorVarsHelper(t->f, v);
  return v;
}

//mirrorVars takes a set of literals and creates VARMAP for them
//by assigning each variable to it's f->depth. [for f's from JJParser]
//if literal has no variables, it still has VarList, only empty
VARLIST_LIST Problem::mirrorVars(FormulaTerm *t) {
  if (!t) return NULL;
  VARLIST_LIST vars = new VarList_List;
  while (t->f) {
    assert(t->f->ftype != _universal);
    vars->push(mirrorVarsHelper(t->f, new VarList));
    t = t->prevTerm();
  }
  return vars;
}

pIndex Problem::find(SYMBOLNODE s) {
  pIndex res;
  for (res = CONNECTIVES_NUM; res != symbols.size(); res++)
    if (symbols[res].data == s) return res;
  return -1;
}

Formula* Problem::parse (FORMULA f) {
  assert(f);
  Formula* temp_result;
  FormulaTerm* t;
  BinaryFormulaType fbin = f->FormulaUnion.BinaryFormula;
  switch (f->Type) {
  case quantified:
    stack[last_in_stack].jjnode = f->FormulaUnion.QuantifiedFormula.\
      Variable->TheSymbol.Variable->VariableName;
    if (f->FormulaUnion.QuantifiedFormula.Quantifier == existential) {
      //skoelomizing - construct skoelom's terms
      FormulaTerm* terms = &dataStore[3];
      int termsArity = 0;
      for (int i = 0; i < last_in_stack; i++)
	if (symbols[stack[i].fnode->parent].type == _universal) {
	  terms = terms->nextNode(&univars[i]);
	  termsArity++;
        }
      //add a new HB skoelom (actually, _existential (for future ref.))
      symbols.push_back(HB(LS, stack[last_in_stack].jjnode, _functor));
      symbols[LS].NameSymbol = symbols[LS].data->NameSymbol;
      symbols[LS].KBOweight = 2;
      symbols[LS].Arity = termsArity;
      //store this info in temporary place
      stack[last_in_stack].fnode = terms->add(LS, symbols);
      stack[last_in_stack].fnode->parent = LS;
      LS++;
    } else {
      stack[last_in_stack].fnode = &univars[last_in_stack];
    }
    last_in_stack++;
    temp_result = parse(f->FormulaUnion.QuantifiedFormula.Formula);
    last_in_stack--;
    return temp_result;
    break;
  case binary:
    if (fbin.Connective == conjunction) {
      t = parse(&dataStore[andStore], fbin.LHS, fbin.RHS, fbin.Connective);
      return t->add(andStore, symbols, NULL);
    }
    t = parse(&dataStore[forStore], fbin.LHS, fbin.RHS, fbin.Connective);
    return t->add(forStore, symbols, mirrorVars(t));
  case unary:  /* assumes negation is the only unary operator */
    //look for the negation in the symbol list (loc of predicate + #predicates)
    t = dataStore[3].nextNode(parse(f->FormulaUnion.UnaryFormula.Formula));
    return t->add(t->f->parent + predSymbols, symbols);
  case atom:
    return parse(f->FormulaUnion.Atom);
  default: ; }
  assert(false);
}

Formula* Problem::handleArgument(TERM arg) {
  if (arg->Type != variable)   //non variables
    return parse(arg);
  //variables
  SYMBOLNODE thisSymbol = arg->TheSymbol.Variable->VariableName;
  bool found = false;
  int p;
  for (p = 0; p < last_in_stack; p++)
    if (stack[p].jjnode == thisSymbol) {
      found = true;
      break;	  
    }
  if (found)
    return stack[p].fnode;
  //this variable is not quantified --- treat as universal
  //later - allow this only in cnf nodes
  stack[last_in_stack].jjnode = thisSymbol;
  stack[last_in_stack].fnode = &univars[last_in_stack];
  Formula* f = &univars[last_in_stack];
  last_in_stack++;
  return f;
}

Formula* Problem::parse(TERM t) {
  assert(t);
  int sarity = t->TheSymbol.NonVariable->Arity;
  pIndex thesymbol = find(t->TheSymbol.NonVariable);
  if (sarity == 0)
    return dataStore[3].add(thesymbol, symbols); 
  FormulaTerm* terms = &dataStore[3];
  if (thesymbol == equalsIndex) {
    Formula* f1 = handleArgument(t->Arguments[0]);
    Formula* f2 = handleArgument(t->Arguments[1]);
    if (f2->wKBO < f1->wKBO) {
      //      cout << "o 1" << endl;
      terms = terms->nextNode(f1);
      terms = terms->nextNode(f2);
    } else {
      terms = terms->nextNode(f2);
      terms = terms->nextNode(f1);
    }
  } else {
    for (int i = 0; i < sarity; i++)
      terms = terms->nextNode(handleArgument(t->Arguments[i]));
  }
  return terms->add(thesymbol, symbols);
}

FormulaTerm* Problem::parse(FormulaTerm* t, 
			    FORMULA left, FORMULA right, ConnectiveType cont) {
  if (left->Type == binary && left->jjf.Connective == cont)
    t = parse(t, left->jjf.LHS, left->jjf.RHS, cont);
  else
    t = t->nextNode(parse(left));

  if (right->Type == binary && right->jjf.Connective == cont)
    t = parse(t, right->jjf.LHS, right->jjf.RHS, cont);
  else
    t = t->nextNode(parse(right));

  return t;
}
