#include "strategy.h"
#include "sorting.h"

template <class R>
bool BasicSOS<R>::satisfiable() {
  return TBU->size() == 0;	
}

template <class R>
BasicSOS<R>::BasicSOS(Problem* problem) : problem(problem) {
  problem->getFacts(ToBeUsed);
  problem->getConjecture(ToBeUsed, false);
  first_chosen = second_chosen = NULL;
  unifier = NULL;
  TBU = new R;
  CBUsubset = new R;
  CBUset = problem->newBag();
  saved_resolve = 0;
  activePreds.resize(problem->symbols.size());
  for (unsigned i = 0; i < activePreds.size(); i++)
    activePreds.set(i, false);
}
  
template <class R>
BasicSOS<R>::~BasicSOS() { 
  delete TBU;
  delete CBUsubset;
  delete CBUset;
  if (unifier)
    delete unifier;
}

template <class R>
bool BasicSOS<R>::notPureClause(Formula* f) {
  for (FormulaTerm* t1 = f->term; t1->f; t1 = t1->prevTerm())
    if (activePreds[t1->f->parent] == true)
      return true;
  return false;
}

template <class R>
void BasicSOS<R>::addToActive(Formula* f) {
  for (FormulaTerm* t1 = f->term; t1->f; t1 = t1->prevTerm())
    activePreds.set(problem->getOpposite(t1->f->parent), true);
}

template <class R>
void BasicSOS<R>::removePureClauses() {
  bool updated = true;
  TermList TBUsmall;
  for (unsigned i = 0; i < ToBeUsed.size(); i++)
    TBUsmall.push(ToBeUsed[i]);
  ToBeUsed.resize(0);
  while(updated == true) {      
    updated = false;
    for (unsigned i = 0; i < TBUsmall.size(); i++) {
      if (notPureClause(TBUsmall[i])) {
	addToActive(TBUsmall[i]);
	ToBeUsed.push(TBUsmall[i]);
	TBUsmall.erase(i);
	updated = true;
	break;
      }
    }
  }
}

template <class R>
bool BasicSOS<R>::first() {
  //discarding identical/subsumed clauses.
  Formula *g;
  first_chosen = NULL;
  while (!TBU->empty()) {
    g = TBU->top()->getId();
    TBU->pop();
    //if f==g then we are choosing unsubsumed f
    //if f!=g then we are choosing subsumer of f, but only if
    //it has never been chosen before (foreign or elsewhere in TBU)
    if (!g->TBU.checked_out((void*)this)) {
      first_chosen = g;
      break;
    }
  }
  if (first_chosen == NULL)
    return false;
  //check in before checking out
  putInCBU(first_chosen);
  return true;
}

template <class R>
void BasicSOS<R>::setFirst(Formula* f) {
  if (f != NULL) first_chosen = f;
  if (unifier) delete unifier;   
  unifier = new Unifier(first_chosen, &problem->symbols, state_machine);
}

template <class R>
void BasicSOS<R>::setMatch(Formula* f) {
  if (f != NULL) second_chosen = f;
  unifier->restore_state(-1);
  unifier->setPeer(second_chosen);
  unifier->flush();
}

template <class R>
void BasicSOS<R>::queryForMatch(FaustInferenceRules infrule) {
  while (!CBUsubset->empty())
    CBUsubset->pop();
  FSet ToTBU;
  FSet ToCBU;
  if (infrule == _paramodulation)	//superposition query
    CBUset->queryEquality(first_chosen, ToCBU, ToTBU);
  else if (infrule == _resolution)
    CBUset->query(first_chosen, ToCBU, ToTBU);
  else 
    assert(false);
  
  for (FSetIter i = ToCBU.begin(); i != ToCBU.end(); i++)
    CBUsubset->push(*i);
  //ToTBU subsumed some of those supposed to be in CBUsubset
  for (FSetIter i = ToTBU.begin(); i != ToTBU.end(); i++)
    putInTBU(*i);
  second_chosen = NULL;
}

template <class R>
bool BasicSOS<R>::putInTBU(Formula *f) {
  //this formula is /was in TBU, do not add it
  if (f->TBU.checked_in((void*)this))
    return false;
  TBU->push(f);
  f->TBU.arrive((void*)this);
  return true;
}

template <class R>
void BasicSOS<R>::putInCBU(Formula *f) {
  f->TBU.depart((void*)this);
  CBUset->add(f);
}

template <class R>
FSET BasicSOS<R>::addStored(FaustInferenceRules infrule) {
  FSET new_clauses = NULL;
  switch(infrule) {
  case _resolution:
    new_clauses = problem->fullResolutionTable.get(first_chosen, second_chosen);
    break;
  case _paramodulation:
    new_clauses = problem->paramodulationTable.get(first_chosen, second_chosen);
    break;
  case _equality_resolution:
    new_clauses = problem->equalityResolutionTable.get(first_chosen);
    break;
  default:
    assert(false);
  }
  if (!new_clauses)
    return NULL;
  saved_resolve++;
  for (FSetIter i = new_clauses->begin(); i != new_clauses->end(); i++)
    putInTBU(*i);
  return new_clauses;
}

template <class R>
void BasicSOS<R>::storeAdded(FSET new_clauses, FaustInferenceRules infrule) {
  switch(infrule) {
  case _resolution:
    problem->fullResolutionTable.add(first_chosen, second_chosen, new_clauses);
    break;
  case _paramodulation:
    problem->paramodulationTable.add(first_chosen, second_chosen, new_clauses);
    break;
  case _equality_resolution:
    problem->equalityResolutionTable.add(first_chosen, new_clauses);
    break;
  default:
    assert(false);
  }      
  for (FSetIter i = new_clauses->begin(); i != new_clauses->end(); i++)
    putInTBU(*i);
}

template <class R>
Formula* BasicSOS<R>::generate(FaustInferenceRules infrule, bool &stop_all) {
  
  //do subsumption
  subsumed_clauses.resize(0);
  int old_state = unifier->state;
  bool good_result = false;
  FInference* ancestors = new FInference;
  Formula* best_result = 
    problem->doSubsumption(unifier, ancestors, subsumed_clauses);

  if (!best_result) {
    ancestors->push(first_chosen);
    if (infrule != _equality_resolution)
      ancestors->push(second_chosen);
    ancestors->push(infrule);
    if (infrule == _paramodulation)
      ancestors->push(&problem->theoryequality);
    //--- create the actual resolvant flatterm
    VARLIST_LIST varset = new VarList_List;
    TermList flatterm;
    unifier->createResolvant(flatterm, varset);
    best_result = problem->addFlatClause(flatterm, *varset, ancestors);
    delete varset;
    //as a result of subsumption resolution, contradiction appeared
    if (unifier->active_literals() == 0) {
      stop_all = true;
      return NULL;
    }
    good_result = true;
  } else {
    delete ancestors;
    //since the resolvant was subsumed by something, check if that something 
    //is from another thread or a half/full brother. if it is - take it
    if (!best_result->TBU.checked_in((void*)this) &&
	!best_result->TBU.checked_out((void*)this)) {
      good_result = true;
    }
  }
  //--- update all formulas that this one subsumed
  unsigned scs = subsumed_clauses.size();
  for (unsigned i = 0; i < scs; i++) {
    //if the chosen clause is subsumed by it's child, stop resolution
    //the only clause in new_clauses will be the new resolvant
    if (subsumed_clauses[i] == first_chosen)
      stop_all = true;
    subsumed_clauses[i]->setId(best_result);
  }
  //if subsumption was tried, push was done. pop it
  unifier->restore_state(old_state);          
  //  problem->printFormulaList(&subsumed_clauses);
  if (good_result == true)
    return best_result;
  return NULL;
}    

template <class R>
bool BasicSOS<R>::doResolution(FaustInferenceRules infrule) {
  bool stop_all = false;
  FSET new_clauses;
  while (!stop_all && !CBUsubset->empty()) {
    second_chosen = CBUsubset->top();
    CBUsubset->pop();
    //if this pair was generated by other thread, process without resolution
    if (addStored(infrule))
      continue;
    setMatch();
    new_clauses = new FSet;   
    do {
      if (infrule == _resolution && !unifier->getNextResolvant())
	break;
      else if (infrule == _paramodulation && !unifier->getNextSuperposition())
	break;      
      Formula* newClause = generate(infrule, stop_all);
      //if proof was found return immidiaetly
      if (newClause != NULL) 
 	new_clauses->insert(newClause);      
    } while (!stop_all);
    unifier->restore_state(-1);
    storeAdded(new_clauses, infrule);
  }
  return stop_all;
}

template <class R>
bool BasicSOS<R>::doExtraction(FaustInferenceRules infrule) {
  if (addStored(infrule))
    return false;
  bool stop_all = false;
  FSET new_clauses = new FSet;
  do {
    if (infrule == _equality_resolution && !unifier->getNextEqualityResolvant())
      break;
    Formula* newClause = generate(infrule, stop_all);
    //if proof was found return immidiaetly
    if (newClause != NULL) 
      new_clauses->insert(newClause);
  } while (!stop_all);
  unifier->restore_state(-1);
  storeAdded(new_clauses, infrule); 
  return stop_all;
}

//-----------------------------------------------------------------------------

template <class R>
SimpleSOS<R>::SimpleSOS(Problem* problem) : BasicSOS<R>(problem) {
  for (unsigned i = 0; i < BasicSOS<R>::ToBeUsed.size(); i++)
    BasicSOS<R>::putInTBU(BasicSOS<R>::ToBeUsed[i]);
}

//-----------------------------------------------------------------------------

template <class R>
ModelSOS<R>::ModelSOS(Problem* problem, bool my_sign) : BasicSOS<R>(problem) {
  for (unsigned i = 0; i < BasicSOS<R>::ToBeUsed.size(); i++) {
    bool good = true;
    Formula &ft = *BasicSOS<R>::ToBeUsed[i];
    for (FormulaTerm* t = ft.term; t && t->isTerm(); t = t->prevTerm()) {
      if (t->f->ftype == my_sign) {
	good = false;
	BasicSOS<R>::putInTBU(&ft);
	break;
      }
    }
    if (good) {
      BasicSOS<R>::putInCBU(&ft);
    }
  }
}

template <class R>
ModelSOS<R>::ModelSOS(Problem* problem, Formula* seed):BasicSOS<R>(problem){
  //pure literal deletion
  BasicSOS<R>::addToActive(seed);
  BasicSOS<R>::removePureClauses();  
  for (unsigned i = 0; i < BasicSOS<R>::activePreds.size(); i++)
    BasicSOS<R>::activePreds.set(i, false);
  for (FormulaTerm* t1 = seed->term; t1->f; t1 = t1->prevTerm())
    BasicSOS<R>::activePreds.set(t1->f->parent, true);
  for (unsigned i = 0; i < BasicSOS<R>::ToBeUsed.size(); i++) {
    if (interpretation(BasicSOS<R>::ToBeUsed[i]))
      BasicSOS<R>::putInCBU(BasicSOS<R>::ToBeUsed[i]);
    else
      BasicSOS<R>::putInTBU(BasicSOS<R>::ToBeUsed[i]);
  }
}

template <class R>
bool ModelSOS<R>::interpretation(Formula* f) {
  if (f->interpretation.checked_out((void*)this)) {
    return false;
  }
  if (f->interpretation.checked_in((void*)this)) {
    return true;
  }
  for (FormulaTerm* t1 = f->term; t1->f; t1 = t1->prevTerm())
    if (BasicSOS<R>::activePreds[t1->f->parent] == true) {
      f->interpretation.arrive((void*)this);
      return true;      
    }
  f->interpretation.depart((void*)this);
  return false;
}

//-----------------------------------------------------------------------------

template <class R>
bool UnitResolutionSOS<R>::first() {
  R * lcbu = BasicSOS<R>::CBUsubset;
  do {
    if (!lcbu || lcbu->empty()) {	
      if(!BasicSOS<R>::first())
	return false;      
      BasicSOS<R>::queryForMatch(_resolution);	
    }
    if (lcbu->empty())
      continue;
    BasicSOS<R>::second_chosen = lcbu->top();
    lcbu->pop();
  } while (!BasicSOS<R>::second_chosen);
  BasicSOS<R>::first_chosen = BasicSOS<R>::second_chosen;
  return true;
}

template <class R>
void UnitResolutionSOS<R>::doResolution() {
  vector<Formula*> stackF;
  vector<int> stackI;

  if (CBUsubset2.size() == 0)
    return;
  
  stackF.push_back(BasicSOS<R>::first_chosen);
  stackI.push_back(0);

  while (stackI.size() > 0) {
    BasicSOS<R>::setFirst(stackF.back());	
    int mysecond = stackI.back();
    BasicSOS<R>::setMatch(CBUsubset2[mysecond]);
    stackF.pop_back();
    stackI.pop_back();
    bool stop_all = false;
    FSET new_clauses = BasicSOS<R>::addStored(_resolution);   
    bool generateC = false;
    if (!new_clauses) {
      generateC = true;
      new_clauses = new FSet;
      while(BasicSOS<R>::unifier->getNextResolvant()) {
        Formula* f = BasicSOS<R>::generate(_resolution, stop_all);
	if (BasicSOS<R>::problem->problemSolved())
	  break;
	if (f)
	  new_clauses->insert(f);
      }
      BasicSOS<R>::storeAdded(new_clauses, _resolution);
    }
    for (FSetIter i = new_clauses->begin(); i != new_clauses->end(); i++) {
      if (mysecond + 1 < CBUsubset2.size()) {
	stackF.push_back(*i);
	stackI.push_back(mysecond+1);
      }
      if (generateC == true) {
	putInTBU(*i);
      }
    }	    
  }
}

template <class R>
UnitResolutionSOS<R>::UnitResolutionSOS(Problem* problem) : BasicSOS<R>(problem) { 
  for (unsigned i = 0; i < BasicSOS<R>::ToBeUsed.size(); i++) {
    Formula *ft = BasicSOS<R>::ToBeUsed[i];
    if (ft->arity == 1) {
      putInTBU(ft);
    } else { 
      BasicSOS<R>::putInCBU(ft);
    }
  }	
}

template <class R>
void UnitResolutionSOS<R>::queryForMatch(FaustInferenceRules infrule) {
  CBUsubset2.clear();
  FSet ToTBU, ToCBU;
  //this first_chosen is one of the non-unit clauses extracted
  //after querying the real first_chosen from it's CBUsubset
  BasicSOS<R>::CBUset->query(BasicSOS<R>::first_chosen, ToCBU, ToTBU);
  for (FSetIter i = ToCBU.begin(); i != ToCBU.end(); i++) {      
    if ((*i)->arity == 1)
      CBUsubset2.push(*i);
  }
  //ToTBU subsumed some of those supposed to be in CBUsubset
  //(only unit clauses are allowed in TBU in UR)
  //no need to limit arity here because maxTBUarity is set to 1
  for (FSetIter i = ToTBU.begin(); i != ToTBU.end(); i++)
    putInTBU(*i);
  BasicSOS<R>::second_chosen = NULL;   
}

template <class R>
bool UnitResolutionSOS<R>::putInTBU(Formula* f) {
  if (f->arity > 1)
    return false;
  return BasicSOS<R>::putInTBU(f);    
}

//-----------------------------------------------------------------------------

template <class R>
void SemanticResolutionSOS<R>::queryForMatch(FaustInferenceRules infrule) {
  FSet ToTBU, ToCBU;
  assert(TBUstack.empty());
  //this first_chosen is one of the non-unit clauses extracted
  //after querying the real first_chosen from it's CBUsubset
  BasicSOS<R>::CBUset->query(BasicSOS<R>::first_chosen, ToCBU, ToTBU);
  for (FSetIter i = ToCBU.begin(); i != ToCBU.end(); i++)
    TBUstack.push(*i);
  //need to put all those in TBUstack too because they might be
  //some of those that were dropped after subsuming the original
  for (FSetIter i = ToTBU.begin(); i != ToTBU.end(); i++)
    TBUstack.push(*i);
  BasicSOS<R>::second_chosen = NULL;   
}

template <class R>
void SemanticResolutionSOS<R>::doResolution() {
  FSet TBUpast;
  while (!TBUstack.empty()) {
    BasicSOS<R>::first_chosen = TBUstack.top()->getId();
    TBUstack.pop();    
    //making sure that similar things are not processed
    if (TBUpast.find(BasicSOS<R>::first_chosen) != TBUpast.end())
      continue;
    TBUpast.insert(BasicSOS<R>::first_chosen);
    //getting match for the first chosen
    BasicSOS<R>::queryForMatch(_resolution);
    BasicSOS<R>::setFirst();
    bool stop_all = false;
    FSET new_clauses;
    while (!stop_all && !BasicSOS<R>::CBUsubset->empty()) {
      BasicSOS<R>::second_chosen = BasicSOS<R>::CBUsubset->top();
      BasicSOS<R>::CBUsubset->pop();
      if (ModelSOS<R>::interpretation(BasicSOS<R>::second_chosen))
	continue;
      //if this pair was generated by other thread, process without resolution
      new_clauses = BasicSOS<R>::problem->fullResolutionTable.get(BasicSOS<R>::first_chosen, 
								  BasicSOS<R>::second_chosen);
      if (new_clauses != NULL) {
	for (FSetIter i = new_clauses->begin(); i != new_clauses->end(); i++)
	  if (ModelSOS<R>::interpretation(*i)) 
	    TBUstack.push(*i) ;
	  else
	    BasicSOS<R>::putInTBU(*i);
	continue;
      }
      //else, properly set up the second chosen and start getting results
      BasicSOS<R>::setMatch();
      new_clauses = new FSet;   
      do {
	if (!BasicSOS<R>::unifier->getNextResolvant())
	  break;
	Formula* newClause = BasicSOS<R>::generate(_resolution, stop_all);	
	if (newClause != NULL) { 
	  new_clauses->insert(newClause);
	  if (ModelSOS<R>::interpretation(newClause)) 
	    TBUstack.push(newClause);
	  else
	    BasicSOS<R>::putInTBU(newClause);
	}
      } while (!stop_all);
      BasicSOS<R>::unifier->restore_state(-1);
      BasicSOS<R>::problem->fullResolutionTable.add(BasicSOS<R>::first_chosen, 
						    BasicSOS<R>::second_chosen, new_clauses);
    }
  }
}

template <class R>
bool SemanticResolutionSOS<R>::putInTBU(Formula *f) {
  if (ModelSOS<R>::interpretation(f))
    return false;	    
  return BasicSOS<R>::putInTBU(f);
}

template <class R>
SemanticResolutionSOS<R>::SemanticResolutionSOS(Problem* problem, Formula* seed): ModelSOS<R>(problem, seed) {  }

//-----------------------------------------------------------------------------

template class SemanticResolutionSOS<WEIGHTq>;
template class SemanticResolutionSOS<CRAPq>;
template class UnitResolutionSOS<WEIGHTq>;
template class UnitResolutionSOS<CRAPq>;
template class ModelSOS<WEIGHTq>;
template class ModelSOS<CRAPq>;
template class ModelSOS<AGEq>;
template class BasicSOS<WEIGHTq>;
