

// Code written by Hans de Nivelle, November 2005. 



#ifndef MODEL_INCLUDED
#define MODEL_INCLUDED   1


#include <vector>
#include <iostream>


#include "element.h"
#include "predicate.h"
#include "groundatom.h"




class model 
{

   std::vector< element > arguments; 
      // Arguments of the atoms. 

   std::vector< unsigned int > parents; 
      // Parents of the atoms. 

   struct atom
   {
      bool usedinproof;   

      predicate p; 
      unsigned int args1;
      unsigned int args2;   
         // The arguments are in arguments [ args1 .. args2 ].

      unsigned int rule_nr;
         // Number of rule with which the atom was constructed.

      unsigned int pars1;
      unsigned int pars2;
        // The indices of the parents are in predicates 
	// [ parents1 .. parents 2 ].       

      unsigned int timestamp;
         // This allows the user to see how old an atom is, 
	 // so that he can have some idea where the prover
	 // is spending its time. 

      unsigned int alternative;
      unsigned int nr_alternatives;
         // So that the user can see where we are 
	 // in the search.

      atom( predicate p, 
            unsigned int args1, unsigned args2,
	    unsigned int rule_nr, 
	    unsigned int pars1, unsigned int pars2,
	    unsigned int timestamp,
	    unsigned int alternative,
	    unsigned int nr_alternatives )
         : usedinproof( false ),
	   p(p), args1( args1 ), args2( args2 ),
	   rule_nr( rule_nr ), 
	   pars1( pars1 ), pars2( pars2 ),
	   timestamp( timestamp ),
	   alternative( alternative ),
	   nr_alternatives( nr_alternatives ) 
      {
      }
   };

   std::vector< atom > atoms; 

public:
   unsigned int closure_counter;

   model( ):
      closure_counter(0)
   {
   }


   bool addatom( groundatom at, unsigned int rule_nr, 
		 const std::vector< unsigned int > & parents,
		 unsigned int alternative = 0,
		 unsigned int nr_alternatives = 1 );
      // The atom is only added when it is unique.
      // addatom returns true if the atom is really added.


   bool canbeadded( groundatom at ) const; 
      // True if the atom at can be added to this model. This is the case
      // when all its variables V occur in a create atom #V.


public: 
   bool backtrack( unsigned int s );
      // backtrack( ) backtracks the model, and keeps also track of the
      // used atoms. When a used atom is deleted, its parents
      // are marked as used. 
      // True is returned if one of the deleted atoms
      // was used.

   unsigned int size( ) const
   {
      return atoms. size( );
   } 

   unsigned int nrelements( ) const;
      // Number of elements in the model.


   struct atom_iterator
   {
      std::vector< atom > :: const_iterator a;

      explicit atom_iterator( std::vector< atom > :: const_iterator a )
         : a(a)
      {
      }


      atom_iterator operator ++ ( int )
      {
         return atom_iterator::atom_iterator( a ++ ); 
      }

      atom_iterator operator -- ( int )
      {
          return atom_iterator::atom_iterator( a -- );
      }

      atom_iterator operator ++ ( )
      {
         return atom_iterator::atom_iterator( ++ a ); 
      }

      atom_iterator operator -- ( ) 
      {
         return atom_iterator::atom_iterator( -- a );
      }

      bool operator == ( atom_iterator other ) const 
      {
         return ( this -> a ) == other. a;
      }

      bool operator != ( atom_iterator other ) const 
      {
	 return ( this -> a ) != other. a;
      }

   };


   atom_iterator atoms_begin( ) const 
   {
      return atom_iterator::atom_iterator( atoms. begin( )); 
   }

   atom_iterator atoms_end( ) const 
   {
      return atom_iterator::atom_iterator( atoms. end( ));
   } 

   atom_iterator make_atom_iterator( unsigned int nr ) const
   {
      ASSERT( nr < atoms. size( ));
      return atom_iterator::atom_iterator( atoms. begin( ) + nr );
   }

   groundatom getatom( atom_iterator at ) const 
   {
      return groundatom::groundatom( at. a -> p, 
         arguments. begin( ) + at. a -> args1,
	 arguments. begin( ) + at. a -> args2 );
   }


   unsigned int gettimestamp( atom_iterator at ) const
   {
      return at. a -> timestamp;
   }

   unsigned int getrule_nr( atom_iterator at ) const
   {
      return at. a -> rule_nr;
   }

   bool isused( atom_iterator at ) const
   {
      return at. a -> usedinproof;
   }

   unsigned int getalternative( atom_iterator at ) const
   {
      return at. a -> alternative;
   }

   unsigned int getnralternatives( atom_iterator at ) const
   {
      return at. a -> nr_alternatives;
   }

   void contradiction( const std::vector< unsigned int > & contr ); 
      // Mark the atoms in contr as used. contr stands
      // for contradiction. 



   ///////////////////////////////////////////////////////////////////
   // A parent_iterator iterates through the parents of an atom.    //
   ///////////////////////////////////////////////////////////////////

   struct parent_iterator
   {
      std::vector< unsigned int > :: const_iterator p;

      explicit parent_iterator( 
		          std::vector< unsigned int > :: const_iterator p )
         : p(p)
      {
      }


      parent_iterator operator ++ ( int )
      {
         return parent_iterator::parent_iterator( p ++ ); 
      }

      parent_iterator operator ++ ( )
      {
         return parent_iterator::parent_iterator( ++ p ); 
      }

      bool operator == ( parent_iterator other ) const 
      {
         return ( this -> p ) == other. p;
      }

      bool operator != ( parent_iterator other ) const 
      {
	 return ( this -> p ) != other. p;
      }

      unsigned int operator * ( ) const
      {
         return *p;
      }

   };


   parent_iterator parents_begin( atom_iterator at ) const 
   {
      return parent_iterator::parent_iterator( 
		         parents. begin( ) + at.a -> pars1 ); 
   }

   parent_iterator parents_end( atom_iterator at ) const 
   {
      return parent_iterator::parent_iterator( 
		         parents. begin( ) + at.a -> pars2 );
   } 


};


std::ostream& operator << ( std::ostream& , const model & );


#endif


