

// Code written by Hans de Nivelle, November 2005.
// Modified in June 2006, so that it can deal with
// breadth-first search.



#include "search.h"
#include "calculus.h"
#include "application.h"
#include "subsumption.h"



namespace 
{

   // The returned partial object is always defined.

   partial< application > processnewlemma( 
		    false_instance& lemma, 
		    rulesystem::type_of_derivation derivedwith,
	            std::vector< unsigned int > & parentrules,
	            model& m, rulesystem& rs )
   {
      std::cout << "\n"; 

      std::cout << "Created a lemma:  ";
      std::cout << lemma. r. nr_premisses( ) << ", ";
      std::cout << lemma. s. size( ) << "\n";

      // At this point we have created a new lemma which
      // we are going to add to the rulesystem. 
      // Before we add it, we make a sequence of simplifications. 

      std::vector< predicate > predicates_base_first;
      std::vector< variable > arguments_base_first;

      false_instance first = calculus::sortlemma(
                  predicates_base_first,
		  arguments_base_first,
		  lemma, m, 1 );

      std::cout << "Lemma is sorted:  ";
      std::cout << first. r. nr_premisses( ) << ", ";
      std::cout << first. s. size( ) << "\n";


      std::vector< predicate > predicates_base_red;
      std::vector< variable > arguments_base_red;

      false_instance red = calculus::functional_reduction( 
			 predicates_base_red, 
                         arguments_base_red,
			 first, m );

      std::cout << "Lemma is functionally reduced:  ";
      std::cout << red. r. nr_premisses( ) << ", ";
      std::cout << red. s. size( ) << "\n";


      std::vector< predicate > predicates_base_cond;
      std::vector< variable > arguments_base_cond;

      false_instance cond = calculus::condense(
			 predicates_base_cond,
			 arguments_base_cond,
			 red, m );

      std::cout << "Lemma is condensed:  ";
      std::cout << cond. r. nr_premisses( ) << ", ";
      std::cout << cond. s. size( ) << "\n";


      std::vector< predicate > predicates_base_sort;
      std::vector< variable > arguments_base_sort;

      false_instance sort = calculus::sortlemma( 
	                          predicates_base_sort,
		                  arguments_base_sort,
		                  cond, m, 1 );
     
      std::cout << "Lemma is resorted:  ";
      std::cout << sort. r. nr_premisses( ) << ", ";
      std::cout << sort. s. size( ) << "\n";

      ASSERT( sort. r. iswellformed( )); 

      rs. deletesubsumed( sort. r );
 
      std::cout << "Adding lemma " << rs. size( ) << " : ";
      std::cout << sort.r << "\n\n";

      rs. addrule( sort. r, derivedwith, parentrules );

      std::cout << "\n"; 
      partial< application > res = application::findclosinglemma( 
		                          rs, m, false );
      ASSERT( res. isdefined( ));
      return res; 
   }


   // Create lemma and add it to the rulesystem rs. 

   partial< application > lemmafromdisjunction( 
	         false_instance& disj, 
		 rulesystem::type_of_derivation derivedwith, 
		 std::vector< unsigned int > & parentrules,
		 model& m, unsigned int m_size,
                 rulesystem& rs )
   {
      ASSERT( disj. r. isdisjunctive( ));

      if( disj. r. nr_conclusions( ) == 0 )
      {
         return processnewlemma( disj, derivedwith, parentrules, m, rs ); 
      }

      // We add the first member of the disjunction in order
      // to catch a lemma.

      rule::conc_iterator at = disj. r. concs_begin( );
      m. addatom( disj. s. apply_on( *at ), 1000000, disj. prem_match );
      partial< application > lemma = 
	          application::findclosinglemma( rs, m, true ); 
      m. backtrack( m_size );

      ASSERT( lemma. isdefined( ) ); 
      ASSERT( lemma. contents( ). f. sup_prem_match( ) > m_size ); 

      std::vector< predicate > predicates_base;
      std::vector< variable > arguments_base;

      parentrules. push_back( lemma. contents( ). rule_nr );

      false_instance disj2 = calculus::disjunction_resolvent(
                predicates_base,
	        arguments_base,
	        disj, 0,
	        lemma. contents( ). f, m, m_size );

      return lemmafromdisjunction( disj2, derivedwith, parentrules, 
		                   m, m_size, rs );
   }



   partial< application > lemmafromexistential( 
		 false_instance& existential,
		 rulesystem::type_of_derivation derivedwith, 
		 std::vector< unsigned int > & parentrules, 
		 model& m, unsigned int m_size, 
		 variable quantvar, element sup_elements,
		 rulesystem& rs )
   {

      ASSERT( existential. r. isexistential( )); 

      unsigned int s_size = existential. s. size( ); 
      existential. s. assign( quantvar, sup_elements );
      m. addatom( existential. s. apply_on( existential. r. getcreate( )),
			          100000, existential. prem_match );
      m. addatom( existential. s. apply_on( existential. r. getbody( )),
			          100000, existential. prem_match );

      partial< application > lemma = 
	         application::findclosinglemma( rs, m, true ); 
      m. backtrack( m_size );
      existential. s. backtrack( s_size );

      ASSERT( lemma. isdefined( ) );
      ASSERT( lemma. contents( ). f. sup_prem_match( ) > m_size ); 

      std::vector< predicate > predicates_base;
      std::vector< variable > arguments_base;

      parentrules. push_back( lemma. contents( ). rule_nr );

      ASSERT( calculus::ext_res_possible( existential, 
	             lemma. contents( ). f, m_size ));

      // The existential resolvent is a disjunction 
      // the conclusions of which need to be further resolved away.

      false_instance disj = calculus::existential_resolvent(
           predicates_base,
	   arguments_base,
           existential, lemma. contents( ). f, m, m_size );

      return lemmafromdisjunction( disj, derivedwith, parentrules, 
		                   m, m_size, rs );
   }


}


partial< application >
search::extendstomodel( model& m, rulesystem& rs )
{
   rulestack critical; 
   return extendstomodel( m, rs, critical );
}


partial< application > 
search::extendstomodel( model& m, rulesystem& rs, rulestack& critical )
{

#if 0
   std::cout << "====================\n"; 
   std::cout << "Entering extendstomodel \n";
   std::cout << m << "\n";
#endif

   // We look for closing lemma, not necessarily the best:

   partial< application > existinglemma = 
	           application::findclosinglemma( rs, m, false );

   if( existinglemma. isdefined( ))
   {
      if( m. closure_counter % show_closures == 0 )
      {
         std::cout << "-------------------------------------\n\n";
	 std::cout << rs << "\n\n\n";

	 std::cout << m;

         std::cout << "The closing lemma is " << existinglemma << "\n\n"; 
	 std::cout << "Critical Rules: " << critical << "\n"; 
         std::cout << "\n\n\n\n"; 
      }

#if 1 
      if( m. closure_counter % 10 == 0 )
      {
         std::cout << "closure  " <<  m. closure_counter << ",  ";
	 std::cout << "rule number  ";
	 std::cout << existinglemma. contents( ). rule_nr << ",  ";
	 std::cout << "atoms  " << m. size( ) << ",  "; 
         std::cout << "elements  " << m. nrelements( ) << ".\n"; 
      }
#endif 

      m. closure_counter ++ ; 
      rs. rules [ existinglemma. contents( ). rule_nr ]. nr_closures ++ ;

      return existinglemma. contents( ); 
   }

   partial< application > forward = application::bestforward( rs, m );
  
   if( forward. isundefined( ))
   {
      // This means that there is no forward rule. Since there was 
      // also no closing lemma, we have found a model! 

      return partial< application > :: partial( ); 
   }
 
   application forw = forward. contents( );

   unsigned int m_size = m. size( );
   unsigned int m_nr_elements = m. nrelements( );
   unsigned int s_size = forw. f. s. size( );
   unsigned int rs_size = rs. size( );
   unsigned int critical_size = critical. size( );

   if( forw. f. r. isexistential( ) )
   {
      variable quantvar = * ( forw. f. r. getcreate( ). a1 );
         // The quantified variable.

      // Try existing elements:

      for( unsigned int alt = 0; alt < m_nr_elements; ++ alt )
      {
         element e = element::constructfromunsigned( alt ); 

         forw. f. s. assign( quantvar, e );

	 m. addatom( forw. f. s. apply_on( forw. f. r. getbody( )), 
		     forw. rule_nr, forw. f. prem_match,
		     alt, m_nr_elements + 1 );

	 partial< application > res = extendstomodel( m, rs, critical );

	 if( res. isundefined( ))
            return res;

	 m. backtrack( m_size );
	 forw. f. s. backtrack( s_size );

	 if( res. isdefined( ))
         {
             partial< application > 
             res2 = application::makeclosurewith( m, res. contents( ). f. r, 
		                           res. contents( ). rule_nr ); 
	     if( res2. isdefined( ))
             {
	        std::cout << "sharpening!\n";
	        return res2; 
             }
         }

	 critical. append( res. contents( ). rule_nr ); 
      }

      // Try a new element:

      {
         element e = element::constructfromunsigned( m_nr_elements ); 

         forw. f. s. assign( quantvar, e );
         m. addatom( forw. f. s. apply_on( forw. f. r. getcreate( )), 
		     forw. rule_nr, forw. f. prem_match,
		     m_nr_elements, m_nr_elements + 1 );
         m. addatom( forw. f. s. apply_on( forw. f. r. getbody( )), 
		     forw. rule_nr, forw. f. prem_match,
		     m_nr_elements, m_nr_elements + 1 );

         partial< application > res = extendstomodel( m, rs, critical );

         if( res. isundefined( ))
            return res;

         m. backtrack( m_size );
	 forw. f. s. backtrack( s_size );
  
         if( res. isdefined( ))
         {
            partial< application >
            res2 = application::makeclosurewith( m, res. contents( ). f. r,
			                   res. contents( ). rule_nr ); 
            if( res2. isdefined( ))
            {
               std::cout << "sharpening!\n";
	       return res2; 
            }
         }

	 critical. append( res. contents( ). rule_nr ); 
      }

      std::vector< unsigned int > parentrules;
      parentrules. push_back( forw. rule_nr ); 

      // The existential rule is the first parent of the derivation
      // that we are going to make.

      element newelement = element::constructfromunsigned( m_nr_elements );

      critical. backtrack( critical_size ); 

      return lemmafromexistential( forw. f, 
		       rulesystem::rule_existential, parentrules, 
		       m, m_size,
		       quantvar, newelement, rs );
   }
   else
   {
      unsigned int deg = forw. f. r. nr_conclusions( ); 

      for( unsigned alt = 0; alt < deg; ++ alt ) 
      {
         rule::conc_iterator at = forw. f. r. concs_begin( );
         for( unsigned i = 0; i < alt; ++ i )
            ++ at;
         m. addatom( forw. f. s. apply_on( *at ), 
		     forw. rule_nr, forw. f. prem_match,
		     alt, deg );

         partial< application > res = extendstomodel( m, rs, critical );

	 if( res. isundefined( )) 
            return res;  

	 m. backtrack( m_size );

	 if( res. isdefined( )) 
         {
            partial< application >
            res2 = application::makeclosurewith( m, res. contents( ). f. r,
			                    res. contents( ). rule_nr ); 
	    if( res2. isdefined( ))
            {
               std::cout << "sharpening!\n";
	       return res2;
            }
         }

	 critical. append( res. contents( ). rule_nr ); 
      }

      std::vector< unsigned int > parentrules;
      parentrules. push_back( forw. rule_nr ); 
         // The disjunction is first parent of the derivation 
	 // that we are going to make.

      critical. backtrack( critical_size ); 

      return lemmafromdisjunction( forw. f, 
		       rulesystem::rule_disjunction, parentrules, 
		       m, m_size, rs );

   }
}


