// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/
package org.mathweb.atp;

import java.io.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.jdom.*;

import org.mathweb.core.BasicService;
import org.mathweb.util.*;
import org.mathweb.onto.OWLJDom;

/**
@author Juergen Zimmer University of Edinburgh
@version $Date: 2006/07/28 08:35:19 $ $Revision: 1.63.2.5 $
*/
public class E extends ATP {

  protected boolean verbose=true;

  private String E_HOME = getSystemProperty("e.home", "E_HOME");
  private String[] e_envs = new String[] {"E_HOME="+E_HOME};
  private String eExecutable = findSystemExecutable(E_HOME, "eprover", "EProver.exe");
  //  private String eExecutable = findSystemExecutable(E_HOME, "eproof", "EProof.exe");

  private static final int defaultNice = 15;

  private static Logger logger = Logger.getLogger("Ep");
  private static String logFileName = getSystemProperty("e.log.file", "E_LOG_FILE");
  private static FileAppender appender = newAppender(logFileName);
  
  public E()
  {
    super();
    name = "EP_0_91";
    /* The MathServ ontology class that is accepted by the service as an input */
    acceptedConcept = "TstpProblem";
    /* The MathServ ontology class of the OWL individual that is returned by the service */
    resultConcept = "FoAtpResult";
    /* The MathServ ontology class of the proof object. */
    proofConcept = "TstpCnfRefutation";
    /* The ID of the calculus individual in the MathServ ontology representing the calculus of the prover. */
    calculusConcept = "ep_res_calc";
    
    logger.setLevel(Level.ALL);
    logger.setLevel(Level.OFF);
    if (!logger.getAllAppenders().hasMoreElements()) logger.addAppender(appender);
  }  
  
  /** 
   * @param problemFile a file containing a TSTP proving problem.
   * @param language the language the proving problem is in. Must be TPTP.
   * @param format the TSTP format, either "CNF" or "FOF".
   * @param time the time limit for the theorem prover in seconds.
   * @param options additional command line options for the prover command if needed.
   * @return A String array of size 2 where the first entry is the status determined by the ATP system. 
   *  The second entry can contain a proof object in TSTP or XTSTP format.
   */
  public String[] callProver(File problemFile, String language, String format, int time, String options) throws Exception
  {
    File inputFile = null;
    File outputFile = null;
    try{
      outputFile = File.createTempFile("eprover", ".out");
            
      if (language.equals(OWLJDom.MW_NS_STR+"TSTP")) {
        
//         System.out.println("probfile: " + problemFile);
//         System.out.println("probfile: " + readFile(problemFile));
        
        /*  first transform the problem into TPTP CNF form without include statements */
        String toSyntax= "tptp";
        String transform = "none";
        String newProb;

        inputFile = tstp2X.tstp2X(problemFile, toSyntax, transform);
        //        logger.info("new problem file: " + inputFile);
        String p = readFile(inputFile);
        //        System.out.println("new problem: " + p );
        
        /*  write the new problem description in the input file */
        //        writeFile(inputFile,newProb);
        
        /*  create the shell command to call the prover and execute it */
        //        String eproofOptions = " --print-statistics -xAuto -tAuto --memory-limit=Auto --tstp-in --tstp-out ";
        //        String eOptions = " -s --print-statistics  -xAuto -tAuto --memory-limit=Auto --tstp-in ";
        String eOptions = " -s --print-statistics  -xAuto -tAuto --memory-limit=1512 --tstp-in ";
        //String eOptions = " -s --print-statistics -xAuto -tAuto --tstp-in --tstp-out -l4 --memory-limit=100 --cpu-limit=" + time + " ";        
        
        String command = eExecutable + " " + options + " " + eOptions + " " +
          inputFile.getAbsolutePath() + " > " + outputFile.getAbsolutePath() + " 2>&1";
                  
        //        logger.info("[E] command: " + command);
        String[] commands = {"sh", "-c", command};
        //        logger.info("[E] starting prover...");
        this.start(command, e_envs, defaultNice, time);
        //myThread = this.startThreaded(command, e_envs, defaultNice, time, true);

                
        /*  read the output written by the prover call */
        //        logger.info("[E] reading result..." + outputFile.getAbsolutePath());
        if (outputFile.exists() && (outputFile.length() > ATP.maxOutputLength)){
          System.out.println("length: " + outputFile.length());
          return  new String[] {"Unknown", null, null};
        }
        String output = this.readFile(outputFile);
        
        //        logger.info("[E] output: " + output);
        //         logger.info("[E] transforming to TSTP...");
        //         String tstp = x2tstp.x2tstp(output, "E---0.82");
        //         logger.info("[E] TSTP..." + tstp);
        //         return tstp;
        
        /*  split the output in lines and examine them to extract the status */
        String[] split = output.split("\n");
        StringBuffer buf = new StringBuffer();
        String status = null;
        
        for(int i=0; i<split.length; i++){
          //logger.info("[line] " + split[i]);
          if (split[i].length()>18){
            //              logger.info("[line] " + split[i].substring(0,2));
            //              System.out.println("status " + split[i].substring(0,20));
            if (status == null)
              if ((split[i].substring(0,20).equals("# TSTP exit status: "))){
                //              System.out.println("found status ");
                status = split[i].substring(20);
              } else 
                if ((split[i].substring(0,14).equals("# SZS status: ")))
                  status = split[i].substring(14);
                else 
                  if (split[i].startsWith("# Cannot determine problem status: Terminated by signal."))
                    status = "Timeout";
            
            //             if ((split[i].substring(0,3).equals("cnf"))){
            //               buf.append(split[i]);
            //               buf.append("\n");
            //             }
            
          }
        }
        if (status == null) status = "Unknown";
//         else if (status.equals("Saturation")) status = "CounterSatisfiable";
//         else if (status.equals("Refutation")) status = "Unsatisfiable";
        //        System.out.println("status " + status);
        
        /*  create and return the resulting String array */
        //        System.out.println("output " + resArray);
        return new String[] {status, buf.toString(), output};
      } else{
        throw new Exception("Input language "+language+" not known or not supported.");
      }
    } catch(Exception e){
      logger.debug("[E] Exception " + e);
      throw e;
    } finally{
      try{ 
        if (inputFile != null) inputFile.delete();
        if (outputFile != null) outputFile.delete();
      } catch(Exception e){ logger.debug("[E] Exception " + e); }
    }
    
  }
  
  public static void main(String args[]) throws Exception {
    String msHome = System.getProperty("mathserv.home");
    E e = new E();
    OWLJDom owlJDom = OWLJDom.getInstance();
    TSTPCreator creator = new TSTPCreator();

    String timeInd; 
    //    timeInd  = owlJDom.createTimeIndi(300,300);

    timeInd  = owlJDom.createTimeIndi(5,5);
    //       String problem = e.readFile(servF1);
    //String servF1 = msHome + "/examples/owl/cnf.owl";
    String servF1 = msHome + "/examples/owl/fof3.owl";
    String ind = e.readFile(servF1);
    
    //    timeInd =  e.readFile(msHome + "/examples/owl/time.owl");
    
    //    ind = e.readFile(msHome + "/examples/owl/cnf3.owl");
    //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/LCL/LCL181-3.p");
    //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRP/GRP012-2.p");
    //      ind = creator.tstp2OWLProblem("/home/jzimmer/t.p");
    ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/NLP/NLP123+1.p");
      //      ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SYN/SYN052+1.p");
    //    ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/ALG/ALG104+1.p");
    // ind = creator.tstp2OWLProblem("/home/jzimmer/systems/TPTP-v3.1.0/Problems/SYN/SYN830-1.p");
    //    ind = creator.tstp2OWLProblem("/home/jzimmer/casc/Active/Problems/EPS/NoTransform/tptp/SYN/SYN853-1+short+ran.p");
    // ind = creator.tstp2OWLProblem(msHome + "/foo.tstp");
    //ind = creator.tstp2OWLProblem("/home/jzimmer/error.p");
    //ind = creator.tstp2OWLProblem(msHome +"/examples/tptp/_10_t5_thm.p");
    
    //ind = creator.tstp2OWLProblem("/home/jzimmer/bath.p");
      //    String tptpProb = readFile(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRP/GRP012-2.p");
      String tptpProb = readFile("/home/jzimmer/core1.p");
      tptpProb = readFile("/home/jzimmer/import/MathServe/examples/tptp/RelationsAndOrders_Nat_trans.tptp");
      
      //tptpProb = readFile("/home/jzimmer/out");
    String result;
    //result = e.proveTPTPProblem(tptpProb, new Integer(60));
    result = e.proveProblem(ind, timeInd);
    System.out.println(result);
    //     try {
//       //String language = "TSTP";
//       //String result = e.callProver(ind, language, 300, "");
//       String result = e.proveProblem(ind, timeInd);
//       //String result = e.proveProblemXml(ind, timeInd);
//       e.writeFile(msHome + "/examples/owl/e-proof2.owl", result);
      
//       System.out.println("result: " + result);
//     }
//     catch(Exception ex){
//       ex.printStackTrace();
//     }
      
//       String listFileName = args[1];
//       String prefix = args[2];
//       String timeStr = args[3];
//       int time =  Integer.parseInt(timeStr);
//       String timeInd = owlJDom.createTimeIndi(time, time);
      
//       String[] problems = e.readFile(listFileName).split("\n");
//       for(int i = 0; i< problems.length; i++){        
//         String probFile = prefix + "/" + problems[i].trim();
//         //        System.out.println("prob: " + probFile);
//         System.gc();
//         String prob = creator.tstp2OWLProblem(probFile);
        
//         try {
//           String result = e.proveProblem(prob, timeInd);
//           Element resultRoot = owlJDom.getRootElement(result);
//           Double wcDouble = (Double) OWLJDom.TIME_RESOURCE_WC.numberValueOf(resultRoot);
//           Double cpuDouble = (Double) OWLJDom.TIME_RESOURCE_CPU.numberValueOf(resultRoot);
//           String status = OWLJDom.STATUS.valueOf(resultRoot);
//           if ((status == null) || (status.equals("")))
//             status = "ResourceOut";
          
//           String rdfID = OWLJDom.POTENTIAL_ID.valueOf(resultRoot);
//           System.out.println(problems[i].trim() + "\t" + cpuDouble + "\t" + wcDouble + "\t" + status);
//         } catch (Exception ex) {
//           System.err.println(ex.toString());
//         }
//       }
      /* runt org.mathweb.atp.E foo /home/jzimmer/casc/unsolved-solved-by-e /home/jzimmer/casc 600 */
  }
  
  
} // end E.java

      
/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */


    /**   Special settings for non-auto mode without splitting!: */
//     eOptions = " --tptp-in --tstp-out --forward-context-sr -l4 --print-statistics --destructive-er-aggressive --destructive-er -winvfreqrank -c1 -Ginvfreq -F1 -WSelectMaxLComplex --prefer-initial-clauses -H'(4*Refinedweight(SimulateSOS,1,1,2,1.5,2),3*Refinedweight(PreferNonGoals,1,1,2,1.5,1.5),1*Clauseweight(PreferProcessed,1,1,1),1*FIFOWeight(PreferProcessed))' --memory-limit=192 --delete-bad-limit=150000000 --cpu-limit=" + time + " ";

//       /**
// --forward-context-sr
// --destructive-er-aggressive
// --destructive-er
// -winvfreqrank -c1 -Ginvfreq
// -F1
// -WSelectMaxLComplex
// --prefer-initial-clauses
// -H'(4*Refinedweight(SimulateSOS,1,1,2,1.5,2),3*Refinedweight(PreferNonGoals,1,1,2,1.5,1.5),1*Clauseweigh\
// t(PreferProcessed,1,1,1),1*FIFOWeight(PreferProcessed))'
// --memory-limit=192
// --delete-bad-limit=150000000
//       */

        /*  Command:   eproof --print-statistics -xAuto -tAuto --memory-limit=Auto --tstp-in --tstp-out %s*/
