// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

package org.mathweb.atp;

import java.io.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout;

import org.mathweb.util.TSTPCreator;
import org.mathweb.core.BasicService;
import org.mathweb.onto.OWLJDom;


/**
@author Juergen Zimmer University of Edinburgh
@version $Date: 2006/05/26 13:47:08 $ $Revision: 1.45 $
*/
public class Otter extends ATP {
  
  public String OTTER_HOME = getSystemProperty("otter.home", "OTTER_HOME");
  public String TPTP_HOME = getSystemProperty("tptp.home.dir", "TPTP.HOME.DIR");
  public String[] otter_envs = new String[] {"OTTER_HOME="+OTTER_HOME};
  public String otterExecutable = findSystemExecutable(OTTER_HOME, "otter", "Otter.exe");
  public String limitedRunExecutable = findSystemExecutable(TPTP_HOME+"/SystemExecution", "TreeLimitedRun", "TreeLimitedRun.exe");

  private static Logger logger = Logger.getLogger("Otter");
  private static String logFileName = getSystemProperty("otter.log.file", "OTTER_LOG_FILE");
  private static FileAppender appender = newAppender(logFileName);
    
  private static final int defaultNice = 15;
  
  public Otter()
  {
    super();
    name = "Otter_3_3";
    acceptedConcept = "TstpProblem";
    resultConcept = "FOBrFPResult";
    proofConcept = "TstpCnfBrFPRefutation";
    calculusConcept = "otter_calc";
    
    logger.setLevel(Level.ALL);
    logger.setLevel(Level.OFF);
    if (!logger.getAllAppenders().hasMoreElements()) logger.addAppender(appender);
  }
  
  public String[] callProver(File problemFile, String language, String format, int time, String options) throws Exception
  {
    File inputFile = null;
    File outputFile = null;
    try{
      outputFile = File.createTempFile("otter", ".out");
      
      String otterOptions = options;
      //      System.out.println(inputFile + " " + outputFile);
      
      if (language.equals(OWLJDom.MW_NS_STR+"TSTP")) {
        inputFile = tstp2X.tstp2X(problemFile, 
                                  "\"otter:hypothesis:set(auto),clear(print_given)\"",
                                  "add_equality:r");
        //        logger.debug("Otter format: " + inputFile);
        
        
//         String command = limitedRunExecutable + " -q4 " + time + " " + time+1 + " " +otterExecutable + " " +  otterOptions + " " +
//           "< " + inputFile.getAbsolutePath() + " > " + outputFile.getAbsolutePath() + " 2>&1";
        String command = otterExecutable + " " +  otterOptions + " " +
          "< " + inputFile.getAbsolutePath() + " > " + outputFile.getAbsolutePath() + " 2>&1";
        logger.debug("[OTTER] command: " + command);
        String[] commands = {"sh", "-c", command};
        logger.debug("[OTTER] starting prover...");
        //        myThread = this.startThreaded(command, otter_envs, defaultNice, time, true);
        //      this.exec(commands, otter_envs, false, true);
        this.start(command, otter_envs, defaultNice, time);

        logger.debug("reading result..." + outputFile.getAbsolutePath());
        if (outputFile.exists() && (outputFile.length() > ATP.maxOutputLength)){
          System.out.println("length: " + outputFile.length());
          return  new String[] {"Unknown", null, null};
        }
        
        String output = this.readFile(outputFile);
        //        logger.debug("[OTTER] output..." + output);

        /* Hack: The character 0x7 is in the Otter output and is not allowed in XML */
        byte[] bytes = output.getBytes();
        for(int i=0; i < bytes.length; i++)
          if (bytes[i] == 7) bytes[i] = 32;
        
        logger.debug("[OTTER] transforming to TSTP...");
        String tstp = x2tstp.x2tstp(output, "Otter---3.3");
        //        System.out.println("[OTTER] TSTP:\n" + tstp);
        String status = extractStatusFromTSTP(tstp);
        return new String[] {status, tstp+"\n", new String(bytes)};
      } 
      else{
        throw new Exception("Input format "+language+" not known or not supported.");
      }
    } catch(Exception e){
      logger.debug("[Otter] Exception " + e);
      throw e;
    } finally{
      try{ 
        //        if (inputFile != null) inputFile.delete();
        if (outputFile != null) outputFile.delete();
      } catch(Exception e){ logger.debug("[Otter] Exception " + e); }
    }
  }
  
  
  public static void main(String argv[]) throws Exception {
    String msHome = System.getProperty("mathserv.home");
      Otter e = new Otter();
      OWLJDom owlJDom = OWLJDom.getInstance();
      TSTPCreator creator = new TSTPCreator();
      String timeInd = owlJDom.createTimeIndi(20,20);
      //      timeInd = e.readFile(msHome + "/examples/owl/fault-time.owl");
      String servF1;
      System.out.println("path: " +System.getProperty("java.library.path"));
                         //       String problem = e.readFile(servF1);
      servF1 = msHome + "/examples/owl/fof2.owl";
      //servF1 = msHome + "/examples/owl/fault.owl";
      servF1 = msHome + "/examples/owl/cnf2.owl";
      //        servF1 = msHome + "/examples/owl/cnf-exec.owl";
      //      servF1 = msHome + "/examples/owl/fof3.owl";
      //        String servF1 = msHome + "/src/b";
      //        logger.info("result1: " + result);

      String ind = e.readFile(servF1);
      //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SYN/SYN880-1.p");
      ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/PUZ/PUZ001+1.p");
      //      ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/ALG/ALG001-2.p");
      //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRA/GRA001-1.p");
//       ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SET/SET062+3.p");
      //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRP/GRP012-2.p");
      
      System.out.println(ind);
      
      try {
        System.out.println(timeInd);
        String result = null;
//         for (int i=0; i< 700; i++) 
//         {
//             System.out.println(" call " + i);            
//             e = new Otter();
        result = e.proveProblem(ind, timeInd);
        //result = e.proveProblemXml(ind, timeInd);
        //System.out.println(result);
        
        e.writeFile(msHome + "/examples/owl/otter-proof2.owl", result);
        
        //        logger.info("result1: " + result);
        
        //         result = e.proveProblem(ind);
        
        //        logger.info("result2: " + result);
        System.out.println(result);
      }
      catch(Exception ex){
        ex.printStackTrace();
        System.out.println(ex);
      }
  }
  
  
} // end Otter.java


// // Procedure to extract the actual proof of Otter
//       StringTokenizer lines = new StringTokenizer(result, "\n");
//       boolean inProof = false; 
//       String actualProof = "";
//       String line = "";
//       while (lines.hasMoreTokens()){
// 	line = lines.nextToken();
// 	logger.debug("[OTTER] parsing in mode " + inProof + ": " + line );
// 	if (line.equals(proofStart))
// 	  inProof = true;
// 	else if (line.equals(proofEnd))
// 	  { inProof = false;
// 	  break;
// 	  }
// 	else if (inProof) 
// 	  actualProof = actualProof + line + "\n";
//       }
//      logger.debug("[OTTER] actual proof: " + actualProof);
//      logger.debug("[OTTER] result: " + result);

/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */
