// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

package org.mathweb.atp;

import java.io.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout;

import org.mathweb.util.TSTPCreator;
import org.mathweb.core.BasicService;
import org.mathweb.onto.OWLJDom;


/**
@author Juergen Zimmer University of Edinburgh
@version $Date: 2006/07/25 14:54:32 $ $Revision: 1.2.2.1 $
*/
public class Paradox extends ATP {
  
  public String PARADOX_HOME = getSystemProperty("paradox.home", "PARADOX_HOME");
  public String TPTP_HOME = getSystemProperty("tptp.home.dir", "TPTP.HOME.DIR");
  public String[] paradox_envs = new String[] {"PARADOX_HOME="+PARADOX_HOME};
  public String paradoxExecutable = findSystemExecutable(PARADOX_HOME, "paradox", "Paradox.exe");

  private static Logger logger = Logger.getLogger("Paradox");
  private static String logFileName = getSystemProperty("paradox.log.file", "PARADOX_LOG_FILE");
  private static FileAppender appender = newAppender(logFileName);
    
  private static String sep = File.separator;
    
  private static final int defaultNice = 15;
  
  public Paradox()
  {
    super();
    name = "Paradox_1_3";
    acceptedConcept = "TstpProblem";
    resultConcept = "FOBrFPResult";
    proofConcept = "TstpCnfRefutation";
    calculusConcept = "paradox_calc";
    
    logger.setLevel(Level.ALL);
    //    logger.setLevel(Level.OFF);
    if (!logger.getAllAppenders().hasMoreElements()) logger.addAppender(appender);
  }
  
  /** 
   * @param problemFile a file containing a TSTP proving problem.
   * @param language the language the proving problem is in. Must be TPTP.
   * @param format the TSTP format, either "CNF" or "FOF".
   * @param time the time limit for the theorem prover in seconds.
   * @param options additional command line options for the prover command if needed.
   * @return A String array of size 2 where the first entry is the status determined by the ATP system. 
   *  The second entry can contain a proof object in TSTP or XTSTP format.
   */
  public String[] callProver(File problemFile, String language, String format, int time, String options) throws Exception
  {
    File inputFile = null;
    File outputFile = null;
    try{
      outputFile = File.createTempFile("paradox", ".out");
      
      //      writeFile(origFile, problem);
      
      
      if (language.equals(OWLJDom.MW_NS_STR+"TSTP")) {
        
        /*  first transform the problem into TPTP CNF form without include statements */
        String toSyntax= "oldtptp";
        String transform = "none";
        String newProb;

        inputFile = tstp2X.tstp2X(problemFile, toSyntax, transform);
        newProb = readFile(inputFile);
        
        //        logger.info("new problem: " + newProb);
        
        /*  write the new problem description in the input file */
        //        writeFile(inputFile,newProb);
        
        /*  create the shell command to call the prover and execute it */
         String paradoxOptions = options + " --print Model";
 
        String command = paradoxExecutable + " " +  paradoxOptions + " " +
          inputFile.getAbsolutePath() + " > " + outputFile.getAbsolutePath() + " 2>&1";
                  
        //        logger.info("[Paradox] command: " + command);
        String[] commands = {"sh", "-c", command};
        //        logger.info("[Paradox] starting prover...");
        this.start(command, paradox_envs, defaultNice, time);
        
        /*  read the output written by the prover call */
        //        logger.info("[Paradox] reading result..." + outputFile.getAbsolutePath());
        if (outputFile.exists() && (outputFile.length() > ATP.maxOutputLength)){
          System.out.println("length: " + outputFile.length());
          return  new String[] {"Unknown", null, null};
        }
        String output = this.readFile(outputFile);
        
        //        logger.info("[Paradox] output: " + output);
        
        /*  split the output in lines and examine them to extract the status */
        String[] split = output.split("\n");
        StringBuffer buf = new StringBuffer();
        String status = null;
        
        for(int i=0; i<split.length; i++){
          //          logger.info("[line] " + split[i]);
          if ((split[i].length() > 10) && (split[i].indexOf("SATISFIABLE") > -1)) {
            status = "Satisfiable";
            i = split.length;
          } else            
            if ((split[i].length() > 12) && (split[i].indexOf("CONTRADICTION") > -1)){
              //              System.out.println("found status ");
              status = "Unsatisfiable";
              i = split.length;
            }
        }
        /*  create and return the resulting String array */
        //        System.out.println("output " + resArray);
        return new String[] {status, null, output};
      } else{
        throw new Exception("Input language "+language+" not known or not supported.");
      }
    } catch(Exception e){
      //      logger.debug("[Paradox] Exception " + e);
      throw e;
    } finally{
      try{ 
        if (inputFile != null) inputFile.delete();
        if (outputFile != null) outputFile.delete();
      } catch(Exception e){ logger.debug("[Paradox] Exception " + e); }
    }
    
  }
  
  
  public static void main(String argv[]) throws Exception {
    String msHome = System.getProperty("mathserv.home");
    Paradox paradox = new Paradox();
    OWLJDom owlJDom = OWLJDom.getInstance();
    TSTPCreator creator = new TSTPCreator();
    //      String timeInd = owlJDom.createTimeIndi(600,600);
    String timeInd = owlJDom.createTimeIndi(20,20);
    //      timeInd = paradox.readFile(msHome + "/examples/owl/fault-time.owl");
    String servF1;
    System.out.println("path: " +System.getProperty("java.library.path"));
    
    String ind;
    //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/BOO/BOO019-1.p");
    //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/MGT/MGT052-1.p");
    //  ind = creator.tstp2OWLProblem("/home/jzimmer/casc/Active/Problems/EPS/NoTransform/tptp/SYN/SYN816-1+short+ran.p");  
    //        ind = creator.tstp2OWLProblem("/home/jzimmer/casc/Active/Problems/EPS/NoTransform/tptp/SYN/SYN831-1+short+ran.p");
    //      ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/ALG/ALG001-2.p");
    //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRA/GRA001-1.p");
    ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRP/GRP394+3.p");
    //       ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SET/SET062+3.p");
    //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRP/GRP012-2.p");
    //    ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SYN/SYN890-1.p");
    //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SYN/SYN455+1.p");
    //        ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/ALG/ALG111+1.p");
    //ind = creator.tstp2OWLProblem("/home/jzimmer/bath.p");
    String result = paradox.proveProblem(ind, timeInd);
    
    System.out.println("result2: " + result);
  }
  
  
} // end Paradox.java


/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */
