// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

package org.mathweb.atp;

import java.io.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.mathweb.core.BasicService;
import org.mathweb.util.TSTPCreator;
import org.mathweb.onto.OWLJDom;


/**
@author Juergen Zimmer University of Edinburgh
@version $Date: 2006/08/09 12:51:51 $ $Revision: 1.47.2.5 $
*/
public class Spass extends ATP {
  
  public String SPASS_HOME = getSystemProperty("spass.home", "SPASS_HOME");
  public String[] spass_envs = new String[] {"SPASS_HOME="+SPASS_HOME};
  public String spassExecutable = findSystemExecutable(SPASS_HOME, "spass", "Spass.exe");
  private static String axiomsUsed = "Formulae used in the proof :";
  
  private static String sep = File.separator;
  private static String spassOptions = " -PGiven=0 -PProblem=0 -DocProof  -TimeLimit=";
    
  private static final int defaultNice = 15;

  private static Logger logger = Logger.getLogger("Spass");
  private static String logFileName = getSystemProperty("spass.log.file", "SPASS_LOG_FILE"); 
  private static FileAppender appender = newAppender(logFileName);
  
  public Spass()
  {
    super();
    name = "SPASS_2_2";
    acceptedConcept = "TstpProblem";
    resultConcept = "FoAtpResult";
    proofConcept = "TstpCnfRefutation";
    calculusConcept = "spass_res_calc";
    
    if (!logger.getAllAppenders().hasMoreElements()) logger.addAppender(appender);
    logger.setLevel(Level.OFF);
  }
  
  /** 
   * @param problemFile a file containing a TSTP proving problem.
   * @param language the language the proving problem is in. Must be TPTP.
   * @param format the TSTP format, either "CNF" or "FOF".
   * @param time the time limit for the theorem prover in seconds.
   * @param options additional command line options for the prover command if needed.
   * @return A String array of size 2 where the first entry is the status determined by the ATP system. 
   *  The second entry can contain a proof object in TSTP or XTSTP format.
   */
  public String[] callProver(File problemFile, String language, String format, int time, String options) throws Exception
  {
    File inputFile = null;
    File outputFile = null;    
    try{
      outputFile = File.createTempFile("spass", ".out");
      
      /*      -PGiven=0 -PProblem=0 -DocProof -TimeLimit=%d %s */
      //      System.out.println(inputFile + " " + outputFile);
      
      if (language.equals(OWLJDom.MW_NS_STR+"TSTP")) {
        inputFile = tstp2X.tstp2X(problemFile,  
                                  "dfg",
                                  "rm_equality:rstfp");
        if (inputFile == null){
          File testFile = File.createTempFile("problem", ".p");
          String problem = readFile(problemFile);
          writeFile(testFile, problem);
          throw new Exception("tptp2X could not create the DFG problem file for "+testFile.getAbsolutePath() + ".");
        }
        
        String dfg = readFile(inputFile);
        //        System.out.println(dfg);
//        System.out.println("path " + inputFile.getPath());
        
        String command = spassExecutable + " " + options + " " +  spassOptions  + time + " " +
          inputFile.getAbsolutePath() + " > " + outputFile.getAbsolutePath() + " 2>&1";
        logger.debug("[SPASS] command: " + command);
        String[] commands = {"sh", "-c", command};
        //        logger.debug("[SPASS] starting prover...");
        this.start(command, spass_envs, defaultNice, time);
        //      this.exec(commands, spass_envs, false, true);
        logger.debug("reading result..." + outputFile.getAbsolutePath());
        if (outputFile.exists() && (outputFile.length() > ATP.maxOutputLength)){
          System.out.println("length: " + outputFile.length());
          return  new String[] {"Unknown", null, null, null};
        }
        
        String output = this.readFile(outputFile);
        //        System.out.println(output);
        
        logger.debug("[SPASS] transforming to TSTP...");
//         //        String tstp = x2tstp.x2tstp(buf.toString(), "SPASS---2.1");
//         String tstp = x2tstp.x2tstp(output, "SPASS---2.1");
//         logger.info("[SPASS] TSTP..." + tstp);
        
        String[] split = output.split("\n");
        String status = null;
        
        for(int i=0; i<split.length; i++){
          //logger.info("[line] " + split[i]);
          if (split[i].startsWith("SPASS beiseite: Ran out of time."))
            status = "Timeout";
          else 
            if (split[i].startsWith("SPASS beiseite: Proof found."))
              status = "Unsatisfiable";
            else 
              if (split[i].startsWith("SPASS beiseite: Completion found."))
                status = "Satisfiable";
          
        }
        
//        String[] split = output.split("\n");
//         String axiomsStr = null;
//         for(int i=0; i<split.length; i++){
//           if ((split[i].startsWith(axiomsUsed)) && (split[i].length() > 28)){
//             axiomsStr = split[i].substring(28, split[i].length());
//             i = split.length;
//           }
//         }
//         System.out.println("axioms: " +axiomsStr);
        
        //         File tstpFile = File.createTempFile("spass", ".tstp");      
//         writeFile(tstpFile, tstp);
        //String status = extractStatusFromTSTP(tstp);
        //        return new String[] {status, tstp+"\n", output, axiomsStr};
        return new String[] {status, null, output, null};
      } 
      else{
        throw new Exception("Input format "+language+" not known or not supported.");
      }
    } catch(Exception e){
      logger.debug("[Spass] Exception " + e);
      throw e;
    } finally{
      try{ 
        if (inputFile != null) inputFile.delete();
        if (outputFile != null) outputFile.delete();
      } catch(Exception e){ logger.debug("[Spass] Exception " + e); }
    }
  }
  
  
  public static void main(String argv[]) throws Exception {
    String msHome = System.getProperty("mathserv.home");
      Spass e = new Spass();
      OWLJDom owlJDom = OWLJDom.getInstance();
      TSTPCreator creator = new TSTPCreator();
      String timeInd = owlJDom.createTimeIndi(10,10);
      {
        String servF1;
        String tptpProb;
        String result;
        String ind; 
        //       String problem = e.readFile(servF1);
        //servF1 = msHome + "/examples/owl/fof.owl";
        servF1 = msHome + "/examples/owl/fof3.owl";
        //        servF1 = msHome + "/examples/owl/cnf-exec.owl";
        // servF1 = msHome + "/src/b";
//        logger.info("result1: " + result);

//        String ind = e.readFile(servF1);
        //        System.out.println("problem: " + ind);
        
        //       String language = "tptp";
        //       TimeResource tr = new TimeResource();
        //       tr.setCpuTime(300);
        //       tr.setWallClockTime(300);
        //      logger.info("prob: " + ind);
        //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/PUZ/PUZ001+1.p");
        //ind = creator.tstp2OWLProblem("/home/jzimmer/serge.tptp");
        //        ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/NLP/NLP043+1.p");
                ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/TOP/TOP020+1.p");
        
        //ind = creator.tstp2OWLProblem("/home/jzimmer/test.p");        
        //ind = creator.tstp2OWLProblem("/home/jzimmer/t1.tptp");        
        //        ind = creator.tstp2OWLProblem(msHome+ "/foo.tstp");
        //System.out.println(ind);
        
        //String result = e.callProver(problem, language, tr, "");
        result = e.proveProblem(ind, timeInd);
        //String result = e.proveProblemWithOptions(ind, timeInd, "-RFSub");
        //        System.out.println(timeInd);
        //        tptpProb = readFile(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRP/GRP012-2.p");
//         tptpProb = readFile("/home/jzimmer/test.p");
//         tptpProb = readFile("/home/jzimmer/t.p");
        tptpProb = readFile("/home/jzimmer/core1.p");
        //result = e.proveTPTPProblem(tptpProb, new Integer(60));
        //result = e.proveTPTPProblem(tptpProb, new Integer(100));
        
        //String result = e.proveProblemXml(ind, timeInd);
        //          e.writeFile(msHome + "/examples/owl/spass-proof2.owl", result);
        //        e.writeFile(msHome + "/examples/owl/proof3.owl", result);
        
          //        logger.info("result1: " + result);
        
          //         result = e.proveProblem(ind);
        
        System.out.println("result2: " + result);
        
      
      }
  }
  
  

} // end Spass.java

//         String[] split = result.split("\n");
//         StringBuffer buf = new StringBuffer();
//         boolean started = false;
//         for(int i=0; i<split.length; i++){
//           logger.info("[line] " + split[i]);
//           if (!started) {
//             started = split[i].startsWith(startProof);
//             if (started) {
//               buf.append(split[i]);
//               buf.append("\n");
//             }
//           } else {
//             buf.append(split[i]);
//             buf.append("\n");
//             if (split[i].startsWith(endProof)) 
//               i = split.length;
//           }
//         }
//         logger.debug("[SPASS] result: "+ buf.toString());


/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */
