// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

package org.mathweb.atp;

import java.io.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout; 

import org.mathweb.core.BasicService;
import org.mathweb.util.TSTPCreator;
import org.mathweb.onto.OWLJDom;


/**
@author Juergen Zimmer Universitaet des Saarlandes
@version $Date: 2006/07/28 08:35:19 $ $Revision: 1.27.2.5 $
*/
public class Vampire extends ATP {
  
  public String VAMPIRE_HOME = getSystemProperty("vampire.home", "VAMPIRE_HOME");
  public String[] vampire_envs = new String[] {"VAMPIRE_HOME="+VAMPIRE_HOME};
  public String vampireExecutable = findSystemExecutable(VAMPIRE_HOME, "vampire", "Vampire.exe");
  
  private static String sep = File.separator;
    
  private static final int defaultNice = 15;

  private static Logger logger = Logger.getLogger("Vampire");
  private static String logFileName = getSystemProperty("vampire.log.file", "VAMPIRE_LOG_FILE"); 
  private static FileAppender appender = newAppender(logFileName);
  
  public Vampire()
  {
    super();
    name = "Vampire_8_0";
    acceptedConcept = "TstpProblem";
    resultConcept = "FoAtpResult";
    proofConcept = "TstpCnfRefutation";
    calculusConcept = "vampire_res_calc";
    
    logger.setLevel(Level.ALL);
    logger.setLevel(Level.OFF);
    if (!logger.getAllAppenders().hasMoreElements()) logger.addAppender(appender);
  }
  
  /** 
   * @param problemFile a file containing a TSTP proving problem.
   * @param language the language the proving problem is in. Must be TPTP.
   * @param format the TSTP format, either "CNF" or "FOF".
   * @param time the time limit for the theorem prover in seconds.
   * @param options additional command line options for the prover command if needed.
   * @return A String array of size 2 where the first entry is the status determined by the ATP system. 
   *  The second entry can contain a proof object in TSTP or XTSTP format.
   */
  public String[] callProver(File problemFile, String language, String format, int time, String options) throws Exception
  {
    File inputFile = null;
    File outputFile = null;

    try{
      outputFile = File.createTempFile("vampire", ".out");

      if (language.equals(OWLJDom.MW_NS_STR+"TSTP")) {
        inputFile = tstp2X.tstp2X(problemFile, 
                                  "oldtptp",
                                  "none");
        //System.out.println("path " + inputFile.getPath());
        
        /*         vampire -t %d %s*/
        String newProb = readFile(inputFile);
        //        System.out.println("new: " + newProb);
        writeFile("/home/jzimmer/problem", newProb);
        
        String command = vampireExecutable + " " + options + " -t " + time + " " +
          inputFile.getAbsolutePath() + " > " + outputFile.getAbsolutePath() + " 2>&1";
        //        logger.debug("[VAMPIRE] command: " + command);
        String[] commands = {"sh", "-c", command};
        //        logger.debug("[VAMPIRE] starting prover...");
        this.start(command, vampire_envs, defaultNice, time);
        //      this.exec(commands, vampire_envs, false, true);
        //        logger.debug("reading result..." + outputFile.getAbsolutePath());

        
        //        logger.debug("[VAMPIRE] output..." + output);
//         logger.debug("[VAMPIRE] transforming to TPTP...");
//         String tstp = x2tstp.x2tstp(output, "Vampire---8.0");
//         logger.info("[VAMPIRE] TPTP..." + tstp);
        
        //         File tstpFile = File.createTempFile("vampire", ".tstp");      
        //         writeFile(tstpFile, tstp);
        //        String status = extractStatusFromTSTP(tstp);
        String output = "";
        String status = null;

        if (outputFile.exists() && (outputFile.length() > ATP.maxOutputLength)){
          FileInputStream isFile  = null;
          StringBuffer sbRead = new StringBuffer();
          try {
            isFile = new FileInputStream(outputFile);
            int nChar;
            nChar = isFile.read();
            while (nChar != '\n') {
              sbRead.append((char) nChar);
              nChar = isFile.read();
            }
            isFile.close();
          } catch (IOException ioe) {
            return null;
          } finally {
            if (isFile != null)
              try{ isFile.close(); } catch (Exception e) {}
          }
          output = sbRead.toString();
        } else {
          output = this.readFile(outputFile);
        }
        
        //        System.out.println(output);
        
        if (output.startsWith("Refutation found"))
          status = "Unsatisfiable";
        else 
          if (output.startsWith("Satisfiability detected"))
            status = "Satisfiable";
          else 
            status = "Unknown";
        
        //        return new String[] {status, tstp+"\n", output};
        return new String[] {status, null, null};
      } 
      else{
        throw new Exception("Input format "+language+" not known or not supported.");
      }
    } catch(Exception e){
      logger.debug("[Vampire] Exception " + e);
      throw e;
    } finally{
      try{ 
        if (inputFile != null) inputFile.delete();
        if (outputFile != null) outputFile.delete();
      } catch(Exception e){ logger.debug("[Vampire] Exception " + e); }
    }
  }
  
  
  public static void main(String argv[]) throws Exception {
    String msHome = System.getProperty("mathserv.home");
      Vampire e = new Vampire();
      OWLJDom owlJDom = OWLJDom.getInstance();
      TSTPCreator creator = new TSTPCreator();
      String timeInd = owlJDom.createTimeIndi(10,10);
      {
        String servF1;
        String result;
        
        //       String problem = e.readFile(servF1);
        //servF1 = msHome + "/examples/owl/fof.owl";
        servF1 = msHome + "/examples/owl/fof3.owl";
        //        servF1 = msHome + "/examples/owl/cnf-exec.owl";
        // servF1 = msHome + "/src/b";
//        logger.info("result1: " + result);

        String ind = e.readFile(servF1);
        //        System.out.println("problem: " + ind);
        
        //       String language = "tptp";
        //       TimeResource tr = new TimeResource();
        //       tr.setCpuTime(300);
        //       tr.setWallClockTime(300);
        //      logger.info("prob: " + ind);
        //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRP/GRP012-2.p");
        //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/PUZ/PUZ001+1.p");
        //        ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SYN/SYN890-1.p");
        //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRP/GRP123-1.005.p");
        ind = creator.tstp2OWLProblem("/home/jzimmer/nb/casc-test/Active/Problems/EPT/NoTransform/tptp/SYN/SYN881-1+short+ran.p");
        
// //         //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SYN/SYN455+1.p");
// //         //        ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/ALG/ALG111+1.p");
// //         //ind = creator.tstp2OWLProblem("/home/jzimmer/bath.p");
        //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/SYN/SYN538-1.p");
        
        result = e.proveProblem(ind, timeInd);
        
        String omdocProb;
        //        tptpProb = readFile(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/GRP/GRP012-2.p");
        //omdocProb = readFile("/home/jzimmer/serge.omdoc");
        //        String tptpProb = readFile("/home/jzimmer/DolceSimpl_PreDolce.tptp");
        //        result = e.proveOmdocProblem(omdocProb, new Integer(20));
        
        System.out.println("result2: " + result);
      }
  }
  

} // end Vampire.java

//         String[] split = result.split("\n");
//         StringBuffer buf = new StringBuffer();
//         boolean started = false;
//         for(int i=0; i<split.length; i++){
//           logger.info("[line] " + split[i]);
//           if (!started) {
//             started = split[i].startsWith(startProof);
//             if (started) {
//               buf.append(split[i]);
//               buf.append("\n");
//             }
//           } else {
//             buf.append(split[i]);
//             buf.append("\n");
//             if (split[i].startsWith(endProof)) 
//               i = split.length;
//           }
//         }
//         logger.debug("[VAMPIRE] result: "+ buf.toString());


/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */
