// -*- indent-tabs-mode: nil -*-
/*****************************************************************
This file is part of the MathServe system

Authors: Juergen Zimmer

email: jzimmer@ags.uni-sb.de
   
For information about this program, write to:                          
MathWeb-Project                                                        
AG Siekmann/FB Informatik                                            
Universitaet des Saarlandes                                          
Gebaeude E 1.1                                                     
D-66041 Saarbruecken                                                 
Germany    

For information about the newest version of MathServe, see 
http://www.ags.uni-sb.de/~jzimmer/mathserve.html

See the file "LICENSE.txt" or
http://www.ags.uni-sb.de/~jzimmer/LICENSE.html
for information on usage and redistribution
of this file, and for a DISCLAIMER OF ALL WARRANTIES.
*****************************************************************/

package org.mathweb.atp;

import java.io.*;
import org.apache.log4j.Level;
import org.apache.log4j.Logger; 
import org.apache.log4j.FileAppender; 
import org.apache.log4j.SimpleLayout;

import org.mathweb.util.TSTPCreator;
import org.mathweb.core.BasicService;
import org.mathweb.onto.OWLJDom;


/**
@author Juergen Zimmer Universitaet des Saarlandes
@version $Date: 2006/07/28 08:35:19 $ $Revision: 1.8.2.4 $
*/
public class Waldmeister extends ATP {
  
  public String WALDMEISTER_HOME = getSystemProperty("waldmeister.home", "WALDMEISTER_HOME");
  public String TPTP_HOME = getSystemProperty("tptp.home.dir", "TPTP.HOME.DIR");
  public String[] waldmeister_envs = new String[] {"WALDMEISTER_HOME="+WALDMEISTER_HOME};
  public String waldmeisterExecutable = findSystemExecutable(WALDMEISTER_HOME, "WaldmeisterII.comp", " WaldmeisterII.comp");
  public String limitedRunExecutable = findSystemExecutable(TPTP_HOME+"/SystemExecution", "TreeLimitedRun", "TreeLimitedRun.exe");

  private static Logger logger = Logger.getLogger("Waldmeister");
  private static String logFileName = getSystemProperty("waldmeister.log.file", "WALDMEISTER_LOG_FILE");
  private static FileAppender appender = newAppender(logFileName);
    
  private static String sep = File.separator;
  private static String proofStart = "START OF PROOF";
  private static String proofEnd = "END OF PROOF";
    
  private static final int defaultNice = 15;
  
  public Waldmeister()
  {
    super();
    name = "Waldmeister_704";
    acceptedConcept = "TstpProblem";
    resultConcept = "FoAtpResult";
    //    proofConcept = "TstpCnfRefutation";
    calculusConcept = "waldmeister_calc";
    
    logger.setLevel(Level.ALL);
    logger.setLevel(Level.OFF);
    if (!logger.getAllAppenders().hasMoreElements()) logger.addAppender(appender);
  }
  
  /** 
   * @param problem a proving problem.
   * @param language the language the proving problem is in. Must be TPTP.
   * @param format the TSTP format, either "CNF" or "FOF".
   * @param time the time limit for the theorem prover in seconds.
   * @param options additional command line options for the prover command if needed.
   * @return A String array of size 2 where the first entry is the status determined by the ATP system. 
   *  The second entry can contain a proof object in TSTP or XTSTP format.
   */
  public String[] callProver(File problemFile, String language, String format, int time, String options) throws Exception
  {
    File inputFile = null;
    File outputFile = null;
    try{
      //      origFile = File.createTempFile("waldmeister", ".orig");
      outputFile = File.createTempFile("waldmeister", ".out");
      
      //      writeFile(origFile, problem);
      
      //      logger.info("new problem: " + newProb);
      
//       if ((problem != null) && (problem.length() > 10000000))
//         return  new String[] {"Unknown", ""};
      if (language.equals(OWLJDom.MW_NS_STR+"TSTP")) {
        
        /*  first transform the problem into TPTP CNF form without include statements */
        String toSyntax= "waldmeister";
        String transform = "none";

        inputFile = tstp2X.tstp2X(problemFile, toSyntax, transform);
        
        /*  write the new problem description in the input file */
        //        writeFile(inputFile,newProb);
        
        /*  create the shell command to call the prover and execute it */
        //        String eOptions = " --print-statistics -xAuto -tAuto --memory-limit=Auto --tstp-in --tstp-out --cpu-limit=" + time + " ";
        String waldmeisterOptions = options;
        //        String eOptions = " -s --print-statistics -xAuto -tAuto --tstp-in --tstp-out -l4 --memory-limit=100 --cpu-limit=" + time + " ";        

        String command = waldmeisterExecutable + " " +  waldmeisterOptions + " " +
          inputFile.getAbsolutePath() + " > " + outputFile.getAbsolutePath() + " 2>&1";
                  
        logger.info("[WALDMEISTER] command: " + command);
        String[] commands = {"sh", "-c", command};
        logger.info("[WALDMEISTER] starting prover...");
        this.start(command, waldmeister_envs, defaultNice, time);
        
        /*  read the output written by the prover call */
        //        logger.info("[WALDMEISTER] reading result..." + outputFile.getAbsolutePath());
        
        //         if ((result != null) && (result.length() > 1000000)){
        if (outputFile.exists() && (outputFile.length() > ATP.maxOutputLength)){
          System.out.println("length: " + outputFile.length());
          return  new String[] {"Unknown", null, null};
        }
       
        //        logger.info("[WALDMEISTER] result: " + result);
        //         logger.info("[WALDMEISTER] transforming to TSTP...");
        //         String tstp = x2tstp.x2tstp(result, "E---0.82");
        //         logger.info("[WALDMEISTER] TSTP..." + tstp);
        //         return tstp;
        
        String output = this.readFile(outputFile);
        //        System.out.println("output: " + output);
        
        /*  split the output in lines and examine them to extract the status */
        String[] split = output.split("\n");
        StringBuffer buf = new StringBuffer();
        String status = "Timeout";
        for(int i=0; i<split.length; i++){
          if (split[i].length() > 10){
            if (split[i].startsWith("Waldmeister states: Goal proved"))
              //              System.out.println("found status ");
              { status = "Unsatisfiable"; i = split.length;
              }
            else if (split[i].startsWith("Waldmeister states: System completed."))
              {  status = "Satisfiable";i = split.length;
              }
            else if (split[i].startsWith("Waldmeister states: Unfair completion run out of critical pairs."))
              {  status = "GaveUp";i = split.length;
              }
            else if (split[i].trim().equals("Terminated: no specification"))
              {   status = "Unknown"; i = split.length;              
              }
          }
        }
        //        logger.info("[status] " + status);
        
        /*  create and return the resulting String array */
        //        return new String[] {status, null, output};
        return new String[] {status, null, output};
      } else{
        throw new Exception("Input language "+language+" not known or not supported.");
      }
    } catch(Exception e){
      logger.debug("[WALDMEISTER] Exception " + e);
      throw e;
    } finally{
      try{ 
        if (inputFile != null) inputFile.delete();
        if (outputFile != null) outputFile.delete();
      } catch(Exception e){ logger.debug("[WALDMEISTER] Exception " + e); }
    }
    
  }
  
  
  
  public static void main(String argv[]) throws Exception {
    String msHome = System.getProperty("mathserv.home");
      Waldmeister wald = new Waldmeister();
      OWLJDom owlJDom = OWLJDom.getInstance();
      TSTPCreator creator = new TSTPCreator();
      String timeInd = owlJDom.createTimeIndi(10,10);
      //      timeInd = wald.readFile(msHome + "/examples/owl/fault-timwald.owl");
      String servF1;
      System.out.println("path: " +System.getProperty("java.library.path"));
                         //       String problem = wald.readFile(servF1);
      servF1 = msHome + "/examples/owl/fof2.owl";
      //servF1 = msHome + "/examples/owl/fault.owl";
      servF1 = msHome + "/examples/owl/cnf2.owl";
      //        servF1 = msHome + "/examples/owl/cnf-exec.owl";
      //      servF1 = msHome + "/examples/owl/fof3.owl";
      //        String servF1 = msHome + "/src/b";
      //        logger.info("result1: " + result);

      String ind = wald.readFile(servF1);
      //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/ALG/ALG005-1.p");
      //      ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/BOO/BOO077-1.p");
      //ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/BOO/BOO027-1.p");
      ind = creator.tstp2OWLProblem(msHome + "/thirdParty/TPTPWorld/TPTP/Problems/COL/COL002-4.p");
      
      //ind = creator.tstp2OWLProblem("/home/jzimmer/t.p");
      
      System.out.println(ind);
      
      try {
        System.out.println(timeInd);
        String result = null;
//         for (int i=0; i< 700; i++) 
//         {
//             System.out.println(" call " + i);            
//             e = new Waldmeister();
        result = wald.proveProblem(ind, timeInd);
        //result = wald.proveProblemXml(ind, timeInd);
        //System.out.println(result);
        
        //        logger.info("result1: " + result);
        
        //         result = wald.proveProblem(ind);
        
        //        logger.info("result2: " + result);
        System.out.println(result);
      }
      catch(Exception ex){
        ex.printStackTrace();
        System.out.println(ex);
      }
  }
  
  
} // end Waldmeister.java


/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */
