package org.mathweb.owls;

import java.io.File;
import java.io.FileInputStream;
import java.io.FileNotFoundException;
import java.io.IOException;
import java.net.URI;
import java.util.Properties;

import org.mindswap.owl.OWLCache;
import org.xml.sax.InputSource;

/**
 * 
 * 
 * 
 * @author Juergen Zimmer
 *
 */
public class OWLCacheImpl implements OWLCache {
  protected Properties cache = new Properties();
  protected String localCacheDirectory = null;
  protected boolean forced = true;
	
  /**
   * Forces the readers to use the cached copies of the files even if the remote
   * file may be available. When the forcing of cache is enabled the reader will
   * first check if the cached copy exists and then only try to use the remote file
   * when there is no cached copy.
   * 
   * @param b
   */
  public void setForced(boolean b) {
    forced = b;
  }
	
  /**
   * Returns if the using of cache is forced.
   * 
   * @return
   */
  public boolean isForced() {
    return forced;
  }
	
  /**
   * Sets the cache dir for the inference engine to find the cached files when a
   * file cannot be downparseed from its original URL. The cache dir should include
   * a file named service.idx. This index file is a text file where each line is in
   * the format
   * [service description url]=[local filename]
   * 
   * The ':' characters in the url's should be escaped as "\:"  
   * 
   * @param dir sets the local cache directory. if null it forces not to use the cache.
   * 			  if the given dir or index file inthat dir does not exist then nothing is
   * 			  done   
   */
  public void setLocalCacheDirectory(String dir) {
    localCacheDirectory = dir;
    if(dir == null) {
      cache.clear();
      System.out.println("INFO: Local cache directory is disabled");
    }
    else {
      String indexFileName = localCacheDirectory + File.separator + "service.idx"; 
      try {
        File indexFile = new File(indexFileName);
	
        cache = new Properties();				
			
        cache.load(new FileInputStream(indexFile));
	
        //				System.out.println("INFO: Cache has been initialized with " + cache.size() + " entries");
      } catch(FileNotFoundException e) {
        System.err.println("ERROR: Cache index file " + indexFileName + " cannot be found");
        localCacheDirectory = null;
      } catch(IOException e) {
        System.err.println("ERROR: Cache index file " + indexFileName + " has an invalid format");
        localCacheDirectory = null;
      }	
    }
  }

	
  /**
   * Returns the cached File object for the given URI. Returns null if there isn't an entry in the
   * cache for the given file or cached file is not found.
   * 
   * @param fileURI
   * @return
   */
  public InputSource getCachedFile(String fileURI) {		
    InputSource in = null;
      
    try {
      if(cache == null) return null;
        
      if(fileURI.endsWith("#")) fileURI = fileURI.substring(0, fileURI.length() - 1);
			
      String localFileName = cache.getProperty(fileURI);
      if(localFileName != null) {
        File file = new File(localFileName);
        if(!file.exists() && localCacheDirectory != null) {
          file = new File(localCacheDirectory + File.separator + localFileName);			    
          if(!file.exists()){
            String msHome = System.getProperty("mathserv.home");
            file = new File(msHome + File.separator + "etc" + File.separator +
                            "ontology" + File.separator + localFileName);			    
          }
          if(!file.exists()) {
            System.out.println("WARNING: Cached file does not exist " + file);
          }
          else {
            //				    System.err.println("INFO: Using cached file " + file + " for URI " + fileURI);
            in = new InputSource(new FileInputStream(file));
          }
        }
      }
    }
    
    catch(Exception e) {
    }
		
    return in;		
  }
	
	
  /**
   * Add a local file to be used as a cached copy for the given URI.
   * 
   * @param fileURI URI for the remote file
   * @param localFile Path for the local cached copy
   */
  public void addCachedFile(String fileURI, File localFile) {	
    cache.setProperty(fileURI, localFile.getName());
  }
	
  public void updateCachedFile(URI uri) {
  }
}

      
/* Settings for Emacs.
 * Local variables:
 * c-basic-offset: 2
 * tab-width: 2
 * indent-tabs-mode: nil
 * End:
 */

