module Check where

import Prelude hiding ( catch )

import Control.Exception            ( catch, throw, evaluate )
import Data.Set as Set              ( Set, empty, member, fromList, insert, delete, union, toList, difference )
import Data.List                    ( isSuffixOf, isPrefixOf, unfoldr, intersperse, nub )
import System.Cmd                   ( system )
import System.Exit                  ( ExitCode(..) )
import System.Environment           ( getArgs )
import System.Directory             ( removeFile )

import Chalmers.QuickCheck         as Q  hiding ( Result, Property, property )
import Chalmers.QuickCheck.Monadic as QM 

import ReadFile

import qualified SatMonad    as M
import           Prop        as P
import           ExtSat      as E
import           SatWrappers 

import Control.Monad

data ShrinkMerge
  = MergeFull
  | MergeNone
    deriving Show

data Property
  = NoCrash
  | Correct
    deriving Show

data Params = Params {
   merge :: ShrinkMerge,
   prop  :: Property
 } deriving Show


shrinkMerge MergeFull = const []
shrinkMerge MergeNone = const []

shrinkProblem ps p = shrink p ++ shrinkMerge (merge ps) p

property ps = case prop ps of
  NoCrash -> prop_simple
  Correct -> prop_full

parseParams :: Params -> [String] -> (Params, [String])
parseParams p [] = (p, [])
parseParams p ("--merge-none"   : args) = parseParams (p { merge = MergeNone }) args
parseParams p ("--merge-full"   : args) = parseParams (p { merge = MergeFull }) args
parseParams p ("--prop-correct" : args) = parseParams (p { prop  = Correct })   args
parseParams p ("--prop-nocrash" : args) = parseParams (p { prop  = NoCrash })   args
parseParams p (a : args)                = let (ps, args') = parseParams p args in (ps, a : args')

defaults = Params { merge = MergeNone, prop = Correct }

main = getArgs >>= \args ->
  case parseParams defaults args of 

     (ps, [sat]) | isSuffixOf ".sat" sat -> do
         commands <- readFile' sat >>= return . map read . lines
         putStr (sat2c commands)
         return True

     -- run quickcheck
     (ps, [solver]) -> Q.quickCheckWith 100 500 100 (Q.forAllShrink Q.arbitrary (shrinkProblem ps) (property ps solver))

     -- run quickcheck
     (ps, ["--repeat", solver]) -> 
            let loop = Q.quickCheckWith 100 500 100 (Q.forAllShrink Q.arbitrary (shrinkProblem ps) (property ps solver)) >>= \r ->
                        if r then loop else return True
                in loop

     -- read a file with satcommands
     (ps, [solver, sat]) | isSuffixOf ".sat" sat -> do
         commands <- readFile' sat >>= return . map read . lines
         Q.quickCheckWith 1 1 1 $ Q.shrinking (shrinkProblem ps) (Arb commands) (property ps solver)
--         Q.quickCheckWith 1 1 1 $ Q.shrinking shrink (Smart 0 (Arb commands)) (\ (Smart _ p) -> prop_simple solver p)

     -- read a dimacs file and convert to satcommands
     (ps, [solver, cnf]) | isSuffixOf ".cnf" cnf -> do
         dimacs <- readFile' cnf
         Q.quickCheckWith 1 1 1 $ Q.shrinking shrink (Smart 0 (Arb $ fromDimacs dimacs)) (\ (Smart _ p) -> property ps solver p)

     (ps, args)  -> do
        putStrLn "No matching action."
        putStrLn ("( params = " ++ show ps ++ " args = " ++ show args ++ " )")
        return True

fromDimacs :: String -> [Command]
fromDimacs = 
   (++ [Simplify True True, Solve []])
 . map (Assert . clause)
 . unfoldr pclause
 . map read
 . words
 . unlines 
 . filter (not . isPrefixOf "c")
 . filter (not . isPrefixOf "p cnf")
 . lines 

 where pclause :: [Int] -> Maybe ([Int],[Int])
       pclause [] = Nothing
       pclause ts = Just . (\ (x,0:y) -> (x,y)) $ break (==0) ts

----------------------------------------------------------------------------------------------------
-- arbitrary implementation

newtype Arb = Arb [Command] deriving Show

instance Arbitrary Command where
    shrink (DefProp v f)      = [ Assert f ]
                             ++ [ DefProp v f'
                                | f' <- shrink f
                                ]
    shrink (DefAdd c s x y z) = [ DefProp s (App Xor [x, App Xor [y,z]])
                                , DefProp c (App And [x, y])
                                , DefProp c (App And [x, z])
                                , DefProp c (App And [y, z])
                                ]
                             ++ [ DefAdd c s x' y' z' | (x',y',z') <- shrink (x,y,z) ]
    shrink (Assert f)         = map Assert (shrink f)
    shrink (Freeze v)         = []
    shrink (Simplify x y)     = [ Simplify x' y' | (x',y') <- shrink (x,y) ]
    shrink (Solve fs)         = [ Solve fs'
                                | fs' <- shrink fs
                                , let isLit (App Not [Var _]) = True
                                      isLit (Var _)           = True
                                      isLit _                 = False
                                , all isLit fs
                                ]

    arbitrary                 = fmap (Assert . clause) $ sized $ \s -> aClause 3 (s `div` 8)

instance Arbitrary Formula where
    shrink (App Not [f])     = [f] ++ [ App Not [f'] | f' <- shrink f ]
    shrink (App op  [f,g])   = [f,g]
                            ++ [ App Not [f], App Not [g] ]
                            ++ [ App op [f',g']    | (f',g')     <- shrink (f,g) ]
    shrink (App op  [f,g,h]) = [f,g,h]
                            ++ [ App And [f,g], App And [App Not [f], h] ]
                            ++ [ App op [f',g',h'] | (f',g', h') <- shrink (f,g,h) ]
    shrink _                 = []
    arbitrary         = sized $ \s -> fgen s s

fgen vars 0 = fmap var (aLit vars)
fgen vars n = frequency [ (2, do x <- fgen vars (n `div` 2)
                                 y <- fgen vars (n `div` 2)
                                 return (mkAnd x y) )
                        , (3, do x <- fgen vars (n `div` 2)
                                 y <- fgen vars (n `div` 2)
                                 return (mkOr x y) )
                        , (3, do x <- fgen vars (n `div` 2)
                                 y <- fgen vars (n `div` 2)
                                 return (mkImp x y) )
                        , (2, do x <- fgen vars (n `div` 2)
                                 y <- fgen vars (n `div` 2)
                                 return (mkXor x y) )
                        , (1, do x <- fgen vars (n - 1)
                                 return (mkNot x) )
                        , (3, do x <- fgen vars (n `div` 3)
                                 y <- fgen vars (n `div` 3)
                                 z <- fgen vars (n `div` 3)
                                 return (mkIte x y z) )
                        ]
             

instance Arbitrary Arb where
   shrink (Arb cs) =
     -- apply basic formula level simplifications (everywhere)
     [ Arb cs'
     | let cs' = map simpC cs
     , cs /= cs'
     ] ++

     -- try to delete operations completely
     [ Arb [ c | Fixed c <- cs' ]
     | cs' <- shrink (map Fixed cs)
     ] ++

     -- replace a variables with true/false or some other variable
     [ Arb cs'
     | v <- varsCs cs
     , e <- [ App F [], App T [] ]
         ++ [ Var w | w <- varsCs cs, v /= w ]
         ++ [ App Not [Var w] | w <- varsCs cs, v /= w ]
     , let cs' = substCs v e cs
           
           sound defd (Assert f    : cs) = sound (nub (varsF f ++ defd)) cs
           sound defd (Solve fs    : cs) = sound (nub (concat (map varsF fs) ++ defd)) cs
           sound defd (DefProp x f : cs) = let defd' = nub (varsF f ++ defd) in (x `notElem` defd) && sound (x : defd') cs
           sound defd (DefAdd x y f g h : cs) = let defd' = nub (varsF f ++ varsF g ++ varsF h ++ defd)
                                                 in (x `notElem` defd) && (y `notElem` defd) && sound (x : y : defd') cs
           sound defd _                  = True
     , sound [] cs'
     ] ++

     -- apply basic formula level simplifications (one at a time)
     [ Arb cs'
     | k <- [0..length cs-1]
     , let c = cs !! k
     , c' <- [ c' | let c' = simpC c, c /= c' ] ++ shrink c
     , let cs' = take k cs ++ [c'] ++ drop (k+1) cs
     ]
    where
     varsCs cs = nub (concat (map varsC cs))

     varsC (Assert f)         = varsF f
     varsC (Solve fs)         = varsFs fs
     varsC (DefProp x f)      = nub (x : varsF f)
     varsC (DefAdd x y f g h) = nub (x : y : varsF f ++ varsF g ++ varsF h)
     varsC _                  = []
     
     varsFs fs = nub (concat (map varsF fs))

     varsF (Var x)    = [x]
     varsF (App _ fs) = varsFs fs
     
     substCs v e cs = [ c' | c <- cs, Just c' <- [substC v e c] ]
     
     substC v e (Assert f)         = Just (Assert (substF v e f))
     substC v e (Solve fs)         = Just (Solve [ f | f <- fs, v `elem` varsF f ])
     substC v e (DefProp x f)      | v /= x           = Just (DefProp x (substF v e f))
     substC v e (DefAdd x y f g h) | v /= x && v /= y = Just (DefAdd x y (substF v e f) (substF v e g) (substF v e h))
     substC v e c                  = Nothing
     
     substF v e (Var x)     = if x == v then e else Var x
     substF v e (App op fs) = App op (map (substF v e) fs)
     
     simpC (Assert f)         = Assert (simpF f)
     simpC (Solve fs)         = Solve (map simpF fs)
     simpC (DefProp x f)      = DefProp x (simpF f)
     simpC (DefAdd x y f g h) = DefAdd x y (simpF f) (simpF g) (simpF h)
     simpC c                  = c
     
     simpF (App Not [App Not [f]]) = simpF f
     simpF (App Not [App T []]) = App F []
     simpF (App Not [App F []]) = App T []
     
     simpF (App And [App T [], g]) = simpF g
     simpF (App And [App F [], g]) = App F []
     simpF (App And [f, App T []]) = simpF f
     simpF (App And [f, App F []]) = App F []
     simpF (App And [f, g])        | f == g = simpF f
     
     simpF (App Or [App T [], g]) = App T []
     simpF (App Or [App F [], g]) = simpF g
     simpF (App Or [f, App T []]) = App T []
     simpF (App Or [f, App F []]) = simpF f
     simpF (App Or [f, g])        | f == g = simpF f
     
     simpF (App Imp [App T [], g]) = simpF g
     simpF (App Imp [App F [], g]) = App T []
     simpF (App Imp [f, App T []]) = App T []
     simpF (App Imp [f, App F []]) = App Not [simpF f]
     simpF (App Imp [f, g])        | f == g = App T []
     
     simpF (App Equ [App T [], g]) = simpF g
     simpF (App Equ [App F [], g]) = App Not [simpF g]
     simpF (App Equ [f, App T []]) = simpF f
     simpF (App Equ [f, App F []]) = App Not [simpF f]
     simpF (App Equ [f, g])        | f == g = App T []
     
     simpF (App Xor [App T [], g]) = App Not [simpF g]
     simpF (App Xor [App F [], g]) = simpF g
     simpF (App Xor [f, App T []]) = App Not [simpF f]
     simpF (App Xor [f, App F []]) = simpF f
     simpF (App Xor [f, g])        | f == g = App F []
     
     simpF (App Ite [App T [], g, h]) = simpF g
     simpF (App Ite [App F [], g, h]) = simpF h
     simpF (App Ite [f, App T [], h]) = App Or [simpF f, simpF h]
     simpF (App Ite [f, App F [], h]) = App And [App Not [simpF f], simpF h]
     simpF (App Ite [f, g, App T []]) = App Or [App Not [simpF f], simpF g]
     simpF (App Ite [f, g, App F []]) = App And [simpF f, simpF g]
     simpF (App Ite [f, g, h])        | g == h = simpF g
     simpF (App Ite [f, g, h])        | f == g = App Or [simpF f, simpF h]
     
     simpF (App op fs) = App op (map simpF fs)
     simpF f           = f
     
   arbitrary = fmap (Arb . (++[Solve []])) $ sized $ \s -> 
      arbProgram (1 `max` (s `div` 2)) (s `div` 2 + 1)
{-
      do cs <- fmap (map clause . filter okClause) $ vectorOf s (aClause 3 (s `div` 8))
         fs <- fmap (filter (not . isSimple)) $ vectorOf (s `div` 4) (fgen (s `div` 8) (s `div` 20))
         return (map Assert (cs ++ fs))
-}

    where
     arbProgram vars 0 =
       do return []
     
     arbProgram vars s =
       do c <- frequency
               [ (10, liftM2 DefProp (return (var (vars+1))) arbForm)
               , (5, liftM5  DefAdd (return (var (vars+1))) (return (var (vars+2))) arbForm arbForm arbForm)
               , (30, liftM  Assert (arbWeakForm True))
               , (4,  liftM2 Simplify arbitrary arbitrary)
               , (4,  liftM  Solve (take 8 `fmap` listOf arbLit))
               ]
          cs <- arbProgram (case c of
                              DefAdd _ _ _ _ _ -> vars+2
                              _                -> vars+1) (s-1)
          return (c:cs)
      where
       arbForm = sized arbF
        where
         arbF n = frequency $
           [ (1,  do liftM Var arbVar)
           , (1,  do liftM (App Not . (:[]) . Var) arbVar)
           ] ++
           concat
           [ [ (10, do op <- arbOp
                       xs <- vectorOf (arity op) (arbF' (arity op))
                       return (App op xs))
             ]
           | n > 0
           ]
          where
           arbF' 1 = arbF (n-1)
           arbF' k = arbF (n `div` k)
           
       arbWeakForm pos = sized (arbF pos)
        where
         arbF pos n = frequency $
           [ (1,  do liftM Var arbVar)
           , (1,  do liftM (App Not . (:[]) . Var) arbVar)
           ] ++
           concat
           [ [ (4, do xs <- vectorOf 2 (arbF' pos 2)
                      return (App (if pos then Or else And) xs))
             , (2, do xs <- vectorOf 3 (arbF' pos 3)
                      return (App Ite xs))
             , (1, do op <- arbOp
                      xs <- vectorOf (arity op) (arbF' (neg op /= pos) (arity op))
                      return (App op xs))
             ]
           | n > 0
           ]
          where
           arbF' pos 1 = arbF pos (n-1)
           arbF' pos k = arbF pos (n `div` k)
           
       arbOp =
         frequency
           [ (1 + arity op, return op)
           | op <- [T, F, Not, And, Or, Imp, Equ, Xor, Ite]
           ]

       neg op =
         op `elem` [ Not, Xor, Equ ]

       arity op
         | op `elem` [T,F] = 0
         | op `elem` [Not] = 1
         | op `elem` [Ite] = 3
         | otherwise       = 2
       
       var i =
         "v" ++ show i
       
       arbVar =
         do i <- choose (1,vars)
            return (var i)

       arbLit =
         do x <- arbVar
            elements [ App Not [ Var x ], Var x ]

okClause [] = False
okClause ls = not (isTrivial ls)
        
isTrivial [] = False
isTrivial (x : xs) | elem x    xs = True
                   | elem (-x) xs = True
                   | otherwise    = isTrivial xs

isSimple (App _   [f,g])   = equals f g
isSimple (App Ite [c,t,f]) = equals c t || equals c f || equals t f
isSimple _                 = True

equals f g = f == g || f == mkNot g || mkNot f == g

clause xs         = foldr1 (\x y -> App Or [x,y]) (map var xs)

var x | x > 0     = Var ('x' : show x)
      | otherwise = App Not [Var ('x' : show (-x))]


aLit :: Int -> Gen Int
aLit n = do i <- elements [ 1 .. max n 1 ]
            b <- arbitrary
            return (if b then i else -i)

aClause :: Int -> Int -> Gen [Int]
aClause k n = vectorOf k (aLit n)


----------------------------------------------------------------------------------------------------
-- properties

prop_simple solver (Arb cs) = monadicIO $
  do monitor $ whenFail' $ writeFile' "bug.sat" (unlines (map show cs))
     rs <- QM.run (evalInterp solver cs)
     assert (modelCheck cs rs)

prop_full solver (Arb cs) = monadicIO $
  do monitor $ whenFail' $ writeFile' "bug.sat" (unlines (map show cs))
     rs <- QM.run (evalInterp solver cs)
     monitor $ whenFail $ print ("model",cs,rs)
     assert (modelCheck cs rs)

     let first5 = take 5
          [ taBort r
            | let taBort (Model _) = Model []
                  taBort (Contr _) = Contr []
          , r <- rs
          , r /= Ok
          ]
     monitor $ collect $ length [ m | Model m <- rs ]

     let ws = weaken cs rs
     wrs <- QM.run (evalExt limmat ws)
     monitor $ whenFail $ print ("contr",ws,wrs)
     assert (contrCheck ws wrs)

----------------------------------------------------------------------------------------------------
-- evaluation of a sequence of commands

evalCommandFile = "tmp-satcheck.commands"
evalResultFile  = "tmp-satcheck.result"

cleanup = mapM_ removeFile [ evalCommandFile, evalResultFile ]

evalInterp :: FilePath -> [Command] -> IO [Result]
evalInterp solver commands = do
  writeFile' evalCommandFile (unlines . map show $ commands)
--  r <- system (solver ++ " 2>/dev/null " ++ evalCommandFile ++ " > " ++ evalResultFile)
  r <- system (solver ++ " " ++ evalCommandFile ++ " > " ++ evalResultFile)
  case r of
    ExitSuccess   -> do
      str <- readFile' evalResultFile
      r <- mapM (\l -> evaluate (read l) `catch` \e -> putStrLn ("Could not parse line: " ++ l) >> throw e) (lines str)
--      r <- return $ map read . lines $ str
      return r
    ExitFailure i -> return []

evalExt :: ExtSolver -> [Command] -> IO [Result]
evalExt solver cs = E.run solver (P.run (mapM interpret cs))

----------------------------------------------------------------------------------------------------
-- correctness checking

modelCheck :: [Command] -> [Result] -> Bool
modelCheck = modelsok []
    where modelsok fs (DefProp _ _      : cs) (Ok      : rs) = modelsok fs cs rs
          modelsok fs (DefAdd _ _ _ _ _ : cs) (Ok      : rs) = modelsok fs cs rs
          modelsok fs (Freeze _         : cs) (Ok      : rs) = modelsok fs cs rs
          modelsok fs (Simplify _ _     : cs) (Ok      : rs) = modelsok fs cs rs
          modelsok fs (Assert f         : cs) (Ok      : rs) = modelsok (f:fs) cs rs
          modelsok fs (Solve fs'        : cs) (Model m : rs) = all (satisfies (Set.fromList [ v | Var v <- m ])) (fs ++ fs') 
                                                            && modelsok fs cs rs
          modelsok fs (Solve _          : cs) (Contr _ : rs) = modelsok fs cs rs
          modelsok fs (Assert _         : cs) (Contr []: rs) = modelsok fs cs rs
          modelsok fs (Simplify _ _     : cs) (Contr []: rs) = modelsok fs cs rs
          modelsok _  []                      []             = True
          modelsok _  _                       _              = False


weaken :: [Command] -> [Result] -> [Command]
weaken []             []              = []
weaken (Solve _      : cs) (Contr ls : rs) = Solve (map mkNot ls) : weaken cs rs
weaken (Solve _      : cs) (r        : rs) = weaken cs rs
weaken (Assert f     : cs) (Contr [] : rs) = [ Assert f, Solve [] ]
weaken (Simplify _ _ : cs) (Contr [] : rs) = [ Solve [] ]
weaken (Simplify _ _ : cs) (Ok       : rs) = weaken cs rs
weaken (c            : cs) (r        : rs) = c : weaken cs rs

contrCheck :: [Command] -> [Result] -> Bool
contrCheck (Solve _ : _ ) (Model _ : rs) = False
contrCheck (c       : cs) (r       : rs) = contrCheck cs rs
contrCheck []             []             = True
          

satisfies :: Set Var -> Formula -> Bool
satisfies set = value
   where value (Var v)           = Set.member v set
         value (App T   [])      = True
         value (App F   [])      = False
         value (App Not [f])     = not (value f)
         value (App And [f,g])   = value f && value g
         value (App Or  [f,g])   = value f || value g
         value (App Imp [f,g])   = not (value f) || value g
         value (App Equ [f,g])   = value f == value g
         value (App Xor [f,g])   = value f /= value g
         value (App Ite [c,t,f]) = if value c then value t else value f
         value x                 = error ("value: [" ++ show x ++"]")

----------------------------------------------------------------------------------------------------
-- convert to c++

sat2c :: [Command] -> String
sat2c cs = header ++ body ++ trailer
  where header  = "#include \"Prop.h\"\n\
                   \\n\
                   \int main(void){\n\
                   \   PropSolver " ++ solver ++ ";\n\
                   \   vec<Lit> __as;\n"
        trailer = "}\n"
        solver  = "__s"
        body    = commands2c solver Set.empty cs []


cmd2c :: String -> Command -> String -> String
cmd2c solver (DefProp v f     ) = ("   Lit "++) . (v++) . (" = "++) . formula2c solver f . (";\n"++)
cmd2c solver (DefAdd c s x y z) = ("   Lit "++) . (c++) . (", "++) . (s++) . (";\n"++) . ("   "++) . call solver "mkAdd" [x,y,z,Var c, Var s] . (";\n"++)
cmd2c solver (Freeze v        ) = ("   "++) . (solver++) . ('.':) . ("freezeVar(var("++) . (v++) . ("));\n"++)
cmd2c solver (Simplify x y    ) = ("   "++) . call solver "simplify" [Var (bool x), Var (bool y)] . (";\n"++)
cmd2c solver (Assert f        ) = 
  let fs = getClause f
  in ("   "++) . 
     call "__as" "clear" [] . ("; "++) .
     foldr (.) id (map (\l -> ("__as.push("++) . formula2c solver l . ("); "++)) fs) . ("\n"++) .
     ("   "++) . 
     call solver "addClause" [Var "__as"] . (";\n"++)

cmd2c solver (Solve fs        ) = ("   "++) . 
                                  call "__as" "clear" [] . ("; "++) .
                                  foldr (.) id (map (\l -> ("__as.push("++) . formula2c solver l . ("); "++)) fs) . ("\n"++) .
                                  ("   "++) . 
                                  call solver "solve" [Var "__as"] . (";\n"++)

commands2c :: String -> Set Var -> [Command] -> String -> String
commands2c _      _    [] = id
commands2c solver defs (c : cs) = define solver (Set.toList fv) . cmd2c solver c .
                                  commands2c solver (Set.union fv (boundVars c defs)) cs

  where fv = freeVars c Set.empty `Set.difference` defs

        boundVars (DefProp v _)      = Set.insert v
        boundVars (DefAdd c s _ _ _) = Set.insert c . Set.insert s
        boundVars _                  = id

        freeVars (DefProp _ f)       = formulaVars f
        freeVars (DefAdd _ _ x y z)  = formulaVars x . formulaVars y . formulaVars z
        freeVars (Assert f)          = formulaVars f
        freeVars (Freeze v)          = Set.insert v
        freeVars (Solve fs)          = foldr (.) id (map formulaVars fs)
        freeVars (Simplify _ _)      = id

        formulaVars (Var v)          = Set.insert v
        formulaVars (App _ fs)       = foldr (.) id (map formulaVars fs)

define solver vs = foldr (.) id (map (\v -> ("   Lit "++) . (v++) . (" = "++) . (solver++) . (".mkLit();\n"++)) vs)

bool :: Bool -> String
bool True  = "true"
bool False = "false"

params2c :: String -> [Formula] -> String -> String
params2c solver fs = foldr (.) id (intersperse (", "++) (map (formula2c solver) fs))

formula2c :: String -> Formula -> String -> String
formula2c solver (Var v)           = (v++)
formula2c solver (App T   [])      = ("lit_True"++)
formula2c solver (App F   [])      = ("lit_False"++)
formula2c solver (App Not [f])     = ('~':) . formula2c solver f
formula2c solver (App And [f,g])   = call solver "mkAnd" [f,g]
formula2c solver (App Or  [f,g])   = call solver "mkOr"  [f,g]
formula2c solver (App Imp [f,g])   = call solver "mkImp" [f,g]
formula2c solver (App Equ [f,g])   = call solver "mkEqu" [f,g]
formula2c solver (App Xor [f,g])   = call solver "mkXor" [f,g]
formula2c solver (App Ite [c,t,f]) = call solver "mkIte" [c,t,f]
formula2c _      _                 = error "illegal formula"
    
call :: String -> String -> [Formula] -> String -> String
call solver method fs = (solver++) . ('.':) . (method++) . ('(':) . params2c solver fs . (')':)

