module Hardware where

import SatMonad
import ExtSat
import SatWrappers

import Control.Monad ( zipWithM, foldM )
import Data.List     ( sortBy, groupBy )

import Debug.Trace ( trace )

----------------------------------------------------------------------------------------------------
-- Odd-Even sorter/merger

type Comp m a = (a -> a -> m (a,a))

merge :: SatMonad m => Comp m a -> a -> [a] -> [a] -> m [a]
merge comp minelem xs ys = 
    fmap (drop extra) (pow2merge comp (pad (2^power) minelem xs) (pad (2^power) minelem ys))
  where maxlen = max (length xs) (length ys)
        power  = ceiling (logBase 2 (fromIntegral maxlen))
        extra  = 2^(power+1) - length xs - length ys

pow2merge comp []  []             = return []
pow2merge comp [x] [y]            = fmap list (comp x y)
pow2merge comp xs@(_:_)  ys@(_:_) =
    do (e:es) <- pow2merge comp (evens xs) (evens ys)
       (os,o) <- pow2merge comp (odds  xs) (odds  ys) >>= return . \xs -> (init xs, last xs)
       ps     <- zipWithM comp es os
       return (e : [ z | (x,y) <- ps, z <- [x,y] ]  ++ [o])

sort :: SatMonad m => Comp m a -> a -> [a] -> m [a]
sort _    _   [] = return []
sort comp min xs = fmap (drop ((2^power) - length xs)) (pow2sort (pad (2^power) min xs))
  where power = ceiling (logBase 2 (fromIntegral (length xs)))

        pow2sort []    = return []
        pow2sort [x]   = return [x]
        pow2sort [x,y] = fmap list (comp x y)
        pow2sort elems =
            do xs <- pow2sort first
               ys <- pow2sort second
               pow2merge comp xs ys
          where (first, second) = splitAt (length elems `div` 2) elems

bincomp x y = 
    do min <- mkAnd x y
       max <- mkOr  x y
       return (min,max)

invcomp x y = fmap ( \(x,y) -> (y,x)) (bincomp x y)

----------------------------------------------------------------------------------------------------
-- Adder networks

type AdderNetwork m = [[Lit]] -> m [Lit]

dadaNetwork :: SatMonad m => AdderNetwork m
dadaNetwork []                     = return []
dadaNetwork [(x:y:ws)]             = dadaNetwork ((x:y:ws):[]:[])
dadaNetwork ([]:rest)              = dadaNetwork ([mkFalse]:rest)
dadaNetwork ([x]:rest)             = 
    do result <- dadaNetwork rest
       return (x:result)
dadaNetwork ([x,y]:next:rest)      =
    do (c,s)  <- halfAdd x y
       result <- dadaNetwork ((c:next):rest)
       return (s:result)
dadaNetwork ((x:y:z:ws):next:rest) = 
    do (c,s) <- mkAdd x y z
       dadaNetwork ((s:ws):(c:next):rest)

sortNetwork :: SatMonad m => AdderNetwork m
sortNetwork = sortNet [] 
  where 
    sortNet []   []       = return []
    sortNet prev []       = sortNet prev [[]]
    sortNet []   (xs:xss) = 
        do curr <- sort invcomp mkTrue xs
           step curr xss
    sortNet prev ([]:xss) = step prev xss
    sortNet prev (xs:xss) = 
        do curr  <- sort  invcomp mkTrue xs
           total <- merge invcomp mkTrue prev curr
           step total xss

    step []   rest = 
        do rs <- sortNet [] rest
           return (mkFalse:rs)
    step curr rest = 
        do s  <- mod2 curr
           rs <- sortNet (odds curr) rest
           return (s:rs)

    mod2 []       = return mkFalse
    mod2 [x]      = return x
    mod2 (x:y:xs) = 
        do z <- mod2 xs
           w <- mkAnd x (-y)
           mkOr w z
              

netAdd :: [[Lit]] -> [[Lit]] -> [[Lit]]
netAdd = zipp 
   where zipp (x:xs) (y:ys) = (x ++ y) : zipp xs ys
         zipp []     ys     = ys
         zipp xs     []     = xs

----------------------------------------------------------------------------------------------------
-- Squaring and Multiplication

multpp :: SatMonad m => [Lit] -> [Lit] -> m [[Lit]]
multpp xs ys = 
  mapM (mapM (uncurry mkAnd)) $ 
       weighted [ (a+b, (x,y)) | (x,a) <- zip xs [0..], (y,b) <- zip ys [0.. ] ]

squarepp :: SatMonad m => [Lit] -> m [[Lit]]
squarepp xs = 
  mapM (mapM (uncurry mkAnd)) $
       weighted ( [ (a+b+1, (x,y)) | (x,a) <- zip xs [0..], (y,b) <- zip xs [0.. ], a < b ]
                ++[ (a*2,   (x,x)) | (x,a) <- zip xs [0..] ])

squarepp_ xs = 
  map (map (uncurry (&&))) $
       weighted ( [ (a+b+1, (x,y)) | (x,a) <- zip xs [0..], (y,b) <- zip xs [0.. ], a < b ]
                ++[ (a*2,   (x,x)) | (x,a) <- zip xs [0..] ]
                )

weighted :: [(Int,a)] -> [[a]]
weighted = 
    fill 0
  . map (\ xs@((i,_):_) -> (i, map snd xs))
  . groupBy ( \ (x,_) (y,_) -> x == y)
  . sortBy  ( \ (x,_) (y,_) -> compare x y)

fill n [] = []
fill n ((i,xs) : rest) | n < i     = [] : fill (n+1) ((i,xs):rest)
                       | otherwise = xs : fill (n+1) rest
  
----------------------------------------------------------------------------------------------------
-- Binary number adding, parametrized by adder network

binAdder :: SatMonad m => AdderNetwork m -> [Lit] -> [Lit] -> m [Lit]
binAdder network xs ys = network (zipp xs ys)
  where zipp (x:xs) (y:ys) = [x,y] : zipp xs ys
        zipp []     ys     = [ [y] | y <- ys ]
        zipp xs     []     = [ [x] | x <- xs ]

----------------------------------------------------------------------------------------------------
-- simple helpers

list   (x,y)    = [x,y]

evens  (x:y:xs) = x : evens xs
evens  xs       = xs
odds            = evens . tail

pad :: Int -> a -> [a] -> [a]
pad n l ls | length ls < n = replicate (n - length ls) l ++ ls
           | otherwise     = ls

halfAdd x y =
    do c <- mkAnd x y
       s <- mkXor x y
       return (c,s)

----------------------------------------------------------------------------------------------------
-- properties

prop_ascending :: SatMonad m => [Lit] -> m Lit
prop_ascending (x:y:xs) =
  do r <- prop_ascending (y:xs)
     a <- mkOr (-x) y
     mkAnd a r
prop_ascending xs = return mkTrue

prop_equals :: SatMonad m => [Lit] -> [Lit] -> m Lit
prop_equals (x:xs) (y:ys) =
    do rs <- prop_equals xs ys
       r  <- mkEqu x y
       mkAnd r rs
prop_equals []     ys     = foldM mkAnd mkTrue [ -y | y <- ys ]
prop_equals xs     []     = foldM mkAnd mkTrue [ -x | x <- xs ]


zero :: [Lit]
zero = [ mkFalse ]

one :: [Lit]
one = [ mkTrue ]

prop_add_0 adder n = run minisat $
    do xs     <- sequence (replicate n mkLit)
       rs     <- adder zero xs
       eq     <- prop_equals xs rs
       failed <- solve [-eq]
       case failed of 
         True -> do xsv <- mapM modelValue xs
                    rsv <- mapM modelValue rs
                    return (Just (xsv,rsv))
         False -> return Nothing
       --return (not failed)

prop_add_1 adder n = run minisat $
    do xs  <- sequence (replicate n mkLit)
       ys  <- sequence (replicate n mkLit)
       xsp <- adder one xs
       lhs <- adder xsp ys
       xys <- adder xs ys
       rhs <- adder one xys
       eq     <- prop_equals lhs rhs
       failed <- solve [-eq]
       return (not failed)

prop_add_2 adder n = run minisat $
    do xs  <- sequence (replicate n mkLit)
       ys  <- sequence (replicate n mkLit)
       zs  <- sequence (replicate n mkLit)
       xys <- adder xs ys
       lhs <- adder xys zs
       yzs <- adder ys zs
       rhs <- adder xs yzs
       eq     <- prop_equals lhs rhs
       failed <- solve [-eq]
       return (not failed)

bitCount :: SatMonad m => [Lit] -> m [Lit]
bitCount xs = dadaNetwork [ [ x | x <- xs ] ]

prop_mergeok x y = run minisat $
    do xs <- sequence (replicate x mkLit)
       ys <- sequence (replicate y mkLit)
       rs <- pow2merge bincomp xs ys
       p1 <- prop_ascending xs
       p2 <- prop_ascending ys
       p3 <- prop_ascending rs
       failed <- solve [p1, p2, -p3]
       return (not failed)

prop_mergepermutes x y = run minisat $
    do xs <- sequence (replicate x mkLit)
       ys <- sequence (replicate y mkLit)
       rs <- merge bincomp mkFalse xs ys
       cnt1   <- bitCount xs
       cnt2   <- bitCount ys
       cnt3   <- binAdder dadaNetwork cnt1 cnt2
       cnt4   <- bitCount rs
       ok     <- prop_equals cnt3 cnt4
       failed <- solve [-ok]
       case failed of
         True -> do xsv <- mapM modelValue xs
                    ysv <- mapM modelValue ys
                    rsv <- mapM modelValue rs
                    return (Just (xsv, ysv, rsv))
         False -> return Nothing
--       return (not failed)

prop_sortok n = run minisat $
    do inputs  <- sequence (replicate n mkLit)
       outputs <- sort bincomp mkFalse inputs
       sorted  <- prop_ascending outputs
       cnt_in  <- bitCount inputs
       cnt_out <- bitCount outputs
       permutes<- prop_equals cnt_in cnt_out
       ok      <- mkAnd sorted permutes
       failed  <- solve [-ok]
       return (not failed)

assert :: [Lit] -> String -> ExtSatM ()
assert ls str =
       solve ls >>= \r ->
       lift (putStrLn $ "assertion " ++ show str ++ " " ++ if r then "failed" else "succeeded")

prop_multok network n = run minisat $
    do xs <- sequence (replicate n mkLit)
       ys <- sequence (replicate n mkLit)
       rs <- multpp xs ys >>= network

       xzero <- prop_equals zero xs
       rzero <- prop_equals zero rs
       xone  <- prop_equals one  xs
       equal <- prop_equals ys   rs

       assert [xzero, -rzero] "mult axiom 0"
       assert [xone, -equal]  "mult axiom 1"

       sr    <- multpp ys xs >>= network
       equal <- prop_equals rs sr
       assert [-equal] "mult axiom 2"

prop_squareok network n = run minisat $
    do xs  <- sequence (replicate n mkLit)
       lhs <- multpp xs xs >>= network
       rhs <- squarepp xs  >>= network

       equal <- prop_equals lhs rhs
       assert [-equal] "square axiom 0"
--       okay >>= \b -> if b
--         then do xsv  <- mapM modelValue xs
--                 lhsv <- mapM modelValue lhs
--                 rhsv <- mapM modelValue rhs
--                 return (Just (xsv, lhsv, rhsv))
--         else return Nothing
