module Prop 
    ( Var         -- *
    , Op(..)      -- *
    , Formula(..) -- *
    , Command(..) -- *
    , Result(..)  -- *
    , Env         -- *
    , InterpT     -- (* -> *) -> * -> *

    , mkTrue      -- Formula
    , mkFalse     -- Formula
    , mkNot       -- Formula -> Formula
    , mkAnd       -- Formula -> Formula -> Formula
    , mkOr        -- Formula -> Formula -> Formula
    , mkImp       -- Formula -> Formula -> Formula
    , mkEqu       -- Formula -> Formula -> Formula
    , mkXor       -- Formula -> Formula -> Formula
    , mkIte       -- Formula -> Formula -> Formula -> Formula
    , mkAdd       -- Formula -> Formula -> Formula -> (Formula,Formula)

    , getClause   -- Formula -> [Formula]

    , parse       -- String -> Either Command String
    , interpret   -- SatMonad m => Command -> InterpT m Result
    , run         -- SatMonad m => InterpT m a -> m a
    , lift        -- m a -> InterpT m a
    ) where

import Text.ParserCombinators.ReadP 
import Data.Map                     as Map ( Map, empty, member, lookup, insert, toList )
import Control.Monad.State          as S
import Data.Char                           ( isAlpha, isAlphaNum )

import qualified SatMonad as M
import SatMonad ( Lit, SatMonad )

----------------------------------------------------------------------------------------------------
-- Types

type Var = String

data Op = T | F | Not | And | Or | Imp | Equ | Xor | Ite | Add deriving (Eq, Ord)

data Formula
  = Var Var
  | App Op [Formula]
    deriving (Eq, Ord)

data Command 
  = DefProp Var Formula
  | DefAdd Var Var Formula Formula Formula
  | Assert Formula
  | Freeze Var
  | Unfreeze Var
  | Simplify Bool Bool
  | Solve  [Formula]
    deriving (Eq, Ord)

data Result          -- response to a single command
  = Ok
  | Model [Formula]
  | Contr [Formula]
  | Error String
    deriving Eq

----------------------------------------------------------------------------------------------------
-- functions

mkNot (App Not [f]) = f
mkNot f             = App Not [f]

mkTrue      = App T []
mkFalse     = App F []
mkAnd f g   = App And [f,g]
mkOr  f g   = App Or  [f,g]
mkImp f g   = App Imp [f,g]
mkEqu f g   = App Equ [f,g]
mkXor f g   = App Xor [f,g]
mkIte c t f = App Ite [c,t,f]
mkAdd a b c = App Add [a,b,c]

getClause :: Formula -> [Formula]
getClause (App Or [f,g]) = getClause f ++ getClause g
getClause x              = [x]

----------------------------------------------------------------------------------------------------
-- helpers

prec :: Op -> Int
prec Not = 5
prec And = 4
prec Or  = 3
prec o | elem o [Imp, Equ, Xor] = 2
       | otherwise              = 1

----------------------------------------------------------------------------------------------------
-- instance declarations for Read/Show

instance Show Op where
  showsPrec _ T   = ('1':)
  showsPrec _ F   = ('0':)
  showsPrec _ Not = ('~':)
  showsPrec _ And = ('&':)
  showsPrec _ Or  = ('|':)
  showsPrec _ Imp = ("=>"++)
  showsPrec _ Equ = ("=="++)
  showsPrec _ Xor = ('^':)
  showsPrec _ Ite = ("?:"++)
  showsPrec _ Add = ('+':)

instance Show Formula where
  showsPrec n (Var v)              = (v++)
  showsPrec n (App T [])           = ('1':)
  showsPrec n (App F [])           = ('0':)
  showsPrec n (App op@Ite [c,t,f]) = showParen (n >= p) $ showsPrec p c . (" ? "++) . showsPrec p t . (" : "++) . showsPrec p f
       where p = prec op
  showsPrec _ (App op@Add [c,t,f]) = showsPrec 0 c . (" ? "++) . showsPrec 0 t . (" : "++) . showsPrec 0 f
  showsPrec n (App op [f,g])       = showParen (if elem op [Xor, Imp, Equ] 
                                                then n >= p
                                                else n > p) $ showsPrec p f . (' ':) . shows op . (' ':) . showsPrec p g
       where p = prec op
  showsPrec n (App op@Not [f])     = ('~':) . showsPrec (prec op) f
  showsPrec _ (App op fs)          = showParen True (shows op) . shows fs


instance Show Command where
  showsPrec n (DefProp v f) = (v++) . (" = "++) . showsPrec n f
  showsPrec n (DefAdd c s x y z) = ('(':) . (c++) . (',':) . (s++) . (") = add("++)
                                 . showsPrec n x . (',':) . showsPrec n y . (',':) . showsPrec n z
                                 . (')':)
  showsPrec n (Freeze v) = ("freeze "++) . (v++)
  showsPrec n (Unfreeze v) = ("unfreeze "++) . (v++)
  showsPrec n (Simplify x y) = ("simplify"++) . ('(':) . sbool x . (',':) . sbool y . (')':)
     where sbool True  = ("true"++)
           sbool False = ("false"++)
  showsPrec n (Assert f) = showsPrec n f
  showsPrec n (Solve fs) = ("solve "++) . shows fs

instance Read Formula where
  readsPrec _ = readP_to_S p_formula

instance Read Command where
  readsPrec _ = readP_to_S p_command

instance Read Result where
  readsPrec n = readP_to_S p_result

    where 

instance Show Result where
  showsPrec n Ok         = ("ok"++)
  showsPrec n (Contr fs) = ("contr "++) . shows fs
  showsPrec n (Model fs) = ("model "++) . shows fs
  showsPrec n (Error e)  = ("error: "++) . (e++)

----------------------------------------------------------------------------------------------------
-- Parsers for all data types

parse :: String -> Either Command String
parse str = case readP_to_S (p_command >>= \r -> p_done >> return r) str of
  [(c, "")] -> Left c
  _         -> Right ("parse failed")

p_done = do
  skipSpaces
  rest <- look
  if null rest
     then return ()
     else pfail

p_var :: ReadP String
p_var = do
  x  <- satisfy isAlpha
  xs <- munch isAlphaNum
  return (x:xs)

p_formula :: ReadP Formula
p_formula = p_formula0
  where p_formula0  = (p_formula1 >>= \c -> skipSpaces >> char '?' >>
                       p_formula0 >>= \t -> skipSpaces >> char ':' >> p_formula0 >>= \f -> return (mkIte c t f))
                  +++  p_formula1
        p_formula1  = chainr1 p_formula2 (skipSpaces >> p_xor <++ p_imp +++ p_equ)
        p_formula2  = chainr1 p_formula3 (skipSpaces >> p_or)
        p_formula3  = chainr1 p_formula4 (skipSpaces >> p_and)

        p_formula4  = skipSpaces >> p_neg >>= \n -> p_formula4_ >>= \f -> return (n f)
        p_formula4_ = (char '(' >> p_formula0 >>= \f -> char ')' >> return f)
                  <++ (char '1' >> return mkTrue)
                  <++ (char '0' >> return mkFalse)
                  <++ (p_var >>= \v -> return (Var v))

        p_neg       = skipSpaces >> (char '~' >> p_neg >>= \n -> return (mkNot . n)) <++ (return id)
        p_and       = char '&'    >> return mkAnd
        p_or        = char '|'    >> return mkOr
        p_xor       = char '^'    >> return mkXor
        p_imp       = string "=>" >> return mkImp
        p_equ       = string "==" >> return mkEqu

p_command :: ReadP Command
p_command = skipSpaces >> 
          ( fmap Solve p_solve
        <++ fmap Freeze p_freeze
        <++ p_simp
        +++ fmap Assert p_formula
        +++ p_defprop
        +++ p_defadd )
   
  where p_solve   = string "solve" >> skipSpaces >> 
                  ( p_list p_formula <++ (return []) )
        p_simp    = string "simplify" >> p_tuple (skipSpaces >> p_bool) 2 >>= \ [x,y] -> return (Simplify x y)
        p_bool    = (string "true"  >> return True)
                <++ (string "false" >> return False)
          
        p_freeze  = string "freeze" >> skipSpaces >> p_var
        p_defprop = do
           v <- p_var
           skipSpaces
           char '='
           skipSpaces
           f <- p_formula
           return (DefProp v f)


        p_defadd = do
           [c,s] <- p_tuple (skipSpaces >> p_var) 2
           skipSpaces
           char '='
           skipSpaces
           string "add"
           [x,y,z] <- p_tuple p_formula 3
           return (DefAdd c s x y z)

p_result :: ReadP Result
p_result = skipSpaces >>
           (string "ok" >> return Ok)
       +++ (string "contr" >> fmap Contr (p_list p_formula))
       +++ (string "model" >> fmap Model (p_list p_formula))
       +++ (string "error:" >> skipSpaces >> fmap Error (munch ('\n'/=)))

p_list :: ReadP a -> ReadP [a]
p_list p = skipSpaces >> (char '[' >> skipSpaces >> sepBy p (skipSpaces >> char ',') >>= \r -> char ']' >> return r)

p_tuple :: ReadP a -> Int -> ReadP [a]
p_tuple p n = skipSpaces >> char '(' >> p_tuple_ n >>= \r -> skipSpaces >> char ')' >> return r
    where p_tuple_ 1 = fmap (\x -> [x]) p
          p_tuple_ n = 
              do r <- p
                 skipSpaces
                 char ','
                 rs <- p_tuple_ (n-1)
                 return (r:rs)

----------------------------------------------------------------------------------------------------
-- Interpreter for Commands

type Env         = Map String Lit
type InterpT m a = StateT Env m a


run :: SatMonad m => InterpT m a -> m a
run m = evalStateT m Map.empty

interpret :: SatMonad m => Command -> InterpT m Result
interpret (DefProp v f) = do
  l <- interpFormula f
  m <- S.get
  if Map.member v m
    then return (Error $ "variable " ++ v ++ " already defined")
    else do put (Map.insert v l m)
            return Ok
interpret (DefAdd c s x y z) = do
  xl <- interpFormula x
  yl <- interpFormula y
  zl <- interpFormula z
  m  <- S.get
  if Map.member c m
    then return (Error $ "variable " ++ c ++ " already defined")
    else if Map.member s m
           then return (Error $ "variable " ++ s ++ " already defined")
           else do (cl,sl) <- S.lift (M.mkAdd xl yl zl)
                   put (Map.insert c cl $ Map.insert s sl $ m)
                   return Ok
interpret (Freeze v) = 
  do l <- interpFormula (Var v)
     S.lift (M.freezeLit l)
     return Ok
interpret (Simplify x y) = 
  do b <- S.lift (M.simplify x y)
     return (if b then Ok else Contr [])

interpret (Assert f) = 
  do ls <- mapM interpFormula (getClause f)
     b  <- S.lift (M.clause ls)
     return (if b then Ok else Contr [])

interpret (Solve fs) | any (not . isAtom) fs = return (Error "formula not allowed")
                     | otherwise             =
  do ls <- mapM (\f -> interpFormula f >>= \l -> return (f, l)) fs
     b  <- S.lift (M.solve (map snd ls))
     if b
        then do m <- S.get
                model <- getModel m
                return (Model model)
        else do contr <- getContr ls
                return (Contr contr)

 where isAtom (Var _)                        = True
       isAtom (App op _) | elem op [Not,T,F] = True
                         | otherwise         = False

       getModel m = sequence [ value v l | (v,l) <- Map.toList m ]
         where value v l = S.lift (M.modelValue l) >>= \b -> return (if b then Var v else mkNot (Var v))

       getContr ls = do subset <- S.lift M.contr
                        return [ mkNot f | (f,l) <- ls, elem (-l) subset ]


interpFormula :: SatMonad m => Formula -> InterpT m Lit
interpFormula (Var v) = do
 do m <- S.get
    case Map.lookup v m of
      Just l  -> return l
      Nothing -> do l <- S.lift M.mkLit
                    put (Map.insert v l m)
                    return l
interpFormula (App T [])      = return M.mkTrue
interpFormula (App F [])      = return M.mkFalse
interpFormula (App Not [f])   = fmap negate (interpFormula f)
interpFormula (App op [f,g]) = do
  fl <- interpFormula f
  gl <- interpFormula g
  S.lift (binop op fl gl)
    where binop And = M.mkAnd
          binop Or  = M.mkOr
          binop Imp = \x y -> M.mkOr (-x) y
          binop Equ = M.mkEqu
          binop Xor = M.mkXor
interpFormula (App Ite [c,t,f]) = do
  cl <- interpFormula c
  tl <- interpFormula t
  fl <- interpFormula f
  S.lift (M.mkIte cl tl fl)
interpFormula _ = error "Illegal formula"
