module SatMonad 
    ( Lit(..)      -- *
    , SatMonad(..) -- class
    , mkTrue       -- Lit
    , mkFalse      -- Lit
    ) where

import Foreign.C.Types  ( CInt )
import Foreign.Storable ( Storable )


newtype Lit = Lit CInt deriving (Eq, Num, Ord, Storable)

instance Show Lit where
    showsPrec i (Lit l) = showsPrec i l

instance Read Lit where
    readsPrec i = map (\ (x,r) -> (Lit x, r)) . readsPrec i

mkTrue     = Lit 1
mkFalse    = -mkTrue

class (Functor m, Monad m) => SatMonad m where
    mkLit       :: m Lit
    clause      :: [Lit] -> m Bool

    solve       :: [Lit] -> m Bool
    solve_      :: Bool  -> [Lit] -> m Bool
    modelValue  :: Lit   -> m Bool
    value       :: Lit   -> m (Maybe Bool)
    contr       :: m [Lit]
    okay        :: m Bool

    -- for convenience
    lift        :: IO a -> m a

    -- default *with* simplification
    solve = solve_ True

    freezeLit   :: Lit   -> m ()
    unfreezeLit :: Lit   -> m ()
    simplify    :: Bool -> Bool -> m Bool

    freezeLit   l = return ()
    unfreezeLit l = return ()
    simplify  x y = return True
    
    -- formula creation (with simple defaults)
    mkAnd       :: Lit -> Lit -> m Lit
    mkAnd       = defAnd

    mkOr        :: Lit -> Lit -> m Lit
    mkOr x y    = mkAnd (-x) (-y) >>= return . negate

    mkEqu       :: Lit -> Lit -> m Lit
    mkEqu       = defEqu

    mkXor       :: Lit -> Lit -> m Lit
    mkXor x y   = mkEqu (-x) (-y) >>= return . negate

    mkIte       :: Lit -> Lit -> Lit -> m Lit
    mkIte       = defIte

    mkAdd       :: Lit -> Lit -> Lit -> m (Lit, Lit)
    mkAdd       = defAdd

defAnd :: SatMonad m => Lit -> Lit -> m Lit
defAnd x y =
    do r <- mkLit
       clause [-r, x]
       clause [-r, y]
       clause [-x, -y, r]
       return r

defEqu :: SatMonad m => Lit -> Lit -> m Lit
defEqu x y =
    do r <- mkLit
       clause [-r, -x,  y]
       clause [-r,  x, -y]
       clause [ r, -x, -y]
       clause [ r,  x,  y]
       return r

defIte :: SatMonad m => Lit -> Lit -> Lit -> m Lit
defIte c t f =
    do x <- mkAnd   c  t
       y <- mkAnd (-c) f
       mkOr x y

defAdd :: SatMonad m => Lit -> Lit -> Lit -> m (Lit,Lit)
defAdd x y z =
    do xy <- mkAnd x y
       xz <- mkAnd x z
       yz <- mkAnd y z
       c0 <- mkOr xy xz
       c  <- mkOr c0 yz
       s0 <- mkXor x y
       s  <- mkXor s0 z
       return (c,s)



