module Main ( main ) where

import System( getArgs )
import List( isPrefixOf, isSuffixOf, unfoldr, break )

import qualified Chalmers.QuickCheck as Q

import SatCheck hiding ( merge )

data ShrinkList
  = Log
  | Lin
  | LinLog
  | ListNone
    deriving Show

data ShrinkGap
  = All
  | Single
  | Full
  | GapNone
    deriving Show

data ShrinkMerge
  = MergeFull
  | MergeNone
    deriving Show

data Property
  = NoCrash
  | Correct
    deriving Show

data Params = Params {
   list  :: ShrinkList,
   gap   :: ShrinkGap,
   merge :: ShrinkMerge,
   prop  :: Property
 } deriving Show


shrinkList  Log       = Q.shrink
shrinkList  Lin       = shrinkLin
shrinkList  LinLog    = shrinkLinLog
shrinkList  ListNone  = const []

shrinkGap   All       = ungapAll
shrinkGap   Single    = ungapSingle
shrinkGap   Full      = ungapFull
shrinkGap   GapNone   = const []

shrinkMerge MergeFull = merges
shrinkMerge MergeNone = const []

shrinkProblem ps p = shrinkList (list ps) p ++ shrinkGap (gap ps) p ++ shrinkMerge (merge ps) p

--property :: Params -> Q.Property
property ps = case prop ps of
  NoCrash -> prop_NoCrash
  Correct -> prop_Correct

--property :: String -> Params -> Q.Property
--property solver ps = Q.forAllShrink Q.arbitrary (shrinkProblem ps) (p (prop ps) solver)
--  where p NoCrash = prop_NoCrash
--        p Correct = prop_Correct

parseParams :: Params -> [String] -> (Params, [String])
parseParams p [] = (p, [])
parseParams p ("--gap-all"      : args) = parseParams (p { gap   = All })       args
parseParams p ("--gap-single"   : args) = parseParams (p { gap   = Single })    args
parseParams p ("--gap-full"     : args) = parseParams (p { gap   = Full })      args
parseParams p ("--gap-none"     : args) = parseParams (p { gap   = GapNone })   args
parseParams p ("--merge-none"   : args) = parseParams (p { merge = MergeNone }) args
parseParams p ("--merge-full"   : args) = parseParams (p { merge = MergeFull }) args
parseParams p ("--list-log"     : args) = parseParams (p { list  = Log })       args
parseParams p ("--list-lin"     : args) = parseParams (p { list  = Lin })       args
parseParams p ("--list-linlog"  : args) = parseParams (p { list  = LinLog })    args
parseParams p ("--list-none"    : args) = parseParams (p { list  = ListNone })  args
parseParams p ("--prop-correct" : args) = parseParams (p { prop  = Correct })   args
parseParams p ("--prop-nocrash" : args) = parseParams (p { prop  = NoCrash })   args
parseParams p (a : args)                = let (ps, args') = parseParams p args in (ps, a : args')

defaults = Params { list = Log, gap = GapNone, merge = MergeNone, prop = Correct }

main = getArgs >>= \args ->
  case parseParams defaults args of 
     -- run quickcheck
     (ps, [solver]) -> Q.quickCheckWith 100 500 100 $ Q.forAllShrink Q.arbitrary (shrinkProblem ps) (property ps solver)

     -- run quickcheck repeatedly
     (ps, ["--repeat", solver]) -> 
          let loop = do
               b <- Q.quickCheckWith 100 500 100 $ Q.forAllShrink Q.arbitrary (shrinkProblem ps) (property ps solver)
               if b then loop else return False
          in  loop

     -- read a file with satcommands
     (ps, [solver, sat]) | isSuffixOf ".sat" sat -> do
         commands <- readFile sat >>= return . map read . lines
         Q.quickCheckWith 1 1 1 $ Q.shrinking (shrinkProblem ps) (SatProblem (length commands-1) commands) (property ps solver)

     -- read a dimacs file and convert to satcommands
     (ps, [solver, cnf]) | isSuffixOf ".cnf" cnf -> do
         dimacs <- readFile cnf
         Q.quickCheckWith 1 1 1 $ Q.shrinking (shrinkProblem ps) (fromDimacs dimacs) (property ps solver)

     -- read a file with satcommands, test that two solvers behave equally
     (ps, [s1, s2, sat]) | isSuffixOf ".sat" sat -> do
         commands <- readFile sat >>= return . map read . lines
         Q.quickCheckWith 1 1 1 $ Q.shrinking (shrinkProblem ps) (SatProblem (length commands-1) commands) (prop_Equal s1 s2)

     -- read a dimacs file and convert to satcommands, test that two solvers behave equally
     (ps, [s1, s2, cnf]) | isSuffixOf ".cnf" cnf -> do
         dimacs <- readFile cnf
         Q.quickCheckWith 1 1 1 $ Q.shrinking (shrinkProblem ps) (addStatistics . fromDimacs $ dimacs) (prop_Equal s1 s2)

     (ps, args)  -> do
        putStrLn "No matching action."
        putStrLn ("( params = " ++ show ps ++ " args = " ++ show args ++ " )")
        return True
                  


interleaveStatistics (SatProblem _ p) = SatProblem (l-1) p'
  where p' = foldr (.) id (map (\a xs -> a : Statistics : xs) p) []
        l  = length p'

addStatistics (SatProblem c p) = SatProblem (c+1) (p ++ [ Statistics ] )

fromDimacs :: String -> SatProblem
fromDimacs = 
   (\cs -> SatProblem (length cs - 1) cs)
 . (++ [ExpensiveSimplify, Solve []])
 . (\ (v, cs) -> map (Clause . map Lit) cs ++ [ FreezeVar v | v <- [1 .. v] ])
 . (\cs -> (vars cs, cs))
 . unfoldr clause
 . map read
 . words
 . unlines 
 . filter (not . isPrefixOf "c")
 . filter (not . isPrefixOf "p cnf")
 . lines 

 where clause :: [Int] -> Maybe ([Int],[Int])
       clause [] = Nothing
       clause ts = Just . (\ (x,0:y) -> (x,y)) $ break (==0) ts

       vars :: [[Int]] -> Int
       vars = maximum . concat