module SatCheck where

import Prelude hiding ( catch )

import List( intersperse, isPrefixOf, sort, find, (\\) )
import Monad( when, liftM )
import Directory( removeFile, doesFileExist )
import Maybe( isNothing, fromJust )

import Control.Exception ( catch, throw, evaluate )

import System.IO( hClose )
import System.Exit
import System.Cmd 

import qualified Data.Set as Set ( fromList, member )

import Chalmers.QuickCheck
import Chalmers.QuickCheck.Monadic

import Debug.Trace ( trace )

----------------------------------------------------------------------------------------------------
-- Data types

newtype Lit        = Lit Int deriving (Eq, Num, Ord)
data    SatProblem = SatProblem Int [SatCommand] deriving (Eq, Show)

data SatCommand
  = Clause [Lit]
  | Solve  [Lit]
  | FreezeVar Int
  | BasicSimplify
  | ExpensiveSimplify
  | Statistics
  | Set Parameter
    deriving (Eq, Show, Read)

data Parameter
  = NoCoreDump
    deriving (Eq, Show, Read)

data SatResult
  = Model [Lit]
  | Contr [Lit]
  | Stat { clauses :: Int, vars :: Int, conflicts :: Int }
  | Ok
    deriving (Eq,Read,Show)


instance Read Lit where
  readsPrec i = map (\ (x, y) -> (Lit x, y)) . readsPrec i

instance Show Lit where
  show (Lit i) = show i

unLit (Lit i) = i

----------------------------------------------------------------------------------------------------
-- shrink functions

shrinkListLin :: Int -> [a] -> [(Int,[a])]
shrinkListLin _ []                  = []
shrinkListLin n xs | n >= length xs = error "shrinkList"
                   | otherwise      = [ ( if i == 0 then length xs - 2 else i - 1
                                        , dropIndex i xs) | i <- reverse ([n+1 .. length xs-1] ++ [0 .. n]) ]

shrinkListLinLog :: Int -> [a] -> [(Int,[a])]
shrinkListLinLog _ []                  = []
shrinkListLinLog n xs | n >= length xs = error "shrinkList"
                      | otherwise      = [ let r = slice j i xs in
                                           ( if j == 0 then length r - 1 else j - 1
                                           , r)
                                           | i <- reverse ([n+1 .. length xs-1] ++ [0 .. n])
                                           , j <- shrink i ++ [i]
                                         ]

slice :: Int -> Int -> [a] -> [a]
slice x y xs = take x xs ++ drop (y+1) xs

dropIndex :: Int -> [a] -> [a]
dropIndex _ []     = []
dropIndex 0 (x:xs) = xs
dropIndex n (x:xs) = x : dropIndex (n-1) xs

instance Arbitrary Lit where
  arbitrary = Lit `fmap` (arbitrary `suchThat` (/=0))
  shrink    = const []

instance Arbitrary SatProblem where
  arbitrary =
    do cs1 <- arbitrary
       cs2 <- arbitrary
       cs3 <- arbitrary
       cs4 <- arbitrary
       cs5 <- arbitrary
       let commands = cs1 ++ cs2 ++ cs3 ++ cs4 ++ cs5 ++ [Solve []]
       return (SatProblem (length commands - 1) commands)

  -- default shrinking
  shrink (SatProblem i cs) = map (SatProblem i) (shrink cs)

-- alternative shrinking
shrinkLin, shrinkLinLog, ungapAll, ungapSingle, ungapFull :: SatProblem -> [SatProblem]

shrinkLin    (SatProblem i cs) = [ SatProblem i' cs' | (i', cs') <- shrinkListLin i cs ]
shrinkLinLog (SatProblem i cs) = [ SatProblem i' cs' | (i', cs') <- shrinkListLinLog i cs ]
ungapFull    p                 = ungapAll p ++ ungapSingle p
ungapSingle  (SatProblem i cs) = [ SatProblem i (ungap g cs) | g <- gaps cs ]
ungapAll     (SatProblem i cs) | null (gaps cs) = []
                               | otherwise      = [ SatProblem i (ungapAll_ cs) ]
  where ungapAll_ cs | null gs   = cs
                     | otherwise = ungapAll_ (ungap (head gs) cs)
           where gs = gaps cs

----------------------------------------------------------------------------------------------------
-- helpers

variables cs = sort . map (\(Lit l) -> abs l) $ concatMap lits cs
  where
    lits (Clause ls)   = ls
    lits (Solve  ls)   = ls
    lits (FreezeVar v) = [Lit v]
    lits _             = []

----------------------------------------------------------------------------------------------------
-- remove variable gaps

ungap :: Int -> [SatCommand] -> [SatCommand]
ungap g = map ungapCommand
  where ungapCommand (Clause ls)               = Clause (map ungapLit ls)
        ungapCommand (Solve  ls)               = Solve  (map ungapLit ls)
        ungapCommand (FreezeVar v) | v < g     = FreezeVar v
                                   | otherwise = FreezeVar (v-1)
        ungapCommand x                         = x

        ungapLit l@(Lit n) | abs n < g = l
                           | n < 0     = Lit (1 - abs n)
                           | otherwise = Lit (abs n - 1)

gaps :: [SatCommand] -> [Int]
gaps cs = gapsInt (0:variables cs)

gapsInt :: [Int] -> [Int]
gapsInt (x:y:xs) = [x+1 .. y-1] ++ gapsInt (y:xs)
gapsInt _        = []

----------------------------------------------------------------------------------------------------
-- merge variables

merge :: Lit -> Lit -> [SatCommand] -> [SatCommand]
merge x y cs | x == y    = error "Trying to merge equal literals"
             | cs == map substCommand cs = error ("merge " ++ show (unLit x) ++ " and " ++ show (unLit y) ++ " gives fixedpoint")
             | otherwise = map substCommand cs
 where substCommand (Clause ls)                         = Clause (map substLit ls)
       substCommand (Solve  ls)                         = Solve  (map substLit ls)
       substCommand (FreezeVar v) | v ==  abs (unLit x) = FreezeVar (abs (unLit y))
                                  | otherwise           = FreezeVar v
       substCommand c                                   = c

       substLit l | l ==  x   = y
                  | l == -x   = -y
                  | otherwise = l

merges :: SatProblem -> [SatProblem]
merges (SatProblem c cs) = [ SatProblem c (merge (Lit x) (Lit y) cs)
                             | x <- variables cs, z <- [1 .. x-1], y <- [z, -z] ]

----------------------------------------------------------------------------------------------------
-- test functions

evalCommandFile = "satcheck.eval.commands"
evalResultFile  = "satcheck.eval.result"
judgeCnfFile    = "satcheck.judge.cnf"
judgeResultFile = "satcheck.judge.result"

cleanUp :: IO ()
cleanUp = 
  mapM_ remove
    [ evalCommandFile
    , evalResultFile
    , judgeCnfFile
    , judgeResultFile
    ]

  where remove f = do
          b <- doesFileExist f
          when b (removeFile f)

eval :: FilePath -> [SatCommand] -> IO [SatResult]
eval solver commands = do
  writeFile evalCommandFile (unlines . map show $ commands)
  r <- system (solver ++ " 2>/dev/null -i < " ++ evalCommandFile ++ " > " ++ evalResultFile)
  case r of
    ExitSuccess   -> do
      str <- readFile' evalResultFile
      r <- mapM (\l -> evaluate (read l) `catch` \e -> writeFile "apa.sat" (unlines . map show $ commands) >> throw e) (lines str)
--      r <- mapM (\l -> evaluate (read l) `catch` \e -> putStrLn ("Could not parse line: " ++ l) >> throw e) (lines str)
--      r <- return $ map read . lines $ str
      return r
    ExitFailure i -> return []

limmat = "limmat"

contradictory :: [[Lit]] -> String -> PropertyM IO ()
contradictory clauses failwith = do
  r <- run (judge limmat clauses)
  when r (fail failwith)

satisfies :: [Lit] -> [[Lit]] -> Bool
satisfies _  [] = True
satisfies ls cs = modelok && all sat cs
  where modelok = isPrefixOf vars $ map (abs . unLit) ls 
        vars    = [1 .. maximum . map (abs . unLit) . concat $ cs]
        set     = Set.fromList ls
        sat     = any (flip Set.member set)

checkOne :: [[Lit]] -> SatCommand -> SatResult -> PropertyM IO ()
checkOne clauses (Clause c)        (Contr [])   = contradictory (c:clauses) "Incorrect contradiction after <Clause>"
checkOne clauses (Clause _)        Ok           = return ()
checkOne clauses (FreezeVar v)     Ok           = return ()
checkOne clauses ExpensiveSimplify Ok           = return ()
checkOne clauses ExpensiveSimplify (Contr [])   = contradictory clauses "Incorrect contradiction after <ExpensiveSimplify>"
checkOne clauses BasicSimplify     Ok           = return ()
checkOne clauses BasicSimplify     (Contr [])   = contradictory clauses "Incorrect contradiction after <BasicSimplify>"
checkOne clauses Statistics        (Stat _ _ _) = return ()
checkOne clauses (Solve as)        (Contr cs) 
   | not $ all (flip elem as . negate) cs       = fail "Incorrect reason for contradiction"
   | otherwise                                  = contradictory ([[-x] | x <- cs] ++ clauses) "Insufficient reason for contradiction"
checkOne clauses (Solve as)        (Model ls) 
   | not $ all (flip elem ls) as                = fail "Model does not respect assumptions"
   | otherwise                                  = when (not $ satisfies ls clauses) $ fail "Incorrect model"
checkOne clauses (Set _)           Ok           = return ()
checkOne clauses c                 _            = error ("Unexpected response to command: " ++ show c)

checkResult = check []
 where  check _       []     []     = return ()
        check clauses (c:cs) (r:rs) = checkOne clauses c r >> check (extend c clauses) cs rs
        check _       _      _      = error "Size of response differs from number of commands"

        extend (Clause c) cs        = c : cs
        extend _          cs        = cs

check :: FilePath -> [SatCommand] -> [SatResult] -> IO (Maybe String)
check solver cs rs = checkAccum [] cs rs

  where checkAccum :: [[Lit]] -> [SatCommand] -> [SatResult] -> IO (Maybe String)
        checkAccum _       []     []     = return Nothing
        checkAccum clauses (c:cs) (r:rs) = do
           when verbose $ putStr ("checking: " ++ show c ++ " : " ++ show r)
           r <- one clauses c r
           when verbose $ putStrLn (if isNothing r then " (Ok)" else (" (Fail)"))
           if isNothing r then checkAccum (extend c clauses) cs rs else return r
        checkAccum _       _      _      = return (Just "Unexpected result/Crash")

        extend :: SatCommand -> [[Lit]] -> [[Lit]]
        extend (Clause c) clauses = c : clauses
        extend _          clauses = clauses

        one :: [[Lit]] -> SatCommand -> SatResult -> IO (Maybe String)
--        one = undefined
        one clauses (Clause c) (Contr []) = do
             r <- judge solver (c:clauses)
             return $ if not r then Nothing else Just "Incorrect contradiction after <Clause>"
        one clauses (Clause _) Ok         = return Nothing
        one clauses (Clause _) _          = error "Unexpected response to <Clause>" -- return False ?
        one clauses (FreezeVar v) Ok      = return Nothing
        one clauses ExpensiveSimplify  Ok = return Nothing
        one clauses ExpensiveSimplify  (Contr []) = do
            r <- judge solver clauses
            return $ if not r then Nothing else Just "Incorrect contradiction after <ExpensiveSimplify>"
        one clauses BasicSimplify  Ok = return Nothing
        one clauses BasicSimplify  (Contr []) = do
            r <- judge solver clauses
            return $ if not r then Nothing else Just "Incorrect contradiction after <BasicSimplify>"
        one clauses (Solve as) (Contr c)  = do
             r  <- judge solver ([[-x] | x <- c] ++ clauses)
             if r then
               do return $ Just "Insufficient reason for contradiction"
              else
               do {-
                  rs <- sequence [ judge solver ([[-x] | x <- c'] ++ clauses)
                                 | x <- c
                                 , let c' = c \\ [x]
                                 ]
                  if any not rs then
                    do return $ Just "Reason was not minimal"
                   else
                  -}
                     do return $
                          if not $ all (flip elem as . negate) c then
                            Just "Incorrect reason for contradiction"
                           else
                            Nothing

        one clauses (Solve as) (Model m)  = do
             r <- judge solver ([[x] | x <- m ] ++ clauses)
             return $ if not r then
               Just "Incorrect model"
              else if not $ all (flip elem m) as then
                Just "Model does not respect assumptions" 
               else
                 Nothing

        one clauses (Solve  _) _          = error "Unexpected response to <Solve>" -- return False ?
        one clauses (Set _)    Ok         = return Nothing
        one clauses (Set _)    _          = error "Unexpected response to <Set>" -- return False ?


judge :: FilePath -> [[Lit]] -> IO Bool
judge solver clauses = do
  writeFile judgeCnfFile (dimacs clauses)
  system (solver ++ " " ++ judgeCnfFile ++ " >" ++ judgeResultFile)
  strs <- readFile' judgeResultFile >>= return . filter (isPrefixOf "s ") . lines
  case strs of
    ["s SATISFIABLE"]   -> return True
    ["s UNSATISFIABLE"] -> return False
    _                   -> error $ "Weird result from judge <" ++ solver ++ ">" ++ ": " ++ show strs
  

dimacs :: [[Lit]] -> String
dimacs clauses = concat $ ["p cnf ", show vars, " ", show (length clauses), "\n"] ++ 
                 concatMap (\c -> intersperse " " (map show c) ++ [" 0\n"]) clauses
  where vars = maximum (0 : map (abs . (\ (Lit i) -> i)) (concat clauses))

isClause, isSolve :: SatCommand -> Bool
isClause (Clause _) = True
isClause _          = False
isSolve  (Solve _)  = True
isSolve  _          = False

----------------------------------------------------------------------------------------------------
-- test case generator

instance Arbitrary SatCommand where
{-
  arbitrary = frequency 
    [ (1,   return $ Clause [])
    , (500, liftM Clause (arbitrary `suchThat` (not . null)))
    , (100, liftM Solve arbitrary)
    ]
-}

  arbitrary =
    frequency
    [ (9000, arbitraryClause)
    , (100,  arbitrarySolve)
    , (1,    return $ Clause [])
    ]
   where
    arbitrarySolve =
      liftM (Solve . unique) (vect 5)

    arbitraryClause =
      liftM (Clause . unique) ((vect 7) `suchThat` ((> 1) . length))

    unique [] = []
    unique (x:xs) | x     `elem` xs = unique xs
                  | -x    `elem` xs = unique xs
                  | otherwise       = x : unique xs

    vect n =
      sized $ \m ->
        do k <- choose (0,(m `div` 7) `min` n)
           vector k

  shrink (Clause c) = map Clause (shrink c)
  shrink (Solve  a) = map Solve  (shrink a)
  shrink x          = []

----------------------------------------------------------------------------------------------------
-- helper

readFile' :: FilePath -> IO String
readFile' file =
  do s <- readFile file
     (s == s) `seq` return s

----------------------------------------------------------------------------------------------------
-- debug

simpleTestCase = 
 [ Clause [-1, 2]
 , Clause [-2, 3]
 , Clause [-3, 1]
 , Solve  [1]
 , Solve  [-2]
 ]

-- eval "../current-base/minisat" simpleTestCase >>= print
simpleResult = [Ok,Ok,Ok,Model [1,2,3],Model [-1,-2,-3]] 

verbose = False

-- check "limmat" simpleTestCase simpleResult >>= print

----------------------------------------------------------------------------------------------------
-- debug

-- Unexpected result/Crash
program = 
  [ Clause [-439,342]
  , Clause [98,610]
  , Clause [437,-50]
  , Clause [-8,529]
  , Clause [-425,-667]
  , Clause [666,539]
  , Clause [606,598]
  , Clause [303,-23]
  , Clause [341,-114]
  , Clause [-710]
  , Solve []
  ]

----------------------------------------------------------------------------------------------------
-- properties

prop_Equal s1 s2 (SatProblem _ p) = monadicIO $ do
   monitor $ whenFail' $ writeFile "bug.sat" script
   r1 <- run $ eval s1 commands
   r2 <- run $ eval s2 commands
   assert (r1 == r2)
          
 where
   commands = Set NoCoreDump : p
   script   = unlines (map show p)

prop_NoCrash solver (SatProblem _ p) = monadicIO $ do
  monitor $ whenFail' $ writeFile "bug.sat" ("% Crash/Unexpected Result"  ++ script)
  res <- run $ eval solver commands
  assert (length res == length commands)

 where
  commands = Set NoCoreDump : p
  script   = unlines (map show p)

prop_Correct solver (SatProblem _ p) = monadicIO $ do 
  monitor $ whenFail' $ writeFile "bug.sat" script
  res <- run $ eval solver commands
  assert (length res == length commands)

  let first5 = take 5
        [ taBort r
        | let taBort (Model _) = Model []
              taBort (Contr _) = Contr []
        , r <- res
        , r /= Ok
        ]

  monitor $ collect $ first5
  monitor $ cover (Contr [] `elem` first5) 40 "contradiction"

  checkResult commands res

 where
  commands = Set NoCoreDump : p
  script   = unlines (map show p)

loops history x | x `elem` history = (x:history)
                | otherwise        = 
 case find (not . null) $ map (loops (x:history)) (shrink x) of
   Nothing -> []
   Just l  -> l


shrinksWithin 0 x = False
shrinksWithin n x =
  case shrink x of
    [] -> True
    xs -> all (shrinksWithin (n-1)) xs

prop_LooparEj (Fixed (x :: SatProblem)) =
  shrinksWithin 1000 x

prop_LooparEj2 (Fixed (x :: SatProblem)) =
    let loop = loops [] x
    in flip whenFail (null loop) $ do
      putStrLn "Loop: "
      mapM_ (putStrLn . show) (reverse loop)
